SystemOrganization addCategory: #'Zinc-HTTP'!
SystemOrganization addCategory: #'Zinc-HTTP-Client-Server'!
SystemOrganization addCategory: #'Zinc-HTTP-Core'!
SystemOrganization addCategory: #'Zinc-HTTP-Exceptions'!
SystemOrganization addCategory: #'Zinc-HTTP-Logging'!
SystemOrganization addCategory: #'Zinc-HTTP-Streaming'!
SystemOrganization addCategory: #'Zinc-HTTP-Support'!
SystemOrganization addCategory: #'Zinc-HTTP-Variables'!

ZnUrlOperation subclass: #ZnHttpRetrieveContents
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Client-Server'!

!ZnHttpRetrieveContents commentStamp: '' prior: 0!
I am ZnHttpRetrieveContents.
I am a ZnUrlOperation.

I implement the #retrieveContents URL operation for HTTP(S) urls.

  'http://zn.stfx.eu/zn/numbers.txt' asZnUrl retrieveContents.

Part of Zinc HTTP Components.
!

!ZnHttpRetrieveContents class methodsFor: 'accessing' stamp: ''!
operation
	^ #retrieveContents! !

!ZnHttpRetrieveContents class methodsFor: 'accessing' stamp: ''!
schemes
	^ #( #http #https )! !

!ZnHttpRetrieveContents methodsFor: 'operations' stamp: ''!
performOperation
	"Download and return the HTTP(S) resource that url refers to.
	This will typically return a String or a ByteArray (see ZnClient>>#get:).
	If something goes wrong, an exception will be signalled."

	^ ZnClient new
		beOneShot;
		enforceHttpSuccess: true;
		get: self url! !

ZnUrlOperation subclass: #ZnHttpSaveContentsToFile
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Client-Server'!

!ZnHttpSaveContentsToFile commentStamp: '' prior: 0!
I am ZnHttpSaveContentsToFile.
I am a ZnUrlOperation.

I implement the #saveContentsToFile: URL operation for HTTP(S) urls, which downloads the url resource to a file.

  'http://zn.stfx.eu/zn/numbers.txt' asZnUrl saveContentsToFile: Path * 'numbers.txt'.

Part of Zinc HTTP Components.!

!ZnHttpSaveContentsToFile class methodsFor: 'accessing' stamp: ''!
operation
	^ #saveContentsToFile:! !

!ZnHttpSaveContentsToFile class methodsFor: 'accessing' stamp: ''!
schemes
	^ #( #http #https )! !

!ZnHttpSaveContentsToFile methodsFor: 'operations' stamp: ''!
performOperation
	"Download the HTTP(S) resource specified by url to the file in argument.
	Return a file reference on the file that's just been downloaded.
	If something goes wrong, an exception will be signalled."

	ZnClient new
		enforceHttpSuccess: true;
		url: self url;
		downloadTo: argument.
	^ argument asFileReference! !

Announcement subclass: #ZnLogEvent
	instanceVariableNames: 'timestamp id'
	classVariableNames: 'IdCounter LogEventAnnouncer'
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnLogEvent commentStamp: '' prior: 0!
I am ZnLogEvent, the base class of a log events emitted by elements of the Zinc HTTP Components framework.

I add a timestamp and a simple id attribute. The id can wrap around and should only be used to distinguish between events that have the same timestamp.

ZnLogEvents are distributed as Announcement through a singleton Announcer that I maintain.

I have a small convenience API to log to the Transcript or open a simple GUI on the emitted log events.!

ZnLogEvent subclass: #ZnClientLogEvent
	instanceVariableNames: 'clientId'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnClientLogEvent commentStamp: '' prior: 0!
I am ZnClientLogEvent, a ZnLogEvent generated by ZnClient.

I hold an optional clientId attribute.
!

ZnClientLogEvent subclass: #ZnClientConnectionClosedEvent
	instanceVariableNames: 'address port'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnClientConnectionClosedEvent commentStamp: '' prior: 0!
ZnClientConnectionClosedEvent signals that a client side connection to address:port is closed.!

!ZnClientConnectionClosedEvent methodsFor: 'accessing' stamp: ''!
address
	^ address! !

!ZnClientConnectionClosedEvent methodsFor: 'accessing' stamp: ''!
address: anObject
	address := anObject! !

!ZnClientConnectionClosedEvent methodsFor: 'accessing' stamp: ''!
port
	^ port! !

!ZnClientConnectionClosedEvent methodsFor: 'accessing' stamp: ''!
port: anObject
	port := anObject! !

!ZnClientConnectionClosedEvent methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 2/2/2018 09:32'!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Connection Closed '.
	address 
		ifNil: [ 
			stream print: address ]
		ifNotNil: [ 
			address 
				do: [ :each | stream print: each ] 
				separatedBy: [ stream nextPut: $. ] ].
	stream nextPut: $:; print: port! !

ZnClientLogEvent subclass: #ZnClientFollowingRedirectEvent
	instanceVariableNames: 'target'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnClientFollowingRedirectEvent commentStamp: '' prior: 0!
ZnClientFollowingRedirectEvent signals that a redirect to target is being followed. How many redirects to follow, if any is configurable.!

!ZnClientFollowingRedirectEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Following Redirect '; print: target! !

!ZnClientFollowingRedirectEvent methodsFor: 'accessing' stamp: ''!
target
	^ target! !

!ZnClientFollowingRedirectEvent methodsFor: 'accessing' stamp: ''!
target: anObject
	target := anObject! !

ZnClientLogEvent subclass: #ZnClientIgnoringExceptionOnConnectionReuseEvent
	instanceVariableNames: 'exception'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnClientIgnoringExceptionOnConnectionReuseEvent commentStamp: '' prior: 0!
ZnClientIgnoringExceptionOnConnectionReuseEvent signals that a client connection was reused and that an exception occurred. Normally, there is one retry after that, trying to reconnect.
!

!ZnClientIgnoringExceptionOnConnectionReuseEvent methodsFor: 'accessing' stamp: ''!
exception
	^ exception! !

!ZnClientIgnoringExceptionOnConnectionReuseEvent methodsFor: 'accessing' stamp: ''!
exception: anObject
	exception := anObject! !

!ZnClientIgnoringExceptionOnConnectionReuseEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Ignoring Exception On Connection Reuse '; print: exception! !

!ZnClientLogEvent methodsFor: 'accessing' stamp: ''!
clientId
	^ clientId! !

!ZnClientLogEvent methodsFor: 'accessing' stamp: ''!
clientId: anObject
	clientId := anObject! !

!ZnClientLogEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	clientId ifNotNil: [ stream space; << clientId ]! !

ZnClientLogEvent subclass: #ZnClientRetryingEvent
	instanceVariableNames: 'exception'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnClientRetryingEvent commentStamp: '' prior: 0!
ZnClientRetryingEvent signals a client is retrying is main operation. The number of retries is configurable.
!

!ZnClientRetryingEvent methodsFor: 'accessing' stamp: ''!
exception
	^ exception! !

!ZnClientRetryingEvent methodsFor: 'accessing' stamp: ''!
exception: anObject
	exception := anObject! !

!ZnClientRetryingEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Retrying '; print: exception! !

ZnClientLogEvent subclass: #ZnClientTransactionEvent
	instanceVariableNames: 'request response requestDuration responseDuration'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnClientTransactionEvent commentStamp: '' prior: 0!
ZnClientTransactionEvent is a ZnClientLogEvent that notes the completion of a client side HTTP transaction (request/response).

Instance Variables
	request: a ZnRequest
	requestDuration:	 milliseconds
	response:	a ZnResponse
	responseDuration: milliseconds
	
Note that this can be quite large since it includes the entities transferred. See ZnSimplifiedClientTransactionEvent for a more compact representation.!

!ZnClientTransactionEvent methodsFor: 'accessing' stamp: ''!
duration
	^ requestDuration + responseDuration! !

!ZnClientTransactionEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream space.
	request requestLine printMethodAndUriOn: stream.
	stream space; print: response code.
	response hasEntity
		ifTrue: [ 
			stream space; print: response entity contentLength; nextPut: $B ].
	stream space; print: self duration; nextPutAll: 'ms'! !

!ZnClientTransactionEvent methodsFor: 'accessing' stamp: ''!
request
	^ request! !

!ZnClientTransactionEvent methodsFor: 'accessing' stamp: ''!
request: anObject
	request := anObject! !

!ZnClientTransactionEvent methodsFor: 'accessing' stamp: ''!
requestDuration
	^ requestDuration! !

!ZnClientTransactionEvent methodsFor: 'accessing' stamp: ''!
requestDuration: anObject
	requestDuration := anObject! !

!ZnClientTransactionEvent methodsFor: 'accessing' stamp: ''!
response
	^ response! !

!ZnClientTransactionEvent methodsFor: 'accessing' stamp: ''!
response: anObject
	response := anObject! !

!ZnClientTransactionEvent methodsFor: 'accessing' stamp: ''!
responseDuration
	^ responseDuration! !

!ZnClientTransactionEvent methodsFor: 'accessing' stamp: ''!
responseDuration: anObject
	responseDuration := anObject! !

ZnClientLogEvent subclass: #ZnConnectionEstablishedEvent
	instanceVariableNames: 'hostname port address duration proxy'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnConnectionEstablishedEvent commentStamp: '' prior: 0!
ZnConnectionEstablishedEvent signals that a client HTTP connection to hostname:port (address:port) was established in duration milliseconds. This time includes DNS resolving.!

!ZnConnectionEstablishedEvent methodsFor: 'accessing' stamp: ''!
address
	^ address! !

!ZnConnectionEstablishedEvent methodsFor: 'accessing' stamp: ''!
address: anObject
	address := anObject! !

!ZnConnectionEstablishedEvent methodsFor: 'accessing' stamp: ''!
duration
	^ duration! !

!ZnConnectionEstablishedEvent methodsFor: 'accessing' stamp: ''!
duration: anObject
	duration := anObject! !

!ZnConnectionEstablishedEvent methodsFor: 'accessing' stamp: ''!
hostname
	^ hostname! !

!ZnConnectionEstablishedEvent methodsFor: 'accessing' stamp: ''!
hostname: anObject
	hostname := anObject! !

!ZnConnectionEstablishedEvent methodsFor: 'accessing' stamp: ''!
port
	^ port! !

!ZnConnectionEstablishedEvent methodsFor: 'accessing' stamp: ''!
port: anObject
	port := anObject! !

!ZnConnectionEstablishedEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Connection Established '; nextPutAll: hostname; nextPut: $:; print: port; space. 
	proxy ifNotNil: [ stream << 'via proxy '; print: proxy; space ].
	address do: [ :each | stream print: each ] separatedBy: [ stream nextPut: $. ].
	stream space; print: duration; << 'ms '.! !

!ZnConnectionEstablishedEvent methodsFor: 'accessing' stamp: ''!
proxy
	^ proxy! !

!ZnConnectionEstablishedEvent methodsFor: 'accessing' stamp: ''!
proxy: anObject
	proxy := anObject! !

ZnClientLogEvent subclass: #ZnRequestWrittenEvent
	instanceVariableNames: 'request duration'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnRequestWrittenEvent commentStamp: '' prior: 0!
ZnRequestWrittenEvent signals that a request was written in duration milliseconds.!

!ZnRequestWrittenEvent methodsFor: 'accessing' stamp: ''!
duration
	^ duration! !

!ZnRequestWrittenEvent methodsFor: 'accessing' stamp: ''!
duration: anObject
	duration := anObject! !

!ZnRequestWrittenEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Request Written '; print: request; space; print: duration; << 'ms'! !

!ZnRequestWrittenEvent methodsFor: 'accessing' stamp: ''!
request
	^ request! !

!ZnRequestWrittenEvent methodsFor: 'accessing' stamp: ''!
request: anObject
	request := anObject! !

ZnClientLogEvent subclass: #ZnResponseReadEvent
	instanceVariableNames: 'response duration'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnResponseReadEvent commentStamp: '' prior: 0!
ZnResponseReadEvent signals that a response was read in duration milliseconds. This time includes all waiting, server processing and networking.!

!ZnResponseReadEvent methodsFor: 'accessing' stamp: ''!
duration
	^ duration! !

!ZnResponseReadEvent methodsFor: 'accessing' stamp: ''!
duration: anObject
	duration := anObject! !

!ZnResponseReadEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Response Read '; print: response; space; print: duration; << 'ms'! !

!ZnResponseReadEvent methodsFor: 'accessing' stamp: ''!
response
	^ response! !

!ZnResponseReadEvent methodsFor: 'accessing' stamp: ''!
response: anObject
	response := anObject! !

ZnClientLogEvent subclass: #ZnSimplifiedClientTransactionEvent
	instanceVariableNames: 'method url response size duration'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnSimplifiedClientTransactionEvent commentStamp: '' prior: 0!
ZnSimplifiedClientTransactionEvent is a ZnClientLogEvent that notes the completion of a client side HTTP transaction (request/response). Only a limited number of simpler data is held. ZnClientTransactionEvent contains much more detailed data.

Instance Variables
	method: HTTP verb
	url: ZnUrl requested
	response:	numeric response code
	size: number of bytes in response
	duration: milliseconds

ZnSimplifiedClientTransactionEvent is ideal for classic HTTP logging with little overhead through its default string representation.!

!ZnSimplifiedClientTransactionEvent methodsFor: 'accessing' stamp: ''!
duration
	^ duration! !

!ZnSimplifiedClientTransactionEvent methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	duration := 0! !

!ZnSimplifiedClientTransactionEvent methodsFor: 'accessing' stamp: ''!
method
	^ method! !

!ZnSimplifiedClientTransactionEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream space; nextPutAll: method.
	stream space.
	url printPathQueryFragmentOn: stream.
	stream space; print: response.
	size ifNotNil: [ 
		stream space; print: size; nextPut: $B ].
	stream space; print: duration; nextPutAll: 'ms'! !

!ZnSimplifiedClientTransactionEvent methodsFor: 'initialize' stamp: ''!
request: aZnRequest
	method := aZnRequest method.
	url := aZnRequest url! !

!ZnSimplifiedClientTransactionEvent methodsFor: 'initialize' stamp: ''!
requestDuration: requestDuration
	duration := duration + requestDuration ! !

!ZnSimplifiedClientTransactionEvent methodsFor: 'initialize' stamp: ''!
response: aZnResponse
	response := aZnResponse code.
	aZnResponse hasEntity
		ifTrue: [ size := aZnResponse entity contentLength ]! !

!ZnSimplifiedClientTransactionEvent methodsFor: 'accessing' stamp: ''!
responseCode
	^ response! !

!ZnSimplifiedClientTransactionEvent methodsFor: 'initialize' stamp: ''!
responseDuration: responseDuration
	duration := duration + responseDuration ! !

!ZnSimplifiedClientTransactionEvent methodsFor: 'accessing' stamp: ''!
size
	^ size! !

!ZnSimplifiedClientTransactionEvent methodsFor: 'accessing' stamp: ''!
url
	^ url! !

!ZnLogEvent class methodsFor: 'accessing' stamp: ''!
announcer
	^ LogEventAnnouncer ifNil: [ LogEventAnnouncer := Announcer new ]! !

!ZnLogEvent class methodsFor: 'class initialization' stamp: ''!
initialize
	IdCounter := 0.
	self environment 
		at: #SessionManager 
		ifPresent: [ :manager | manager default registerNetworkClassNamed: self name ]
		ifAbsent: [ Smalltalk addToStartUpList: self ]! !

ZnLogEvent initialize!

!ZnLogEvent class methodsFor: 'convenience' stamp: ''!
logToTranscript
	self stopLoggingToTranscript.
	^ self announcer when: ZnLogEvent do: [ :event | Transcript crShow: event ].! !

!ZnLogEvent class methodsFor: 'accessing' stamp: ''!
nextId
	"This should be thread safe because SmallInteger 
	arithmetic primitives cannot be interrupted"
	
	^ IdCounter := IdCounter + 1! !

!ZnLogEvent class methodsFor: 'convenience' stamp: ''!
open
	^ self announcer inspect! !

!ZnLogEvent class methodsFor: 'system startup' stamp: ''!
startUp
	"Reset the id counter"
	
	IdCounter := 0! !

!ZnLogEvent class methodsFor: 'convenience' stamp: ''!
stopLoggingToTranscript
	self announcer unsubscribe: self! !

!ZnLogEvent methodsFor: 'accessing' stamp: ''!
announcer
	^ self class announcer! !

!ZnLogEvent methodsFor: 'actions' stamp: ''!
emit
	self announcer announce: self! !

!ZnLogEvent methodsFor: 'accessing' stamp: ''!
id
	^ id! !

!ZnLogEvent methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	timestamp := DateAndTime now.
	id := self nextId! !

!ZnLogEvent methodsFor: 'accessing' stamp: ''!
nextId
	^ self class nextId! !

!ZnLogEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	"Subclasses should implement this to add output"! !

!ZnLogEvent methodsFor: 'printing' stamp: ''!
printOn: stream
	timestamp printYMDOn: stream.
	stream space.
	timestamp printHMSOn: stream. 
	stream space.
	id \\ 1000 printOn: stream base: 10 length: 3 padded: true.
	self printContentsOn: stream! !

!ZnLogEvent methodsFor: 'beacon' stamp: ''!
printOneLineContentsOn: stream
	self printContentsOn: stream! !

!ZnLogEvent methodsFor: 'beacon' stamp: ''!
printOneLineOn: stream
	self printOn: stream! !

!ZnLogEvent methodsFor: 'accessing' stamp: ''!
timestamp
	^ timestamp! !

ZnLogEvent subclass: #ZnServerLogEvent
	instanceVariableNames: 'serverId processId'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnServerLogEvent commentStamp: '' prior: 0!
I am ZnServerLogEvent, a ZnLogEvent generated by ZnServer and its subclasses.

I hold a processId attribute.
I hold an optional serverId attribute.!

ZnServerLogEvent subclass: #ZnConnectionAcceptedEvent
	instanceVariableNames: 'address'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnConnectionAcceptedEvent commentStamp: '' prior: 0!
ZnConnectionAcceptedEvent is signalled when an HTTP server accepts an incoming connection from a client at address.!

!ZnConnectionAcceptedEvent methodsFor: 'accessing' stamp: ''!
address
	^ address! !

!ZnConnectionAcceptedEvent methodsFor: 'accessing' stamp: ''!
address: anObject
	address := anObject! !

!ZnConnectionAcceptedEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Connection Accepted '.
	address do: [ :each | stream print: each ] separatedBy: [ stream nextPut: $. ]! !

ZnServerLogEvent subclass: #ZnRequestReadEvent
	instanceVariableNames: 'request duration'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnRequestReadEvent commentStamp: '' prior: 0!
ZnRequestReadEvent is signalled when an HTTP server reads a request in duration milliseconds.

Note that the duration can be misleading: when servicing multiple requests over a kept alive connection (the default for HTTP 1.1), the read time includes any wait time (as long as it is below the timeout). The first request read should not contain wait time.!

!ZnRequestReadEvent methodsFor: 'accessing' stamp: ''!
duration
	^ duration! !

!ZnRequestReadEvent methodsFor: 'accessing' stamp: ''!
duration: anObject
	duration := anObject! !

!ZnRequestReadEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Request Read '; print: request; space; print: duration; << 'ms'! !

!ZnRequestReadEvent methodsFor: 'accessing' stamp: ''!
request
	^ request! !

!ZnRequestReadEvent methodsFor: 'accessing' stamp: ''!
request: anObject
	request := anObject! !

ZnServerLogEvent subclass: #ZnRequestResponseHandledEvent
	instanceVariableNames: 'request response duration'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnRequestResponseHandledEvent commentStamp: '' prior: 0!
ZnRequestResponseHandledEvent signals that the handler (delegate) processed a request and produced a response in duration in milliseconds. This duration is thus the time taken by the application itself.!

!ZnRequestResponseHandledEvent methodsFor: 'accessing' stamp: ''!
duration
	^ duration! !

!ZnRequestResponseHandledEvent methodsFor: 'accessing' stamp: ''!
duration: anObject
	duration := anObject! !

!ZnRequestResponseHandledEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Request Handled '; print: request; space; print: duration; << 'ms'! !

!ZnRequestResponseHandledEvent methodsFor: 'accessing' stamp: ''!
request
	^ request! !

!ZnRequestResponseHandledEvent methodsFor: 'accessing' stamp: ''!
request: anObject
	request := anObject! !

!ZnRequestResponseHandledEvent methodsFor: 'accessing' stamp: ''!
response
	^ response! !

!ZnRequestResponseHandledEvent methodsFor: 'accessing' stamp: ''!
response: anObject
	response := anObject! !

ZnServerLogEvent subclass: #ZnResponseWrittenEvent
	instanceVariableNames: 'response duration'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnResponseWrittenEvent commentStamp: '' prior: 0!
ZnResponseWrittenEvent signals that response was written in duration milliseconds.!

!ZnResponseWrittenEvent methodsFor: 'accessing' stamp: ''!
duration
	^ duration! !

!ZnResponseWrittenEvent methodsFor: 'accessing' stamp: ''!
duration: anObject
	duration := anObject! !

!ZnResponseWrittenEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Response Written '; print: response; space; print: duration; << 'ms'! !

!ZnResponseWrittenEvent methodsFor: 'accessing' stamp: ''!
response
	^ response! !

!ZnResponseWrittenEvent methodsFor: 'accessing' stamp: ''!
response: anObject
	response := anObject! !

ZnServerLogEvent subclass: #ZnServerConnectionClosedEvent
	instanceVariableNames: 'address'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnServerConnectionClosedEvent commentStamp: '' prior: 0!
ZnServerConnectionClosedEvent signals that the server closed a client connection to address. !

!ZnServerConnectionClosedEvent methodsFor: 'accessing' stamp: ''!
address
	^ address! !

!ZnServerConnectionClosedEvent methodsFor: 'accessing' stamp: ''!
address: anObject
	address := anObject! !

!ZnServerConnectionClosedEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Server Connection Closed '.
	address do: [ :each | stream print: each ] separatedBy: [ stream nextPut: $. ]! !

ZnServerLogEvent subclass: #ZnServerErrorEvent
	instanceVariableNames: 'exception'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnServerErrorEvent commentStamp: '' prior: 0!
ZnServerErrorEvent is a ZnServerLogEvent related to an exception that occurred.!

!ZnServerErrorEvent methodsFor: 'accessing' stamp: ''!
exception
	^ exception! !

!ZnServerErrorEvent methodsFor: 'accessing' stamp: ''!
exception: anObject
	exception := anObject! !

ZnServerErrorEvent subclass: #ZnServerHandlerErrorEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnServerHandlerErrorEvent commentStamp: '' prior: 0!
ZnServerHandlerErrorEvent signals an error that occurred during handling by the delegate. This is considered a real error.!

!ZnServerHandlerErrorEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Server Handler Error ';  print: exception! !

ZnServerErrorEvent subclass: #ZnServerReadErrorEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnServerReadErrorEvent commentStamp: '' prior: 0!
ZnServerReadErrorEvent signals an error during the reading of a request. These errors are considered mostly benign and the connection gets closed.
!

!ZnServerReadErrorEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Server Read Error ';  print: exception! !

ZnServerErrorEvent subclass: #ZnServerWriteErrorEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnServerWriteErrorEvent commentStamp: '' prior: 0!
ZnServerWriteErrorEvent signals an error during the writing of a response. These errors are considered mostly benign and the connection gets closed.
!

!ZnServerWriteErrorEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Server Write Error ';  print: exception! !

ZnServerLogEvent subclass: #ZnServerGenericLogEvent
	instanceVariableNames: 'subject'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnServerGenericLogEvent commentStamp: '' prior: 0!
I am ZnServerGenericLogEvent, a ZnServerLogEvent and ZnLogEvent that holds a generic subject, possibly not limited to a String.

Here are two examples:

(ZnServerGenericLogEvent subject: 'You can''t do that, Dave') emit.

(ZnServerGenericLogEvent subject: { #id->123. #reason->#timeout } asDictionary) emit.!

!ZnServerGenericLogEvent class methodsFor: 'instance creation' stamp: ''!
subject: object
	"self subject: #(i 10 j -1)."
	"self subject: 'How are you?'."
	
	^ self new
		subject: object;
		yourself! !

!ZnServerGenericLogEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream space; nextPutAll: subject asString! !

!ZnServerGenericLogEvent methodsFor: 'accessing' stamp: ''!
subject
	^ subject! !

!ZnServerGenericLogEvent methodsFor: 'accessing' stamp: ''!
subject: anObject
	subject := anObject! !

!ZnServerLogEvent methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	processId := ZnUtils currentProcessID! !

!ZnServerLogEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	serverId ifNotNil: [ stream space; << serverId ].
	stream space.
	processId printOn: stream base: 10 length: 6 padded: true! !

!ZnServerLogEvent methodsFor: 'accessing' stamp: ''!
processId
	^ processId! !

!ZnServerLogEvent methodsFor: 'initialize' stamp: ''!
processId: anObject
	processId := anObject! !

!ZnServerLogEvent methodsFor: 'accessing' stamp: ''!
serverId
	^ serverId! !

!ZnServerLogEvent methodsFor: 'initialize' stamp: ''!
serverId: anObject
	serverId := anObject! !

ZnServerLogEvent subclass: #ZnServerSocketBoundEvent
	instanceVariableNames: 'address port'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnServerSocketBoundEvent commentStamp: '' prior: 0!
ZnServerSocketBoundEvent signals a new server socket on address:port is bound (opened).!

!ZnServerSocketBoundEvent methodsFor: 'accessing' stamp: ''!
address
	^ address! !

!ZnServerSocketBoundEvent methodsFor: 'accessing' stamp: ''!
address: anObject
	address := anObject! !

!ZnServerSocketBoundEvent methodsFor: 'accessing' stamp: ''!
port
	^ port! !

!ZnServerSocketBoundEvent methodsFor: 'accessing' stamp: ''!
port: anObject
	port := anObject! !

!ZnServerSocketBoundEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Server Socket Bound '.
	address do: [ :each | stream print: each ] separatedBy: [ stream nextPut: $. ].
	stream nextPut: $:; print: port
	! !

ZnServerLogEvent subclass: #ZnServerSocketReleasedEvent
	instanceVariableNames: 'address port'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnServerSocketReleasedEvent commentStamp: '' prior: 0!
ZnServerSocketReleasedEvent signals a new server socket on address:port is released (closed).!

!ZnServerSocketReleasedEvent methodsFor: 'accessing' stamp: ''!
address
	^ address! !

!ZnServerSocketReleasedEvent methodsFor: 'accessing' stamp: ''!
address: anObject
	address := anObject! !

!ZnServerSocketReleasedEvent methodsFor: 'accessing' stamp: ''!
port
	^ port! !

!ZnServerSocketReleasedEvent methodsFor: 'accessing' stamp: ''!
port: anObject
	port := anObject! !

!ZnServerSocketReleasedEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Server Socket Released '.
	address do: [ :each | stream print: each ] separatedBy: [ stream nextPut: $. ].
	stream nextPut: $:; print: port! !

ZnServerLogEvent subclass: #ZnServerStartedEvent
	instanceVariableNames: 'description'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnServerStartedEvent commentStamp: '' prior: 0!
ZnServerStartedEvent notes the startup of an HTTP server is complete.
!

!ZnServerStartedEvent methodsFor: 'accessing' stamp: ''!
description
	^ description! !

!ZnServerStartedEvent methodsFor: 'accessing' stamp: ''!
description: anObject
	description := anObject! !

!ZnServerStartedEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Started '; << description! !

ZnServerLogEvent subclass: #ZnServerStoppedEvent
	instanceVariableNames: 'description'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnServerStoppedEvent commentStamp: '' prior: 0!
ZnServerStoppedEvent notes the shutdown of an HTTP server is complete.!

!ZnServerStoppedEvent methodsFor: 'accessing' stamp: ''!
description
	^ description! !

!ZnServerStoppedEvent methodsFor: 'accessing' stamp: ''!
description: anObject
	description := anObject! !

!ZnServerStoppedEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream << ' Stopped '; << description! !

ZnServerLogEvent subclass: #ZnServerTransactionEvent
	instanceVariableNames: 'request response timing'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnServerTransactionEvent commentStamp: '' prior: 0!
ZnServerTransactionEvent is a ZnServerLogEvent that notes the completion of a server side HTTP transaction (request/response).

Instance Variables
	request: a ZnRequest
	response:	a ZnResponse
	timing: a ZnServerTransactionTiming

Note that this can be quite large since it includes the entities transferred. See ZnSimplifiedServerTransactionEvent for a more compact representation.!

!ZnServerTransactionEvent methodsFor: 'accessing' stamp: ''!
duration
	^ timing totalDuration! !

!ZnServerTransactionEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream space.
	request requestLine printMethodAndUriOn: stream.
	stream space; print: response code.
	response hasEntity
		ifTrue: [ 
			stream space; print: response entity contentLength; nextPut: $B ].
	stream space; print: self duration; nextPutAll: 'ms'! !

!ZnServerTransactionEvent methodsFor: 'accessing' stamp: ''!
request
	^ request! !

!ZnServerTransactionEvent methodsFor: 'accessing' stamp: ''!
request: anObject
	request := anObject! !

!ZnServerTransactionEvent methodsFor: 'accessing' stamp: ''!
response
	^ response! !

!ZnServerTransactionEvent methodsFor: 'accessing' stamp: ''!
response: anObject
	response := anObject! !

!ZnServerTransactionEvent methodsFor: 'accessing' stamp: ''!
timing
	^ timing! !

!ZnServerTransactionEvent methodsFor: 'accessing' stamp: ''!
timing: anObject
	timing := anObject! !

ZnServerLogEvent subclass: #ZnSimplifiedServerTransactionEvent
	instanceVariableNames: 'method url response size duration'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnSimplifiedServerTransactionEvent commentStamp: '' prior: 0!
ZnSimplifiedServerTransactionEvent is a ZnServerLogEvent that notes the completion of a served side HTTP transaction (request/response). Only a limited number of simpler data is held. ZnServerTransactionEvent contains much more detailed data.

Instance Variables
	method: HTTP verb
	url: ZnUrl requested
	response:	numeric response code
	size: number of bytes in response
	duration: milliseconds

ZnSimplifiedServerTransactionEvent is ideal for classic HTTP logging with little overhead through its default string representation.
!

!ZnSimplifiedServerTransactionEvent methodsFor: 'accessing' stamp: ''!
duration
	^ duration! !

!ZnSimplifiedServerTransactionEvent methodsFor: 'accessing' stamp: ''!
method
	^ method! !

!ZnSimplifiedServerTransactionEvent methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	stream space; nextPutAll: method.
	stream space.
	url printPathQueryFragmentOn: stream.
	stream space; print: response.
	size ifNotNil: [ 
		stream space; print: size; nextPut: $B ].
	stream space; print: duration; nextPutAll: 'ms'! !

!ZnSimplifiedServerTransactionEvent methodsFor: 'initialize' stamp: ''!
request: aZnRequest
	method := aZnRequest method.
	url := aZnRequest url! !

!ZnSimplifiedServerTransactionEvent methodsFor: 'initialize' stamp: ''!
response: aZnResponse
	response := aZnResponse code.
	aZnResponse hasEntity
		ifTrue: [ size := aZnResponse entity contentLength ]! !

!ZnSimplifiedServerTransactionEvent methodsFor: 'accessing' stamp: ''!
responseCode
	^ response! !

!ZnSimplifiedServerTransactionEvent methodsFor: 'accessing' stamp: ''!
size
	^ size! !

!ZnSimplifiedServerTransactionEvent methodsFor: 'initialize' stamp: ''!
timing: timing
	duration := timing totalDuration ! !

!ZnSimplifiedServerTransactionEvent methodsFor: 'accessing' stamp: ''!
url
	^ url! !

Notification subclass: #ZnRespond
	instanceVariableNames: 'response'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Exceptions'!

!ZnRespond commentStamp: '' prior: 0!
I am ZnRespond, a Notification to signal the end of #handleRequest: processing with a specific ZnResponse, earlier than normal stack unwinding.

  ZnRespond signalWith: ZnResponse unauthorized

Normal #handleRequest: processing in a ZnServer delegate takes a ZnRequest object as input argument and needs to produce a ZnResponse as output, returning it as a regular result. Sometimes you do not want to keep managing this single result with many levels of condition on a complex or deep stack, but instead directly want to return a response: that is when you can use the ZnRespond notification.!

!ZnRespond class methodsFor: 'instance creation' stamp: ''!
signalWith: response
	^ self new
		response: response;
		signal! !

!ZnRespond methodsFor: 'accessing' stamp: ''!
response
	^ response! !

!ZnRespond methodsFor: 'accessing' stamp: ''!
response: anObject
	response := anObject! !

Error subclass: #ZnEntityTooLarge
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Exceptions'!

!ZnEntityTooLarge commentStamp: '' prior: 0!
ZnEntityTooLarge is signalled when too big an entity is read from a stream.

This is a resumable exception!

!ZnEntityTooLarge methodsFor: 'testing' stamp: ''!
isResumable
	"Determine whether an exception is resumable."

	^ true! !

Error subclass: #ZnHttpUnsuccessful
	instanceVariableNames: 'response'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Exceptions'!

!ZnHttpUnsuccessful commentStamp: '' prior: 0!
ZnHttpUnsuccessful is signalled when an HTTP client receives a non success response.

See ZnResponse>>#isSuccess

I hold the response that was unsuccessful.

Part of Zinc HTTP Components. !

!ZnHttpUnsuccessful class methodsFor: 'instance creation' stamp: ''!
response: response
	^ self new
		response: response;
		yourself! !

!ZnHttpUnsuccessful methodsFor: 'accessing' stamp: ''!
messageText
	"Overwritten to initialiaze the message text to a standard text if it has not yet been set"
	
	^ messageText ifNil: [ messageText := self standardMessageText ]! !

!ZnHttpUnsuccessful methodsFor: 'accessing' stamp: ''!
response
	^ response! !

!ZnHttpUnsuccessful methodsFor: 'accessing' stamp: ''!
response: anObject
	response := anObject! !

!ZnHttpUnsuccessful methodsFor: 'accessing' stamp: ''!
standardMessageText
	"Generate a standard textual description"
	
	^ String streamContents: [ :stream |
		response ifNotNil: [
			response statusLine printCodeAndReasonOn: stream ] ]! !

Error subclass: #ZnMissingHost
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Exceptions'!

!ZnMissingHost commentStamp: '' prior: 0!
ZnMissingHost is signalled when a host should be specified but was not.

Part of Zinc HTTP Components. !

Error subclass: #ZnParseError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Exceptions'!

!ZnParseError commentStamp: '' prior: 0!
ZnParseError is the superclass of all Zinc HTTP parsing related errors.

Part of Zinc HTTP Components.!

ZnParseError subclass: #ZnLineTooLong
	instanceVariableNames: 'limit'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Exceptions'!

!ZnLineTooLong commentStamp: '' prior: 0!
ZnLineTooLong is signalled when an HTTP CRLF terminated input line exceeds a predefined limit.

Part of Zinc HTTP Components. !

!ZnLineTooLong class methodsFor: 'instance creation' stamp: ''!
limit: limit
	^ self new 
		limit: limit; 
		yourself! !

!ZnLineTooLong methodsFor: 'accessing' stamp: ''!
limit
	^ limit! !

!ZnLineTooLong methodsFor: 'accessing' stamp: ''!
limit: anObject
	limit := anObject! !

ZnParseError subclass: #ZnUnknownHttpMethod
	instanceVariableNames: 'method'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Exceptions'!

!ZnUnknownHttpMethod commentStamp: '' prior: 0!
ZnUnknownHttpMethod is signalled when an unknown/unsupported HTTP method is found while parsing a request.

I am resumable to allow non-standard methods.

Part of Zinc HTTP Components. !

!ZnUnknownHttpMethod class methodsFor: 'instance creation' stamp: ''!
method: method
	^ self new
		method: method;
		yourself! !

!ZnUnknownHttpMethod methodsFor: 'testing' stamp: ''!
isResumable
	^ true! !

!ZnUnknownHttpMethod methodsFor: 'accessing' stamp: ''!
method
	^ method! !

!ZnUnknownHttpMethod methodsFor: 'accessing' stamp: ''!
method: anObject
	method := anObject! !

ZnParseError subclass: #ZnUnknownHttpStatusCode
	instanceVariableNames: 'code'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Exceptions'!

!ZnUnknownHttpStatusCode commentStamp: '' prior: 0!
ZnUnknownHttpStatusCode is signalled when an unknown/unsupported HTTP status code is found.

Part of Zinc HTTP Components. !

!ZnUnknownHttpStatusCode class methodsFor: 'instance creation' stamp: ''!
code: code
	^ self new
		code: code;
		yourself! !

!ZnUnknownHttpStatusCode methodsFor: 'accessing' stamp: ''!
code
	^ code! !

!ZnUnknownHttpStatusCode methodsFor: 'accessing' stamp: ''!
code: anObject
	code := anObject! !

ZnParseError subclass: #ZnUnknownHttpVersion
	instanceVariableNames: 'version'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Exceptions'!

!ZnUnknownHttpVersion commentStamp: '' prior: 0!
ZnUnknownHttpVersion is signalled when an unknown/unsupported HTTP version is found.

Part of Zinc HTTP Components. !

!ZnUnknownHttpVersion class methodsFor: 'instance creation' stamp: ''!
version: version
	^ self new
		version: version;
		yourself! !

!ZnUnknownHttpVersion methodsFor: 'accessing' stamp: ''!
version
	^ version! !

!ZnUnknownHttpVersion methodsFor: 'accessing' stamp: ''!
version: anObject
	version := anObject! !

Error subclass: #ZnTooManyRedirects
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Exceptions'!

!ZnTooManyRedirects commentStamp: '' prior: 0!
ZnTooManyRedirects is signalled when an HTTP client has been following more redirects than allowed.

Part of Zinc HTTP Components. !

!ZnTooManyRedirects methodsFor: 'testing' stamp: ''!
isResumable
	"Determine whether an exception is resumable."

	^ true! !

Error subclass: #ZnUnexpectedContentType
	instanceVariableNames: 'expectedContentType actualContentType'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Exceptions'!

!ZnUnexpectedContentType commentStamp: '' prior: 0!
ZnUnexpectedContentType is signalled when a different content type (mime type) was expected.

I hold the expected and actual content types.

Part of Zinc HTTP Components. !

!ZnUnexpectedContentType class methodsFor: 'instance creation' stamp: ''!
expected: expectedMimeType actual: actualMimeType
	^ self new
		expectedContentType: expectedMimeType;
		actualContentType: actualMimeType;
		yourself! !

!ZnUnexpectedContentType methodsFor: 'accessing' stamp: ''!
actualContentType
	^ actualContentType! !

!ZnUnexpectedContentType methodsFor: 'accessing' stamp: ''!
actualContentType: anObject
	actualContentType := anObject! !

!ZnUnexpectedContentType methodsFor: 'accessing' stamp: ''!
expectedContentType
	^ expectedContentType! !

!ZnUnexpectedContentType methodsFor: 'accessing' stamp: ''!
expectedContentType: anObject
	expectedContentType := anObject! !

!ZnUnexpectedContentType methodsFor: 'accessing' stamp: ''!
messageText
	"Overwritten to initialiaze the message text to a standard text if it has not yet been set"
	
	^ messageText ifNil: [ messageText := self standardMessageText ]! !

!ZnUnexpectedContentType methodsFor: 'accessing' stamp: ''!
standardMessageText
	"Generate a standard textual description"
	
	^ String streamContents: [ :stream |
		self expectedContentType ifNotNil: [ :contentType |
			stream << 'expected '; print: contentType ].
		self actualContentType ifNotNil: [ :contentType |
			self expectedContentType ifNotNil: [ stream space ].
			stream << 'actual '; print: contentType ] ]! !

!ZnUrl methodsFor: '*zinc-http' stamp: ''!
retrieveContents
	"Download and return the resource that I refer to.
	This will typically return a String or a ByteArray.
	If something goes wrong, an exception will be signalled."
	
	" 'http://zn.stfx.eu/zn/numbers.txt' asZnUrl retrieveContents. "
	
	^ self performOperation: #retrieveContents! !

!ZnUrl methodsFor: '*zinc-http' stamp: ''!
saveContentsToFile: aFilename
	"Download and return a fileReference to the resource that I refer to.

	If something goes wrong, an exception will be signalled."
	
	" 'http://zn.stfx.eu/zn/numbers.txt' asZnUrl saveContentsToFile: Path * 'numbers.txt'. "
	
	^ self performOperation: #saveContentsToFile: with: aFilename! !

Stream subclass: #ZnBivalentWriteStream
	instanceVariableNames: 'stream binary'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Streaming'!

!ZnBivalentWriteStream commentStamp: '' prior: 0!
I am ZnBivalentWriteStream, wrapping another write stream.

I allow both Characters and Strings as well as Integers (0 to: 255) and ByteArrays to be writtten to me.

The write stream that I wrap can be both binary as well as character based.

Part of Zinc HTTP Components.!

!ZnBivalentWriteStream class methodsFor: 'instance creation' stamp: ''!
on: writeStream
	"Wrap writeStream as a ZnBivalentWriteStream, unless it is already one"
	
	(writeStream isKindOf: self) ifTrue: [ ^ writeStream ].
	
	^ self basicNew
		on: writeStream;
		yourself! !

!ZnBivalentWriteStream methodsFor: 'initialize-release' stamp: ''!
close
	stream close! !

!ZnBivalentWriteStream methodsFor: 'accessing' stamp: ''!
flush
	stream flush! !

!ZnBivalentWriteStream methodsFor: 'private' stamp: ''!
isBinary
	binary isNil
		ifTrue: [ [ binary := stream isBinary ] on: MessageNotUnderstood do: [ :err | binary := false ] ].
	^ binary! !

!ZnBivalentWriteStream methodsFor: 'accessing' stamp: ''!
next: count putAll: collection startingAt: offset
	| buffer |
	self isBinary
		ifTrue: [ 
			(collection isKindOf: ByteArray)
				ifTrue: [ stream next: count putAll: collection startingAt: offset ]
				ifFalse: [ 
					buffer := ByteArray new: count.
					buffer replaceFrom: 1 to: count with: collection startingAt: offset.
					stream nextPutAll: buffer ] ]
		ifFalse: [ 
			(collection isKindOf: ByteArray)
				ifTrue: [ 
					buffer := String new: count.
					buffer replaceFrom: 1 to: count with: collection startingAt: offset.
					stream nextPutAll: buffer ]
				ifFalse: [ stream next: count putAll: collection startingAt: offset ] ]! !

!ZnBivalentWriteStream methodsFor: 'accessing' stamp: ''!
nextPut: object
	self isBinary
		ifTrue: [ stream nextPut: object asInteger ]
		ifFalse: [ stream nextPut: object asCharacter ]! !

!ZnBivalentWriteStream methodsFor: 'accessing' stamp: ''!
nextPutAll: collection
	self isBinary
		ifTrue: [ stream nextPutAll: collection asByteArray ]
		ifFalse: [ stream nextPutAll: collection asString ]! !

!ZnBivalentWriteStream methodsFor: 'initialize-release' stamp: ''!
on: writeStream
	stream := writeStream! !

!ZnBivalentWriteStream methodsFor: 'printing' stamp: ''!
printOn: aStream
	aStream 
		nextPutAll: 'a '; 
		nextPutAll: self class name! !

!ZnBivalentWriteStream methodsFor: 'accessing' stamp: ''!
space
	self nextPut: Character space! !

Stream subclass: #ZnChunkedReadStream
	instanceVariableNames: 'stream chunk position limit atEnd totalSize chunkCount extraHeaders'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Streaming'!

!ZnChunkedReadStream commentStamp: '' prior: 0!
I am ZnChunkedReadStream, implementing HTTP 1.1 chunked transfer encoding on a wrapped streams.

Clients should read me until I am atEnd.

After I am completely read, I can tell you my totalSize and optional extraHeaders.

Part of Zinc HTTP Components.!

!ZnChunkedReadStream class methodsFor: 'instance creation' stamp: ''!
on: stream
	^ self basicNew
		initialize;
		on: stream;
		yourself! !

!ZnChunkedReadStream methodsFor: 'testing' stamp: ''!
atEnd
	self ensureChunkOrAtEnd.
	^ atEnd! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
binary
	stream binary! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
chunkCount
	^ chunkCount! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
collectionSpecies
	^ self isBinary ifTrue: [ ByteArray ] ifFalse: [ String ]
	! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
contents
	"This is technically not correct, but it is better than nothing"
	
	^ self upToEnd! !

!ZnChunkedReadStream methodsFor: 'private' stamp: ''!
ensureChunkBufferOfSize: size
	(chunk notNil and: [ size <= chunk size ]) ifTrue: [ ^ self ].  
	chunk := self collectionSpecies new: size.
! !

!ZnChunkedReadStream methodsFor: 'private' stamp: ''!
ensureChunkOrAtEnd
	atEnd 
		ifTrue: [ ^ self ].
	(chunk isNil or: [ position >= limit ]) 
		ifTrue: [ self getNextChunk ]
! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
extraHeaders
	^ extraHeaders! !

!ZnChunkedReadStream methodsFor: 'private' stamp: ''!
getExtraHeaders
	extraHeaders := ZnHeaders readFrom: stream! !

!ZnChunkedReadStream methodsFor: 'private' stamp: ''!
getNextChunk
	| size |
	(size := self getNextChunkSize) = 0
		ifTrue: [
			atEnd := true.
			chunk := position := limit := nil.
			self getExtraHeaders ]
		ifFalse: [
			totalSize := totalSize + size.
			chunkCount := chunkCount + 1.
			self ensureChunkBufferOfSize: size.  
			stream next: size into: chunk.
			stream next: 2. "consume CRLF"
			limit := size.
			position := 0 ]! !

!ZnChunkedReadStream methodsFor: 'private' stamp: ''!
getNextChunkSize
	| line |
	(line := (ZnLineReader on: stream) nextLine) isEmpty ifTrue: [ ^ 0 ].
	^ Integer readFrom: (line copyUpTo: $;) base: 16! !

!ZnChunkedReadStream methodsFor: 'initialization' stamp: ''!
initialize
	atEnd := false.
	chunkCount := 0.
	totalSize := 0! !

!ZnChunkedReadStream methodsFor: 'testing' stamp: ''!
isBinary
	^ stream isBinary
! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
match: subCollection
	"Try to read and match the elements of subCollection. 
	If successful return true and leave the me positioned after the match.
	If unsuccesful return false and leave me at end."
	
	| buffer pattern bufferPosition bufferLimit |
	pattern := subCollection readStream.
	"we have to use an internal buffer because we are not positionable"
	buffer := subCollection species new: subCollection size.
	bufferPosition := bufferLimit := 0.
	[ pattern atEnd ] whileFalse: [ | nextElement |
		self atEnd ifTrue: [ ^ false ].
		"get the next char to match either from the buffer or from the stream while buffering it"
		(bufferPosition < bufferLimit) 
			ifTrue: [ 
				nextElement := buffer at: bufferPosition + 1 ]
			ifFalse: [
				nextElement := buffer at: bufferPosition + 1 put: self next.
				bufferLimit := bufferLimit + 1 ].
		bufferPosition := bufferPosition + 1.
		pattern next = nextElement
			ifFalse: [
				"shift the buffer down one element and restart"
				2 to: bufferLimit do: [ :each |
					buffer at: each - 1 put: (buffer at: each) ].
				bufferLimit := bufferLimit - 1.
				bufferPosition := 0.
				pattern position: 0 ] ].
	^ true! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
next
	self ensureChunkOrAtEnd.
	self atEnd ifTrue: [ ^ nil ].
	^ chunk at: (position := position + 1)! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
next: requestedCount 
	"Read requestedCount elements into new collection and return it,
	 it could be that less elements were available"

	^ self 
		next: requestedCount 
		into: (self collectionSpecies new: requestedCount)! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
next: requestedCount into: collection
	"Read requestedCount elements into collection,
	returning a copy if less elements are available"
	
	^ self
		next: requestedCount
		into: collection
		startingAt: 1! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
next: requestedCount into: collection startingAt: offset
	"Read requestedCount elements into collection starting at offset,
	returning a copy if less elements are available"
	
	| readCount |
	readCount := self 
		readInto: collection 
		startingAt: offset 
		count: requestedCount.
	^ requestedCount = readCount
		ifTrue: [ collection ]
		ifFalse: [ collection copyFrom: 1 to: offset + readCount - 1 ]! !

!ZnChunkedReadStream methodsFor: 'initialize-release' stamp: ''!
on: readStream
	stream := readStream! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
peek
	self ensureChunkOrAtEnd.
	self atEnd ifTrue: [ ^ nil ].
	^ chunk at: (position + 1)! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
position
	^ self atEnd 
		ifTrue: [ totalSize ]
		ifFalse: [ totalSize - chunk size + position ]! !

!ZnChunkedReadStream methodsFor: 'printing' stamp: ''!
printOn: aStream
	aStream 
		nextPutAll: 'a '; 
		nextPutAll: self class name! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
readInto: collection startingAt: offset count: requestedCount
	"Read requestedCount elements into collection starting at offset.
	Return the actual number of elements read."

	| read toRead |
	read := 0.
	[ read < requestedCount and: [ self atEnd not] ] whileTrue: [
		toRead := (limit - position) min: (requestedCount - read).
		collection 
			replaceFrom: offset + read 
			to: offset + read + toRead - 1 
			with: chunk 
			startingAt: position + 1.
		read := read + toRead.
		position := position + toRead ].
	^ read! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
reset
	"We don't allow it"! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
skip: count
	count timesRepeat: [ self next ]! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
totalSize
	^ totalSize! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
upTo: anObject 
	"We use our own collectionSpecies."
	
	^ self collectionSpecies streamContents: [ :out | | element |
		[ self atEnd or: [ (element := self next) = anObject ] ]
			whileFalse: [ out nextPut: element ] ]! !

!ZnChunkedReadStream methodsFor: 'accessing' stamp: ''!
upToEnd
	"We use our own collectionSpecies and read by chunk"

	^ self collectionSpecies streamContents: [ :out |
		[ self atEnd ] whileFalse: [ 
			out next: limit putAll: chunk.
			position := limit ] ]! !

Stream subclass: #ZnLimitedReadStream
	instanceVariableNames: 'stream position limit binary'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Streaming'!

!ZnLimitedReadStream commentStamp: '' prior: 0!
I am ZnLimitedReadStream, wrapping another read stream delegating to it 
but limiting reading to a fixed number of elements.

I can be atEnd while my wrapped stream is not.

My contentSpecies can be forced to be ByteArray (binary) or ByteString (ascii).

I do byte/char conversions on the fly (support bivalent access).

This implementation is really a kludge.

Part of Zinc HTTP Components.!

!ZnLimitedReadStream class methodsFor: 'instance creation' stamp: ''!
on: stream limit: limit
	^ self basicNew
		on: stream limit: limit;
		yourself! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
ascii
	binary := false! !

!ZnLimitedReadStream methodsFor: 'testing' stamp: ''!
atEnd
	^ position >= limit or: [ stream atEnd ]! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
binary
	binary := true! !

!ZnLimitedReadStream methodsFor: 'initialize-release' stamp: ''!
close
	stream ifNotNil: [ stream close. stream := nil ]! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
collectionSpecies
	"I can be set to binary or ascii explicitely.
	If not set, delegate to our wrapped stream."
	
	^ binary notNil
		ifTrue: [ 
			binary ifTrue: [ ByteArray ] ifFalse: [ ByteString ] ]
		ifFalse: [
			(stream respondsTo: #collectionSpecies)
				ifTrue: [ stream collectionSpecies ]  
				ifFalse: [ stream isBinary ifTrue: [ ByteArray ] ifFalse: [ ByteString ] ] ]! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
contents
	"This is technically not correct, but it is better than nothing"
	
	^ self upToEnd! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
contentsOfEntireFile
	^ self upToEnd! !

!ZnLimitedReadStream methodsFor: 'testing' stamp: ''!
isBinary
	^ binary isNil
		ifTrue: [ stream isBinary ]
		ifFalse: [ binary ]! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
match: subCollection
	"Try to read and match the elements of subCollection. 
	If successful return true and leave the me positioned after the match.
	If unsuccesful return false and leave me at end."
	
	| buffer pattern bufferPosition bufferLimit |
	pattern := subCollection readStream.
	"we have to use an internal buffer because we are not positionable"
	buffer := subCollection species new: subCollection size.
	bufferPosition := bufferLimit := 0.
	[ pattern atEnd ] whileFalse: [ | nextElement |
		self atEnd ifTrue: [ ^ false ].
		"get the next char to match either from the buffer or from the stream while buffering it"
		(bufferPosition < bufferLimit) 
			ifTrue: [ 
				nextElement := buffer at: bufferPosition + 1 ]
			ifFalse: [
				nextElement := buffer at: bufferPosition + 1 put: self next.
				bufferLimit := bufferLimit + 1 ].
		bufferPosition := bufferPosition + 1.
		pattern next = nextElement
			ifFalse: [
				"shift the buffer down one element and restart"
				2 to: bufferLimit do: [ :each |
					buffer at: each - 1 put: (buffer at: each) ].
				bufferLimit := bufferLimit - 1.
				bufferPosition := 0.
				pattern position: 0 ] ].
	^ true! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
next
	"We do byte/char conversions on the fly."
	
	^ self atEnd 
		ifTrue: [ nil ] 
		ifFalse: [ 
			position := position + 1. 
			binary isNil
				ifTrue: [ stream next ] 
				ifFalse: [ 
					binary
					 	ifTrue: [ stream next asInteger ]
						ifFalse: [ stream next asCharacter ] ] ]! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
next: requestedCount 
	"Read requestedCount elements into new collection and return it,
	 it could be that less elements were available"

	^ self 
		next: requestedCount 
		into: (self collectionSpecies new: requestedCount)! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
next: requestedCount into: collection
	"Read requestedCount elements into collection,
	returning a copy if less elements are available"
	
	^ self
		next: requestedCount
		into: collection
		startingAt: 1! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
next: requestedCount into: collection startingAt: offset
	"Read requestedCount elements into collection starting at offset,
	returning a copy if less elements are available"
	
	| readCount |
	readCount := self 
		readInto: collection 
		startingAt: offset 
		count: requestedCount.
	^ requestedCount = readCount
		ifTrue: [ collection ]
		ifFalse: [ collection copyFrom: 1 to: offset + readCount - 1 ]! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
nextInto: collection
	"Read the next elements of the receiver into collection,
	returning a copy if less elements are available"
	
	^ self
		next: collection size
		into: collection! !

!ZnLimitedReadStream methodsFor: 'initialize-release' stamp: ''!
on: readStream limit: integer
	stream := readStream.
	limit := integer.
	position := 0! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
peek
	"We do byte/char conversions on the fly."
	
	^ self atEnd 
		ifTrue: [ nil ] 
		ifFalse: [ 
			binary isNil
				ifTrue: [ stream peek ] 
				ifFalse: [ 
					binary
					 	ifTrue: [ stream peek asInteger ]
						ifFalse: [ stream peek asCharacter ] ] ]! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
peekFor: anElement
	^ self peek = anElement and: [ self next. true ]! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
position
	^ position! !

!ZnLimitedReadStream methodsFor: 'printing' stamp: ''!
printOn: aStream
	aStream 
		nextPutAll: 'a '; 
		nextPutAll: self class name! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
readInto: collection startingAt: offset count: count 
	"Read count elements and place them in collection starting at offset.
	Return the number of elements actually read."
	
	| target actual |
	target := count min: (limit - position).
	actual := stream 
		readInto: collection 
		startingAt: offset 
		count: target.
	position := position + actual.
	^ actual! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
reset
	"We don't allow it"! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
skip: count
	count timesRepeat: [ self next ]! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
upTo: anObject 
	"We use our own collectionSpecies."
	
	^ self collectionSpecies streamContents: [ :out | | element |
		[ self atEnd or: [ (element := self next) = anObject ] ] whileFalse: [ 
			out nextPut: element ] ]! !

!ZnLimitedReadStream methodsFor: 'accessing' stamp: ''!
upToEnd
	^ ZnUtils readUpToEnd: self limit: nil! !

ProcessLocalVariable subclass: #ZnCurrentServerSession
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Variables'!

!ZnCurrentServerSession commentStamp: '' prior: 0!
I am ZnCurrentServerSession.
I am a ProcessLocalVariable and a ProcessSpecificVariable.

I hold a reference to the current ZnServerSession instance during HTTP processing, during the dynamic scope of #handleRequest:

ZnServerSessions are created lazily.
!

DynamicVariable subclass: #ZnConnectionTimeout
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Variables'!

!ZnConnectionTimeout commentStamp: '' prior: 0!
I am ZnConnectionTimeout.
I am a DynamicVariable and a ProcessSpecificVariable.

I can be used to modify the global default socket IO timeout on a per process basis, for example:

ZnConnectionTimeout 
	value: 5 
	during: [ ^ ZnClient new get: 'http://zn.stfx.eu/zn/small.html' ]!

DynamicVariable subclass: #ZnCurrentServer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Variables'!

!ZnCurrentServer commentStamp: '' prior: 0!
I am ZnCurrentServer.
I am a DynamicVariable and a ProcessSpecificVariable.

I hold a reference to the current ZnServer instance during HTTP processing, during the dynamic scope of #handleRequest:!

DynamicVariable subclass: #ZnMaximumEntitySize
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Variables'!

!ZnMaximumEntitySize commentStamp: '' prior: 0!
I am ZnMaximumEntitySize.
I am a DynamicVariable and a ProcessSpecificVariable.

I am used to control the maximum size of entities being read from streams.
!

!ZnMaximumEntitySize methodsFor: 'accessing' stamp: ''!
default
	"By default, allow for entities of up to 16 Mb to be read from streams"
	
	^ 16 * 1024 * 1024! !

DynamicVariable subclass: #ZnSignalProgress
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Variables'!

!ZnSignalProgress commentStamp: '' prior: 0!
I am ZnSignalProgress.
I am a DynamicVariable and a ProcessSpecificVariable.

I can be used to indicate that Zn related code down the call stack has to signal HTTPProgress, for example:

ZnSignalProgress 
	value: true 
	during: [ ^ ZnClient new get: 'http://zn.stfx.eu/zn/small.html' ]!

!ZnSignalProgress class methodsFor: 'accessing' stamp: ''!
enabled
	"Answer if HTTPProgress should be signalled.
	The default is false."
	
	^ self value
		ifNil: [ false ]
		ifNotNil: [ :value | value ]! !

Object subclass: #ZnBasicAuthenticator
	instanceVariableNames: 'credentials realm'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Support'!

!ZnBasicAuthenticator commentStamp: '' prior: 0!
I am ZnBasicAuthenticator.
I help servers handle HTTP Basic Authentication.
I implement the required #authenticateRequest:do:

I have a credentials dictionary with username->password associations.

Part of Zinc HTTP Components.!

!ZnBasicAuthenticator class methodsFor: 'accessing' stamp: ''!
defaultRealm
	^ 'ZincHTTPComponents'! !

!ZnBasicAuthenticator class methodsFor: 'instance creation' stamp: ''!
username: username password: password
	^ self new
		atUsername: username putPassword: password;
		yourself! !

!ZnBasicAuthenticator class methodsFor: 'instance creation' stamp: ''!
username: username password: password realm: realm
	^ self new
		atUsername: username putPassword: password;
		realm: realm;
		yourself! !

!ZnBasicAuthenticator methodsFor: 'accessing' stamp: ''!
atUsername: username putPassword: password
	credentials at: username put: password! !

!ZnBasicAuthenticator methodsFor: 'public' stamp: ''!
authHeader
	^ 'Basic realm=', self realm! !

!ZnBasicAuthenticator methodsFor: 'public' stamp: ''!
authenticateRequest: request do: block
	^ (self isRequestAuthenticated: request)
		ifTrue: [ block value ]
		ifFalse: [ ZnResponse unauthorized: self authHeader ]! !

!ZnBasicAuthenticator methodsFor: 'initialization' stamp: ''!
initialize
	credentials := Dictionary new! !

!ZnBasicAuthenticator methodsFor: 'testing' stamp: ''!
isRequestAuthenticated: request
	| authorization username password |
	authorization := [ request basicAuthentication ] on: Error do: [ ^ false ].
	username := authorization first.
	password := authorization second.
	^ (credentials at: username ifAbsent: [ nil ]) = password
	! !

!ZnBasicAuthenticator methodsFor: 'accessing' stamp: ''!
realm
	^ realm ifNil: [ realm := self class defaultRealm ]! !

!ZnBasicAuthenticator methodsFor: 'accessing' stamp: ''!
realm: anObject
	realm := anObject! !

!ZnBasicAuthenticator methodsFor: 'accessing' stamp: ''!
removeAll
	credentials removeAll! !

!ZnBasicAuthenticator methodsFor: 'accessing' stamp: ''!
removeUsername: username
	credentials removeKey: username! !

ZnBasicAuthenticator subclass: #ZnDigestAuthenticator
	instanceVariableNames: 'nonces'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Support'!

!ZnDigestAuthenticator commentStamp: '' prior: 0!
I am ZnDigestAuthenticator.
I help servers handle HTTP Digest Authentication.

I have a nonces dictionary with nonce->opaque associations. 
I generate a new nonce value for every 401 response I trigger.

Part of Zinc HTTP Components.!

!ZnDigestAuthenticator class methodsFor: 'testing' stamp: ''!
hasMD5Support
	^ (self md5Hash: 'test') isNil not! !

!ZnDigestAuthenticator class methodsFor: 'accessing' stamp: ''!
md5Hash: aString
	| hash |
	"Answer hash of aString as lowercase 32 digit hex String."
	Smalltalk at: #MD5 ifPresent: [ :cls |
		^ ((cls new initialize) hashStream: (aString asByteArray readStream)) hex ].
	hash := Smalltalk at: #CMD5Hasher ifPresent: [ :cls |
		cls hashMessage: aString ].
	hash ifNil: [
		hash := Smalltalk at: #TCryptoRandom ifPresent: [ :cls |
			(cls basicNew perform: #md5HashMessage: with: aString) asInteger ] ].
	hash ifNotNil: [
		hash := hash hex asLowercase.
		(hash beginsWith: '16r') ifTrue: [ hash := hash allButFirst: 3 ].
		hash := hash padLeftTo: 32 with: $0 ].
	^ hash
! !

!ZnDigestAuthenticator class methodsFor: 'accessing' stamp: ''!
parseAuthRequest: headerValue
	| data dict |
	dict := Dictionary new.
	"Chops off the ' Digest ' scheme name, a bit too brittle."
	data := headerValue copyFrom: 8 to: headerValue size.
	(data substrings: ',') do: [ :fragment | | tokens i key value |
		tokens := fragment trimBoth.
		i := tokens indexOf: $=.
		(i > 0) ifTrue: [
			key := tokens copyFrom: 1 to: i - 1.
			value := tokens copyFrom: i + 1 to: tokens size.  
			dict at: key put: (value trimBoth: [ :char | char = $" ]) ] ].
	^ dict! !

!ZnDigestAuthenticator methodsFor: 'private' stamp: ''!
a1for: username
	| password |
	password := credentials at: username ifAbsent: [ ^ nil ].
	^ self class md5Hash: username, ':', self realm, ':', password
	! !

!ZnDigestAuthenticator methodsFor: 'private' stamp: ''!
a2forUrl: uri method: method
	^ self class md5Hash: method, ':', uri
	! !

!ZnDigestAuthenticator methodsFor: 'public' stamp: ''!
authHeader
	| nonce opaque |
	nonce := self createNonce.
	opaque := self createOpaque.
	self nonces at: nonce put: opaque.
	^ 'Digest realm="', self realm, '", nonce="', nonce, '", opaque="', opaque, '"'! !

!ZnDigestAuthenticator methodsFor: 'private' stamp: ''!
createNonce
	| pt |
	pt := DateAndTime now asString, ':', Random new next asString.
	^ self class md5Hash: pt! !

!ZnDigestAuthenticator methodsFor: 'private' stamp: ''!
createOpaque
	| pt |
	pt := Random new next asString, ':', DateAndTime now asString.
	^ self class md5Hash: pt! !

!ZnDigestAuthenticator methodsFor: 'testing' stamp: ''!
isRequestAuthenticated: request
	| authorization response cresponse a1 a2 nonce opaque |
	authorization := self class parseAuthRequest: (request headers at: 'Authorization' ifAbsent: [ ^ false ]).
	nonce := authorization at: 'nonce' ifAbsent: [ ^ false ].
	opaque := self nonces at: nonce ifAbsent: [ ^ false ].
	a1 := self a1for: (authorization at: 'username' ifAbsent: [ ^ false ]).
	a1 ifNil: [ ^ false ].
	a2 := self a2forUrl: (authorization at: 'uri' ifAbsent: [ '/' ]) method: request method.
	response := self class md5Hash: (a1, ':', nonce, ':', a2).
	"(opaque = (authorization at: 'opaque' ifAbsent: [^false])) ifFalse: [^false]."
	cresponse := authorization at: 'response' ifAbsent: [ ^ false ].
	^ (response = cresponse)! !

!ZnDigestAuthenticator methodsFor: 'accessing' stamp: ''!
nonces 
	^ nonces ifNil: [ nonces := Dictionary new ]! !

Object subclass: #ZnChunkedWriteStream
	instanceVariableNames: 'stream extraHeaders totalSize chunkCount'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Streaming'!

!ZnChunkedWriteStream commentStamp: '' prior: 0!
I am ZnChunkedWriteStream, I wrap another write stream and add HTTP 1.1 chunked transfer encoding.

I write a chunk for each #nextPutAll: or #next:putAll:startingAt:
I should be wrapped in a ZnBufferedWriteStream for #next: to work.
When done, send #close or #finish to me.

Part of Zinc HTTP Components.
!

!ZnChunkedWriteStream class methodsFor: 'instance creation' stamp: ''!
on: writeStream
	^ self new
		on: writeStream;
		yourself! !

!ZnChunkedWriteStream methodsFor: 'initialize-release' stamp: ''!
binary
	stream binary! !

!ZnChunkedWriteStream methodsFor: 'accessing' stamp: ''!
chunkCount
	^ chunkCount! !

!ZnChunkedWriteStream methodsFor: 'initialize-release' stamp: ''!
close
	self finish.
	stream close! !

!ZnChunkedWriteStream methodsFor: 'private' stamp: ''!
crlf
	stream nextPutAll: (self isBinary ifTrue: [ #[ 13 10 ] ] ifFalse: [ String crlf ])! !

!ZnChunkedWriteStream methodsFor: 'accessing' stamp: ''!
extraHeaders
	^ extraHeaders ifNil: [ extraHeaders := ZnHeaders new ]! !

!ZnChunkedWriteStream methodsFor: 'initialize-release' stamp: ''!
finish
	self writeChunkSize: 0.
	self crlf.
	extraHeaders ifNotNil: [ extraHeaders writeOn: stream ]! !

!ZnChunkedWriteStream methodsFor: 'initialize-release' stamp: ''!
flush
	stream flush! !

!ZnChunkedWriteStream methodsFor: 'initialization' stamp: ''!
initialize
	super initialize.
	totalSize := 0.
	chunkCount := 0! !

!ZnChunkedWriteStream methodsFor: 'testing' stamp: ''!
isBinary
	^ stream isBinary
! !

!ZnChunkedWriteStream methodsFor: 'testing' stamp: ''!
isStream
	^ true! !

!ZnChunkedWriteStream methodsFor: 'accessing' stamp: ''!
next: count putAll: collection
	self 
		next: count 
		putAll: collection 
		startingAt: 1! !

!ZnChunkedWriteStream methodsFor: 'accessing' stamp: ''!
next: count putAll: collection startingAt: offset
	self writeChunkSize: count.
	stream next: count putAll: collection startingAt: offset.
	self crlf.
	chunkCount := chunkCount + 1.
	totalSize := totalSize + count
! !

!ZnChunkedWriteStream methodsFor: 'accessing' stamp: ''!
nextPut: element
	self error: 'Wrap me into a buffered stream, because each #nexPut: will create a new chunk'! !

!ZnChunkedWriteStream methodsFor: 'accessing' stamp: ''!
nextPutAll: collection
	self 
		next: collection size
		putAll: collection
		startingAt: 1! !

!ZnChunkedWriteStream methodsFor: 'initialize-release' stamp: ''!
on: writeStream
	stream := writeStream! !

!ZnChunkedWriteStream methodsFor: 'accessing' stamp: ''!
position
	^ totalSize! !

!ZnChunkedWriteStream methodsFor: 'private' stamp: ''!
writeChunkSize: size
	stream nextPutAll: (self isBinary ifTrue: [ size printStringHex asByteArray ] ifFalse: [ size printStringHex ]).
	self crlf! !

Object subclass: #ZnClient
	instanceVariableNames: 'request response connection lastUsed options session logLevel'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Client-Server'!

!ZnClient commentStamp: '' prior: 0!
I am ZnClient, an object to build, execute and process HTTP client requests.

I have a rich protocol to construct requests and to access responses.
I have various error handling options.
I can reuse an existing connection to a specific host:port.
I handle sessions, cookies, redirects and authentication.
I have many options (settings) with sensible defaults.

Simplest possible invocation:

	ZnClient new
		get: 'http://zn.stfx.eu/zn/numbers.txt'.
	
Using some features to make a better HTTP request:

	ZnClient new
		systemPolicy;
		accept: ZnMimeType textPlain;
		http;
		host: 'zn.stfx.eu';
		path: 'zn/numbers.txt';
		contentReader: [ :entity | entity contents lines collect: [ :each | each asNumber ] ];
		ifFail: [ :exception | self inform: 'I am sorry: ', exception printString ];
		get.

Part of Zinc HTTP Components.!

!ZnClient methodsFor: 'accessing request' stamp: ''!
accept: mimeTypeObject
	"Set the Accept HTTP header to mimeTypeObject for the current request."

	| mimeType |
	mimeType := mimeTypeObject asZnMimeType.
	self request setAccept: mimeType! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
addPart: mimePart
	"Add mimePart to the multipart/form-data entity of the current request"

	self multiPartFormDataEntity 
		addPart: mimePart ! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
addPath: object
	"Extend the path of the current request with object.
	Either a string or a collection of strings is allowed as argument."
	
	object isString
		ifTrue: [ self addPathSegment: object ] 
		ifFalse: [ object do: [ :each | self addPathSegment: each ] ]! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
addPathSegment: pathSegment
	"Extend the path of the current request with a single pathSegment."
	
	request url 
		ifNil: [ request url: (ZnUrl new addPathSegment: pathSegment; yourself) ]
		ifNotNil: [ request url addPathSegment: pathSegment ]  ! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
applicationFormUrlEncodedEntity
	"Initialize the current request to have an application/x-www-form-urlencoded entity.
	Return the existing entity if there is one. 
	Fail if another type was already set."
	
	request hasEntity
		ifTrue: [
			request entity contentType = ZnMimeType applicationFormUrlEncoded 
				ifFalse: [ self error: 'An incompatible entity was already set for the request' ] ]
		ifFalse: [
			request entity: ZnApplicationFormUrlEncodedEntity new ].
	^ request entity! !

!ZnClient methodsFor: 'options' stamp: ''!
autoResetEntityMethods
	"Return a collection of HTTP methods for which to auto #resetEntity.
	The default is to do this only for HEAD and DELETE, although most people 
	would want it for GET too, but this can be used for forms"
	
	^ self optionAt: #autoResetEntityMethods ifAbsentPut: [ #(HEAD DELETE) ]! !

!ZnClient methodsFor: 'options' stamp: ''!
autoResetEntityMethods: methods
	"Set the collection of HTTP methods for which to auto #resetEntity.
	The default is to do this only for HEAD and DELETE, although most people 
	would want it for GET too, but this can be used for forms"
	
	^ self optionAt: #autoResetEntityMethods put: methods! !

!ZnClient methodsFor: 'options' stamp: ''!
beBinary
	"Configure me to read HTTP responses without interpreting them further,
	their resources/entities will remain bytes."
	
	self reader: [ :stream | ZnResponse readBinaryFrom: stream ]! !

!ZnClient methodsFor: 'options' stamp: ''!
beOneShot
	"Switch me to use connections only for a single request/response cycle.
	By default, I will keep HTTP connections open according to the HTTP 1.1 spec.
	When oneShot is true, I will add a Connection: close header to requests and close after one cycle.
	I cannot be one shot when #streaming is true because the stream would be closed too soon."
	
	self streaming
		ifTrue: [ self error: 'I cannot be one shot when #streaming is true.' ].
	self optionAt: #oneShot put: true   ! !

!ZnClient methodsFor: 'private' stamp: ''!
canReuseConnection
	self isConnected ifFalse: [ ^ false ].
	lastUsed ifNil: [ ^ false ].
	^ Time totalSeconds - lastUsed < self connectionReuseTimeout! !

!ZnClient methodsFor: 'options' stamp: ''!
certificate
	"Return the TLS/SSL certificate to be used, or nil for none."
	
	^ self optionAt: #certificate ifAbsent: [ nil ]! !

!ZnClient methodsFor: 'options' stamp: ''!
certificate: anObject
	"Set the TLS/SSL certificate to be used to anObject, or nil for none.
	See ZdcAbstractSSLSession>>#certificateName: for more information."
	
	^ self optionAt: #certificate put: anObject! !

!ZnClient methodsFor: 'private' stamp: ''!
cleanupConnection
	(self isOneShot or: [ request notNil and: [ request wantsConnectionClose ] ])
		ifTrue: [
			self close ]! !

!ZnClient methodsFor: 'options' stamp: ''!
clientId
	"Return the id for this HTTP client. Defaults to nil.
	Can be used identify different clients' log events."
	
	^ self optionAt: #clientId ifAbsent: [ nil ]! !

!ZnClient methodsFor: 'options' stamp: ''!
clientId: anObject
	"Set the id for this HTTP client to anObject. Defaults to nil.
	Can be used identify different clients' log events."
	
	^ self optionAt: #clientId put: anObject! !

!ZnClient methodsFor: 'cleanup' stamp: ''!
close
	"Close any HTTP network connection that I might have kept open."

	connection
		ifNotNil: [ 
			| address port |
			address := [ connection socket remoteAddress ] on: Error do: [ nil ].
			port := [ connection socket remotePort ] on: Error do: [ nil ].
			connection close.
			connection := nil.
			self logConnectionClosedAddress: address port: port ].
	lastUsed := nil! !

!ZnClient methodsFor: 'accessing' stamp: ''!
connection
	"Return the socket stream connection that I am currently using"
	
	^ connection! !

!ZnClient methodsFor: 'options' stamp: ''!
connectionReuseTimeout
	"Return the timeout in seconds currently used to determine
	if I should try to reuse an open connection. Defaults to 10 seconds."

	^ self optionAt: #connectionReuseTimeout ifAbsent: [ 10 ]! !

!ZnClient methodsFor: 'options' stamp: ''!
connectionReuseTimeout: seconds
	"Set the timeout used to determine if I should try 
	to reuse an open connection to seconds."

	self optionAt: #connectionReuseTimeout put: seconds! !

!ZnClient methodsFor: 'options' stamp: ''!
contentReader
	"Return the current content reader, an object conforming 
	to the #value: protocol to convert a response entity
	as used by #contents. Defaults to nil.
	Can be used to parse incoming data."
	
	^ self optionAt: #contentReader ifAbsent: [ nil ]! !

!ZnClient methodsFor: 'options' stamp: ''!
contentReader: object
	"Set the content reader, an object conforming 
	to the #value: protocol to convert a response entity
	as used by #contents to object.
	Can be used to parse incoming data."

	self optionAt: #contentReader put: object! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
contentType: mimeType
	"Set the content-type of the request entity, provided we have one"
	
	request hasEntity
		ifTrue: [ 
			request entity contentType: mimeType asZnMimeType ]
		ifFalse: [ 
			self error: 'There is no request entity yet, cannot set its content-type' ]! !

!ZnClient methodsFor: 'options' stamp: ''!
contentWriter
	"Return the current content writer, an object conforming 
	to the #value: protocol to convert an object to a request content entity
	as used by #contents:. Defaults to nil.
	Can be used to format data to be uploaded."

	^ self optionAt: #contentWriter ifAbsent: [ nil ]! !

!ZnClient methodsFor: 'options' stamp: ''!
contentWriter: object
	"Set current content writer, an object conforming 
	to the #value: protocol to convert an object to a request content entity
	as used by #contents:. Defaults to nil.
	Can be used to format data to be uploaded."

	self optionAt: #contentWriter put: object! !

!ZnClient methodsFor: 'accessing response' stamp: ''!
contents 
	"Answer the contents or HTTP payload of the last response, 
	assuming the current request was executed.
	If I have a contentReader, it will be used to compute the contents.
	Otherwise I delegate to ZnResponse>>#contents, unless I am #streaming,
	in which case I return the stream of the streaming entity."

	^ response ifNotNil: [ 
		self contentReader 
			ifNil: [ 
				self streaming
					ifTrue: [  response entity stream ]
					ifFalse: [ response contents ] ]
			ifNotNil: [ :reader | reader value: response entity ] ]  ! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
contents: object
	"Set the entity contents of the current request to object.
	I will pass this through my contentWriter if I have one.
	Otherwise I delegate to ZnEntity to convert object to a generic entity."
	
	self contentWriter 
		ifNil: [ self entity: (ZnEntity with: object) ]
		ifNotNil: [ :writer | self entity: (writer value: object) ]! !

!ZnClient methodsFor: 'convenience' stamp: ''!
curl
	"Create and return a curl command line invocation equivalent to the current request."
	"ZnClient new url: 'http://zn.stfx.eu/zn/small.html'; method: #GET; curl."
	"ZnClient new url: 'http://zn.stfx.eu/echo'; contents: #[2 4 6 8]; method: #POST; curl."
	
	self prepareRequest.
	^ String streamContents: [ :stream |
		request hasEntity ifTrue: [ 
			 stream << 'echo ' << request entity asByteArray hex << ' | xxd -r -p | ' ].
		stream << 'curl'.
		request method = #GET ifFalse: [ stream << ' -X ' << request method ].
		stream space; print: request uri.
		request headersDo: [ :key :value |
			stream << ' -H"' << key << ':' << value << '"' ].
		request hasEntity ifTrue: [ stream << ' --data-binary @-' ] ]! !

!ZnClient methodsFor: 'initialization' stamp: ''!
debugPolicy
	self 
		beOneShot! !

!ZnClient methodsFor: 'options' stamp: ''!
defaultEncoder
	"The default character encoder to use when none is set in a mime-type"
	
	^ self optionAt: #defaultEncoder ifAbsent: [ ZnDefaultCharacterEncoder value ] ! !

!ZnClient methodsFor: 'options' stamp: ''!
defaultEncoder: encoder
	"Set the default character encoder to use when none is set in a mime-type"
	
	^ self optionAt: #defaultEncoder put: encoder asZnCharacterEncoder! !

!ZnClient methodsFor: 'operations' stamp: ''!
delete
	"Execute a HTTP DELETE ont the request set up and return the response #contents."
	
	^ self 
		method: #DELETE; 
		execute! !

!ZnClient methodsFor: 'convenience' stamp: ''!
delete: url
	"Execute an HTTP DELETE request on url and return the response #contents."
	
	^ self 
		url: url; 
		delete! !

!ZnClient methodsFor: 'options' stamp: ''!
dontFollowRedirects
	"Disable the default automatic following of HTTP redirects.
	Redirects will be returned as normal responses."

	^ self followRedirects: false! !

!ZnClient methodsFor: 'convenience' stamp: 'SvenVanCaekenberghe 11/1/2018 23:54'!
downloadEntityTo: path
	"If I hold an entity, download it directly to a file.
	If path describes a file, use that. 
	If path describes a directory, use the last path segment of the URI to construct a file in that directory.
	The URI of the last request is used, which could make a difference when following redirects."

	| target |
	self entity ifNil: [ ^ self ].
	target := ZnFileSystemUtils downloadPathTo: path for: self request uri lastPathSegment.
	ZnFileSystemUtils
		binaryWriteStreamFor: target
		do: [ :fileStream |
			self withProgressDo: [ self entity writeOn: fileStream ] ]! !

!ZnClient methodsFor: 'convenience' stamp: ''!
downloadTo: path
	"Setup streaming, do a GET request and download the entity directly to a file.
	If path describes a file, use that. 
	If path describes a directory, use the last path segment of the URI to construct a file in that directory.
	The URI of the first request is used, which could make a difference when following redirects.
	Return true when successful, false otherwise."

	| target |
	target := ZnFileSystemUtils downloadPathTo: path for: self request uri lastPathSegment.
	self
		streaming: true;
		get.
	^ self isSuccess
		ifTrue: [ 
			self downloadEntityTo: target.
			true ]
		ifFalse: [ false ]! !

!ZnClient methodsFor: 'options' stamp: ''!
enforceAcceptContentType
	"Return whether or not I enforce that the content type of the
	response matches the content type specified in the accept request header."
	
	^ self optionAt: #enforceAcceptContentType ifAbsent: [ false ]! !

!ZnClient methodsFor: 'options' stamp: ''!
enforceAcceptContentType: boolean
	"Set whether or not I enforce that the content type of the
	response matches the content type specified in the accept request
	header to boolean."

	self optionAt: #enforceAcceptContentType put: boolean! !

!ZnClient methodsFor: 'options' stamp: ''!
enforceHttpSuccess
	"Return whether or not I enforce that the response 
	to the request is an HTTP success code."
	
	^ self optionAt: #enforceHttpSuccess ifAbsent: [ false ]! !

!ZnClient methodsFor: 'options' stamp: ''!
enforceHttpSuccess: boolean
	"Set whether or not I enforce that the response 
	to the request is an HTTP success code to boolean."

	self optionAt: #enforceHttpSuccess put: boolean! !

!ZnClient methodsFor: 'accessing response' stamp: ''!
entity
	"Answer the entity or HTTP payload of the last response, 
	assuming the current request was executed."
	
	^ response ifNotNil: [ response entity ]  ! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
entity: entity
	"Set the entity of the current request to enity.
	A ZnEntity subclass or compatible object is expected."
	
	request entity: entity! !

!ZnClient methodsFor: 'operations' stamp: ''!
execute
	"Execute the currently set up request to generate a response.
	Return the #contents of the response, if any."

	| result |
	self prepareRequest.
	self withProgressDo: [ result := self executeWithTimeout ].
	self prepareNextRequest.
	^ result! !

!ZnClient methodsFor: 'private protocol' stamp: ''!
executeRequestResponse
	| initialMilliseconds requestDuration responseDuration |
	initialMilliseconds := Time millisecondClockValue.
	self writeRequest.
	requestDuration := self logRequestWrittenStarted: initialMilliseconds.
	initialMilliseconds := Time millisecondClockValue.
	self readResponse.
	responseDuration := self logResponseReadStarted: initialMilliseconds.
	self logClientTransactionRequestDuration: requestDuration responseDuration: responseDuration! !

!ZnClient methodsFor: 'private protocol' stamp: ''!
executeWithRedirectsRemaining: redirectCount
	self getConnectionAndExecute.
	response isRedirect 
		ifTrue: [
			(redirectCount > 0 and: [ self followRedirects ])
				ifTrue: [
					self 
						prepareRedirect; 
						executeWithRedirectsRemaining: redirectCount - 1 ]
				ifFalse: [
					self followRedirects 
						ifTrue: [ ZnTooManyRedirects signal ] ] ].
	^ self handleResponse
	! !

!ZnClient methodsFor: 'private protocol' stamp: ''!
executeWithRetriesRemaining: retryCount
	^ [ self executeWithRedirectsRemaining: self maxNumberOfRedirects ]
		on: self retryExceptionSet
		do: [ :exception |
			retryCount > 0
				ifTrue: [
					self 
						handleRetry: exception;
						executeWithRetriesRemaining: retryCount - 1 ]
				ifFalse: [
					exception pass ] ]! !

!ZnClient methodsFor: 'private protocol' stamp: ''!
executeWithTimeout
	^ self withTimeoutDo: [ 
		[ self executeWithRetriesRemaining: self numberOfRetries ]
			on: Error
			do: self ifFailBlock ]! !

!ZnClient methodsFor: 'options' stamp: ''!
followRedirects
	"Return whether I will automatically following of HTTP redirects.
	This is still subject to #maxNumberOfRedirects. Defaults to true.
	When set to false, redirects will be returned as normal responses."

	^ self optionAt: #followRedirects ifAbsent: [ true ]! !

!ZnClient methodsFor: 'options' stamp: ''!
followRedirects: boolean
	"Set whether or not I will automatically following of HTTP redirects.
	This is still subject to #maxNumberOfRedirects. Defaults to true.
	When set to false, redirects will be returned as normal responses."

	self optionAt: #followRedirects put: boolean ! !

!ZnClient methodsFor: 'initialization' stamp: 'SvenVanCaekenberghe 11/7/2018 14:31'!
forJsonREST
	"Configure me for JSON REST interaction: I assume that the entities that I read and write are JSON.
	This requires either NeoJSON or STON to be present."

	| reader writer |
	reader := self class environment
		at: #NeoJSONObject
		ifAbsent: [ 
			self class environment 
				at: #STONJSON 
				ifAbsent: [ self error: 'Could not find a suitabe JSON parser' ] ].
	writer := self class environment
		at: #NeoJSONWriter
		ifAbsent: [ 
			self class environment 
				at: #STONJSON 
				ifAbsent: [ self error: 'Could not find a suitabe JSON writer' ] ].
	self
		accept: ZnMimeType applicationJson;
		contentReader: [ :entity | reader fromString: entity contents ];
		contentWriter: [ :data | ZnEntity json: (writer toString: data) ]
! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
formAdd: association
	"Add the key->value association to the application/x-www-form-urlencoded 
	entity of the current request."
	
	self applicationFormUrlEncodedEntity 
		add: association ! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
formAddAll: keyedCollection
	"Add all key->value associations of keyedCollection to 
	the application/x-www-form-urlencoded entity of the current request."

	self applicationFormUrlEncodedEntity 
		addAll: keyedCollection ! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
formAt: key add: value
	"Add key equals value to the application/x-www-form-urlencoded 
	entity of the current request. This is for multi-values form fields."

	self applicationFormUrlEncodedEntity 
		at: key add: value! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
formAt: key put: value
	"Set key equal to value in the application/x-www-form-urlencoded 
	entity of the current request."

	self applicationFormUrlEncodedEntity 
		at: key put: value! !

!ZnClient methodsFor: 'operations' stamp: ''!
get
	"Execute an HTTP GET the request set up and return the response #contents."
	
	^ self 
		method: #GET; 
		execute! !

!ZnClient methodsFor: 'convenience' stamp: ''!
get: url
	"Execute an HTTP GET request on url and return the response #contents."

	^ self 
		url: url; 
		get! !

!ZnClient methodsFor: 'private protocol' stamp: ''!
getConnectionAndExecute
	self canReuseConnection 
		ifTrue: [ 
			[ ^ self executeRequestResponse; cleanupConnection ] 
				on: Exception 
				do: [ :exception | self logIgnoringExceptionOnReusedConnection: exception; close ] ].
	self newConnectionTo: request url.
	[ self executeRequestResponse ]
		ensure: [ self cleanupConnection ]! !

!ZnClient methodsFor: 'private protocol' stamp: ''!
handleResponse
	self receiveCookies.
	(self enforceHttpSuccess and: [ response isSuccess not ])
		ifTrue: [ 
			(self followRedirects not and: [ response isRedirect ])
				ifFalse: [ (ZnHttpUnsuccessful response: response) signal ] ].
	(self enforceAcceptContentType and: [ self isContentTypeAcceptable not ])
		ifTrue: [ 
			(ZnUnexpectedContentType 
				expected: request accept asZnMimeType 
				actual: response contentType) signal ].
	^ self contents! !

!ZnClient methodsFor: 'private protocol' stamp: ''!
handleRetry: exception
	self logRetrying: exception.
	self close.
	(Delay forSeconds: self retryDelay) wait.
! !

!ZnClient methodsFor: 'operations' stamp: ''!
head
	"Execute a HTTP HEAD on the request set up and return nil."

	^ self 
		method: #HEAD; 
		execute! !

!ZnClient methodsFor: 'convenience' stamp: ''!
head: url
	"Execute an HTTP HEAD request on url and return nil."

	^ self 
		url: url; 
		head! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
headerAddAll: keyedCollection
	"Add all key/value pairs of keyedCollection as headers.
	Note that #addAllMulti: is used"
	
	request headers 
		addAll: keyedCollection ! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
headerAt: key add: value
	"Add value to key in the HTTP header of the current request.
	This is for multi-values headers."
	
	request headers 
		at: key add: value! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
headerAt: key put: value
	"Set key equals value in the HTTP header of the current request"
	
	request headers 
		at: key put: value   ! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
host: hostName
	"Set the host of the current request to hostName."
	
	self url: (ZnUrl new host: hostName; yourself)! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
http
	"Set the scheme of the current request to HTTP."
	
	self url: (ZnUrl new scheme: #http; yourself)! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
https
	"Set the scheme of the current request to HTTPS."
	
	self url: (ZnUrl new scheme: #https; yourself)! !

!ZnClient methodsFor: 'options' stamp: ''!
ifFail: block
	"Set the block that I will be execute when the request fails.
	The exception of the failure will be passed as argument."
	
	self optionAt: #ifFailBlock put: block! !

!ZnClient methodsFor: 'options' stamp: ''!
ifFailBlock
	"Return the block that I should execute if the request fails.
	Defaults to #pass-ing the exception up. 
	Accepts an exception as argument."
	
	^ self optionAt: #ifFailBlock ifAbsentPut: [ [ :exception | exception pass ] ]! !

!ZnClient methodsFor: 'initialization' stamp: ''!
initialize
	request := ZnRequest empty.
	self loggingOn ! !

!ZnClient methodsFor: 'initialization' stamp: ''!
interactivePolicy
	self 
		beOneShot;
		ifFail: [ :exception | 
			self inform: 'HTTP call failed with ', exception printString.
			nil ]! !

!ZnClient methodsFor: 'testing' stamp: ''!
isBadRequest
	"Return true if the last executed HTTP request returned an HTTP bad request response code."
	
	^ response 
		ifNil: [ self error: 'Request not yet executed' ]
		ifNotNil: [ response isBadRequest ]! !

!ZnClient methodsFor: 'testing' stamp: ''!
isConnected
	"Return true if I hold an open HTTP connection."

	^ connection notNil and: [ connection isConnected ]! !

!ZnClient methodsFor: 'testing' stamp: ''!
isContentTypeAcceptable
	"Return true if the contentType of the response of the last executed 
	HTTP request matches the accept header of the request."
	
	^ response 
		ifNil: [ self error: 'Request not yet executed' ]
		ifNotNil: [ 
			response hasEntity not 
				or: [ response contentType matches: request accept asZnMimeType ] ]! !

!ZnClient methodsFor: 'testing' stamp: ''!
isCreated
	"Return true it the last executed HTTP request returned an HTTP created response code."
	
	^ response 
		ifNil: [ self error: 'Request not yet executed' ]
		ifNotNil: [ response isCreated ]! !

!ZnClient methodsFor: 'testing' stamp: ''!
isNotFound
	"Return true it the last executed HTTP request returned an HTTP not found response code."
	
	^ response 
		ifNil: [ self error: 'Request not yet executed' ]
		ifNotNil: [ response isNotFound ]! !

!ZnClient methodsFor: 'testing' stamp: ''!
isNotModified
	"Return true it the last executed HTTP request returned an HTTP not modified response code."
	
	^ response 
		ifNil: [ self error: 'Request not yet executed' ]
		ifNotNil: [ response isNotModified ]! !

!ZnClient methodsFor: 'options' stamp: ''!
isOneShot
	"Return if I use connections only for a single request/response cycle.
	By default, I will keep HTTP connections open according to the HTTP 1.1 spec.
	When oneShot is true, I will add a Connection: close header to requests."

	^ self optionAt: #oneShot ifAbsent: [ false ]! !

!ZnClient methodsFor: 'testing' stamp: ''!
isRedirect
	"Return true it the last executed HTTP request returned an HTTP redirect response code."
	
	^ response 
		ifNil: [ self error: 'Request not yet executed' ]
		ifNotNil: [ response isRedirect ]! !

!ZnClient methodsFor: 'testing' stamp: ''!
isSuccess 
	"Return true it the last executed HTTP request returned an HTTP success response code."
	
	^ response 
		ifNil: [ self error: 'Request not yet executed' ]
		ifNotNil: [ response isSuccess ]! !

!ZnClient methodsFor: 'private logging' stamp: ''!
logClientTransactionRequestDuration: requestDuration responseDuration: responseDuration
	| logEventClass |
	logLevel < 1 ifTrue: [ ^ nil ].
	logEventClass := logLevel = 1
		ifTrue: [ ZnSimplifiedClientTransactionEvent ]
		ifFalse: [ ZnClientTransactionEvent ].
	^ (self newLogEvent: logEventClass)
		request: request;
		requestDuration: requestDuration;
		response: response; 
		responseDuration: responseDuration;
		emit;
		yourself! !

!ZnClient methodsFor: 'private logging' stamp: ''!
logConnectionClosedAddress: address port: port
	logLevel < 3 ifTrue: [ ^ nil ].
	^ (self newLogEvent: ZnClientConnectionClosedEvent)
		address: address;
		port: port;
		emit! !

!ZnClient methodsFor: 'private logging' stamp: ''!
logConnectionEstablishedTo: url started: initialMilliseconds
	| event |
	logLevel < 3 ifTrue: [ ^ nil ].
	(event := self newLogEvent: ZnConnectionEstablishedEvent)
		hostname: url host;
		port: url portOrDefault;
		address: ([ connection socket remoteAddress ] on: Error do: [ nil ]);
		duration: (Time millisecondsSince: initialMilliseconds).
	(ZnNetworkingUtils shouldProxyUrl: url)
		ifTrue: [ event proxy: ZnNetworkingUtils proxyUrl ].
	event emit! !

!ZnClient methodsFor: 'private logging' stamp: ''!
logIgnoringExceptionOnReusedConnection: exception
	logLevel < 3 ifTrue: [ ^ nil ].
	^ (self newLogEvent: ZnClientIgnoringExceptionOnConnectionReuseEvent)
		exception: exception;
		emit! !

!ZnClient methodsFor: 'logging' stamp: ''!
logLevel: integer
	"Set the log level to integer.
	0 - no logging
	1 - simplified transaction logging
	2 - detailed transaction logging
	3 - log everything"
	
	logLevel := integer! !

!ZnClient methodsFor: 'private logging' stamp: ''!
logRedirect: url
	^ (self newLogEvent: ZnClientFollowingRedirectEvent)
		target: url;
		emit! !

!ZnClient methodsFor: 'private logging' stamp: ''!
logRequestWrittenStarted: initialMilliseconds
	^ logLevel < 3
		ifTrue: [ Time millisecondsSince: initialMilliseconds ]
		ifFalse: [ 
			(self newLogEvent: ZnRequestWrittenEvent)
				request: request;
				duration: (Time millisecondsSince: initialMilliseconds);
				emit;
				duration ]! !

!ZnClient methodsFor: 'private logging' stamp: ''!
logResponseReadStarted: initialMilliseconds
	^ logLevel < 3
		ifTrue: [ Time millisecondsSince: initialMilliseconds ]
		ifFalse: [ 
			(self newLogEvent: ZnResponseReadEvent)
				response: response;
				duration: (Time millisecondsSince: initialMilliseconds);
				emit;
				duration ]! !

!ZnClient methodsFor: 'private logging' stamp: ''!
logRetrying: exception
	logLevel < 3 ifTrue: [ ^ nil ].
	^ (self newLogEvent: ZnClientRetryingEvent)
		exception: exception;
		emit! !

!ZnClient methodsFor: 'logging' stamp: ''!
logToTranscript
	"Setup logging to the Transcript.
	If logging was completely off, turn it on"

	logLevel isZero ifTrue: [ self loggingOn ].
	ZnLogEvent logToTranscript ! !

!ZnClient methodsFor: 'logging' stamp: ''!
loggingOff
	"Turn logging completely off - do not generate log events."
	
	self logLevel: 0! !

!ZnClient methodsFor: 'logging' stamp: ''!
loggingOn
	"Turn full logging on - generate all log events"
	
	self logLevel: 3! !

!ZnClient methodsFor: 'options' stamp: ''!
maxNumberOfRedirects
	"Return the maximum number of HTTP redirect that will be followed.
	Note that when the #followsRedirect option is false, 
	no redirects will ever be followed regardless of the value of this option.
	Defaults to 3. A ZnTooManyRedirects will be signalled when the count reaches zero."

	^ self optionAt: #maxNumberOfRedirects ifAbsent: [ 3 ]! !

!ZnClient methodsFor: 'options' stamp: ''!
maxNumberOfRedirects: count
	"Set the maximum number of HTTP redirect that will be followed to count.
	Note that when the #followsRedirect option is false, 
	no redirects will ever be followed regardless of the value of this option.
	A ZnTooManyRedirects will be signalled when the count reaches zero."

	self optionAt: #maxNumberOfRedirects put: count! !

!ZnClient methodsFor: 'options' stamp: ''!
maximumEntitySize
	"Return the maximum entity size in bytes that I will read from a stream before signalling ZnEntityTooLarge"
	
	^ self optionAt: #maximumEntitySize ifAbsent: [ ZnMaximumEntitySize value ]! !

!ZnClient methodsFor: 'options' stamp: ''!
maximumEntitySize: integer
	"Set the maximum entity size in bytes that I will read from a stream before signalling ZnEntityTooLarge"
	
	^ self optionAt: #maximumEntitySize put: integer! !

!ZnClient methodsFor: 'options' stamp: ''!
maximumNumberOfDictionaryEntries
	"Return the maximum number of entries allowed in ZnMutliValueDictionaries before signalling ZnTooManyDictionaryEntries. This protects us from overflow attacks."
		
	^ self
		optionAt: #maximumNumberOfDictionaryEntries
		ifAbsent: [ ZnMaximumNumberOfDictionaryEntries value ]! !

!ZnClient methodsFor: 'options' stamp: ''!
maximumNumberOfDictionaryEntries: anInteger
	"Set the maximum number of entries allowed in ZnMutliValueDictionaries before signalling ZnTooManyDictionaryEntries. This protects us from overflow attacks."
	
	^ self
		optionAt: #maximumNumberOfDictionaryEntries
		put: anInteger! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
method: httpMethod
	"Set the HTTP method of the current request to httpMethod.
	An uppercase symbol such as #GET is expected.
	Note that for all methods in #autoResetEntityMethods #resetEntity is called automatically,
	since you don't want the entity of a previous request to be sent again"

	(self autoResetEntityMethods includes: httpMethod)
		ifTrue: [ self resetEntity ].
	request method: httpMethod! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
multiPartFormDataEntity
	"Initialize the current request to have an multipart/form-data entity.
	Return the existing entity if there is one. 
	Fail if another type was already set."

	request hasEntity
		ifTrue: [
			(request entity contentType matches: ZnMimeType multiPartFormData)
				ifFalse: [ self error: 'An incompatible entity was already set for the request' ] ]
		ifFalse: [
			request entity: ZnMultiPartFormDataEntity new ].
	^ request entity! !

!ZnClient methodsFor: 'private' stamp: ''!
newConnectionTo: url
	| initialMilliseconds |
	self signalHTTPProgress: 'Connecting to ',  url authority.
	initialMilliseconds := Time millisecondClockValue.
	(connection notNil and: [ connection isConnected ]) 
		ifTrue: [ connection close ].
	connection := ZnNetworkingUtils socketStreamToUrl: url.
	url hasSecureScheme
		ifTrue: [ self setupTLSTo: url ].
	self logConnectionEstablishedTo: url started: initialMilliseconds! !

!ZnClient methodsFor: 'private logging' stamp: ''!
newLogEvent: logEventClass
	^ logEventClass new
		clientId: self clientId;
		yourself! !

!ZnClient methodsFor: 'options' stamp: ''!
numberOfRetries
	"Return how many times I will retry a failed request.
	Defaults to 0."
	
	^ self optionAt: #numberOfRetries ifAbsent: [ 0 ]! !

!ZnClient methodsFor: 'options' stamp: ''!
numberOfRetries: count
	"Set the number of times I will retry a failed request to count."
	
	self optionAt: #numberOfRetries put: count! !

!ZnClient methodsFor: 'accessing' stamp: ''!
optionAt: key ifAbsent: block
	"Return my option/settings stored under key.
	Execute block if I have no such option/setting.
	This is a generic interface, see my options protocol for specific usages."

	options ifNil: [ ^ block value ].
	^ options at: key ifAbsent: block! !

!ZnClient methodsFor: 'accessing' stamp: ''!
optionAt: key ifAbsentPut: block
	"Return my option/settings stored under key.
	If I have no such option/setting, store the result of evaluating block as new value and return it.
	This is a generic interface, see my options protocol for specific usages."

	options ifNil: [ options := Dictionary new ].
	^ options at: key ifAbsentPut: block! !

!ZnClient methodsFor: 'accessing' stamp: ''!
optionAt: key put: value
	"Set my option/setting identified by key to be value.
	This is a generic interface, see my options protocol for specific usages."
	
	options ifNil: [ options := Dictionary new ].
	options at: key put: value! !

!ZnClient methodsFor: 'operations' stamp: ''!
options
	"Execute an HTTP OPTIONS on the request set up and return the response #contents."
	
	^ self 
		method: #OPTIONS; 
		execute! !

!ZnClient methodsFor: 'convenience' stamp: ''!
options: url
	"Execute an HTTP OPTIONS request on url and return the response #contents."

	^ self 
		url: url; 
		options! !

!ZnClient methodsFor: 'operations' stamp: ''!
patch
	"Execute a HTTP PATCH on the request set up and return the response #contents."

	^ self 
		method: #PATCH; 
		execute! !

!ZnClient methodsFor: 'convenience' stamp: ''!
patch: url contents: object
	"Execute an HTTP PATCH request on url with object as request #contents:
	and return the response #contents."

	^ self 
		url: url; 
		contents: object; 
		patch! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
path: pathString
	"Set the path of the current request to pathString.
	The argument will be merged with of the current/previous request."
	
	self url: pathString! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
port: portNumber
	"Set the port of the current request to portNumber."
	
	self url: (ZnUrl new port: portNumber; yourself)! !

!ZnClient methodsFor: 'operations' stamp: ''!
post
	"Execute a HTTP POST on the request set up and return the response #contents."

	^ self 
		method: #POST; 
		execute! !

!ZnClient methodsFor: 'convenience' stamp: ''!
post: url contents: object
	"Execute an HTTP POST request on url with object as request #contents:
	and return the response #contents."

	^ self 
		url: url; 
		contents: object; 
		post! !

!ZnClient methodsFor: 'private protocol' stamp: ''!
prepareNextRequest
	"The request could be modified destructively in the next run,
	but the request can also be part of a log event, hence copy it"
	request := request copy	
! !

!ZnClient methodsFor: 'private protocol' stamp: ''!
prepareRedirect
	"We implement the common behavior that POST/PUT redirects turn into GET request
	and that the body data is dropped, except for a 307 or 308 where this is explicitly forbidden.
	Note the cookie handling too."

	| targetUrl |
	targetUrl := self redirectUrl.
	self logRedirect: targetUrl.
	self receiveCookies.
	self prepareNextRequest.
	(#(307 308) includes: response code)
		ifFalse: [  
			(#(POST PUT) includes: request method)
				ifTrue: [ self method: #GET ].
			request hasEntity
				ifTrue: [ request resetEntity: nil ] ].
	self url: targetUrl.
	self sendCookies! !

!ZnClient methodsFor: 'private protocol' stamp: ''!
prepareRequest
	self isOneShot 
		ifTrue: [ request setConnectionClose ].
	request hasEntity
		ifTrue: [ 
			"This is necessary because entities are often modified after first being added empty,
			and a wrong content length is cached at the request level"
			request resetEntity: request entity ].
	self sendCookies.
	self prepareRequestHook 
		ifNotNil: [ :hook | hook value: request ]! !

!ZnClient methodsFor: 'options' stamp: ''!
prepareRequest: hook
	"Set the current request preparation hook, an object conforming 
	to the #value: protocol that gets the final chance to change a request 
	right before it gets executed. Defaults to nil.
	Typically used to sign requests. See also #prepareRequestHook "

	^ self optionAt: #prepareRequestHook put: hook! !

!ZnClient methodsFor: 'options' stamp: ''!
prepareRequestHook
	"Return the current request preparation hook, an object conforming 
	to the #value: protocol that gets the final chance to change a request 
	right before it gets executed. Defaults to nil.
	Typically used to sign requests. See also #prepareRequest: "

	^ self optionAt: #prepareRequestHook ifAbsent: [ nil ]! !

!ZnClient methodsFor: 'operations' stamp: ''!
put
	"Execute a HTTP PUT on the request set up and return the response #contents."

	^ self 
		method: #PUT; 
		execute! !

!ZnClient methodsFor: 'convenience' stamp: ''!
put: url contents: object
	"Execute an HTTP PUT request on url with object as request #contents:
	and return the response #contents."

	^ self 
		url: url; 
		contents: object; 
		put! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
queryAddAll: keyedCollection
	"Add all key -> value elements of keyedCollection 
	as HTTP query parameters to the current request.
	Note that #queryAt:add: is used to preserve duplicate keys."

	keyedCollection keysAndValuesDo: [ :key :value |
		self queryAt: key add: value ]! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
queryAt: key add: value 
	"Modify the receiver by adding a query variable key=value.
	If key already exists, value is added to it."

	request url 
		ifNil: [ request url: (ZnUrl new queryAt: key add: value; yourself) ]
		ifNotNil: [ request url queryAt: key add: value ]  ! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
queryAt: key put: value
	"Add key equals value as an HTTP query parameter to the current request.
	If the key already exists, its value is overwritten."
	
	request url 
		ifNil: [ request url: (ZnUrl new queryAt: key put: value; yourself) ]
		ifNotNil: [ request url queryAt: key put: value ]  ! !

!ZnClient methodsFor: 'private protocol' stamp: ''!
readResponse
	self signalHTTPProgress: 'Reading response'.
	self withMaximumNumberOfDictionaryEntriesDo: [
		self withDefaultCharacterEncoderDo: [
			response := (#(HEAD CONNECT) includes: request method)
				ifTrue: [ ZnResponse readHeaderFrom: connection ]
				ifFalse: [ 
					self streaming
						ifTrue: [ ZnResponse readStreamingFrom: connection ]
						ifFalse: [ 
							self withMaximumEntitySizeDo: [ self reader value: connection ] ] ] ] ].
	lastUsed := Time totalSeconds! !

!ZnClient methodsFor: 'options' stamp: ''!
reader
	"Return a block that when given a stream reads a response from it."
	
	^ self optionAt: #reader ifAbsentPut: [ [ :stream | ZnResponse readFrom: stream ] ]! !

!ZnClient methodsFor: 'options' stamp: ''!
reader: block
	"Customize how responses are read from a stream, see #reader, see #beBinary"
	
	self optionAt: #reader put: block! !

!ZnClient methodsFor: 'private' stamp: ''!
receiveCookies
	response cookies do: [ :each |
		each defaultDomain: request url host.
		self session cookieJar add: each ]! !

!ZnClient methodsFor: 'private protocol' stamp: ''!
redirectUrl
	"We support relative locations although they are not the standard (http://en.wikipedia.org/wiki/HTTP_location)"

	| redirectUrl |
	redirectUrl := response location asZnUrl.
	^ redirectUrl isAbsolute
		ifTrue: [ redirectUrl ]
		ifFalse: [ redirectUrl inContextOf: request url ]! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
request
	"Return the current request that I am building,
	or the previous request that I executed."
	
	^ request ! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
request: aRequest
	"Replace my current request with aRequest.
	This is a low level operation only needed in special cases."
	
	self validateConnectionForHost: aRequest url host port: aRequest url port.
	request := aRequest! !

!ZnClient methodsFor: 'accessing' stamp: ''!
resetCookies
	session ifNotNil: [ session resetCookies ]! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
resetEntity
	"Make sure that the current request does not hold an entity from a previous request.
	This is sometimes necessary after a POST or PUT. This method will be called automatically
	by #method: for certain requests."
	
	(request notNil and: [ request hasEntity ])
		ifTrue: [ request resetEntity: nil ]! !

!ZnClient methodsFor: 'accessing response' stamp: ''!
response 
	"Answer the last response, assuming the current request was executed."

	^ response ! !

!ZnClient methodsFor: 'options' stamp: ''!
retryDelay
	"Return the number of seconds I will delay retrying a failed request,
	provided I do retry requests. Defaults to 1 seconds."
	
	^ self optionAt: #retryDelay ifAbsent: [ 1 ]! !

!ZnClient methodsFor: 'options' stamp: ''!
retryDelay: seconds
	"Set the delay before I will delay retrying a failed request to seconds,
	provided I do retry requests."

	self optionAt: #retryDelay put: seconds! !

!ZnClient methodsFor: 'private' stamp: ''!
retryExceptionSet
	^ NetworkError, ZnParseError, ZnCharacterEncodingError, ZnUnknownScheme, ZnPortNotANumber! !

!ZnClient methodsFor: 'private' stamp: ''!
sendCookies
	| cookies cookiesString |
	session ifNil: [ ^ self ].
	cookies := (session cookieJar cookiesForUrl: request url).
	cookies isEmpty ifTrue: [ ^ self ]. 
	cookiesString := String streamContents: [ :stream |
		cookies
			do: [ :each | each writeNameAndValueOn: stream ]
			separatedBy: [ stream nextPutAll: '; ' ] ].
	request setCookie: cookiesString! !

!ZnClient methodsFor: 'accessing' stamp: ''!
session
	"Return the current HTTP session object that I am using.
	This session object holds cookies among other things."

	^ session ifNil: [ session := ZnUserAgentSession new ]! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
setAcceptEncodingGzip
	"Give the server the hint that we accept gzip content encoding.
	When the server complies, important bandwidth reductions are possible."

	self request setAcceptEncodingGzip! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
setBasicAuthenticationUsername: username password: password
	"Set the username/password for basic authenticationfor the current request."
	
	self request setBasicAuthenticationUsername: username password: password
! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
setIfModifiedSince: reference
	"Tell the server that we are only interested in the currently requested URI
	if it was modified since reference, a DateAndTime.
	If the URI was not modified since reference, the response will be #isNotModified."

	self request setIfModifiedSince: reference! !

!ZnClient methodsFor: 'private' stamp: 'SvenVanCaekenberghe 3/22/2018 14:49'!
setupTLSTo: url
	(ZnNetworkingUtils shouldProxyUrl: url)
		ifTrue: [ | originalRequest |
			"http://www.ietf.org/rfc/rfc2817.txt (section 5)"
			"https://en.wikipedia.org/wiki/HTTP_tunnel#HTTP_CONNECT_tunneling"
			originalRequest := request copy.
			request entity: nil. "to prevent #resetEntity from being called, resulting in #close of the shared entity"
			self method: #CONNECT; writeRequest; readResponse.
			response isSuccess ifFalse: [ self error: 'Failed to CONNECT to proxy for TLS/SSL' ].
			connection := ZnNetworkingUtils secureSocketStreamOn: connection socket.
			request := originalRequest ].
	
	connection sslSession certificateName: self certificate.
	
	[ connection sslSession serverName: url host ] 
		on: PrimitiveFailed 
		do: [ :error | "log it?" ].
		
	connection connect! !

!ZnClient methodsFor: 'private' stamp: ''!
signalHTTPProgress: message
	self signalProgress
		ifTrue: [ HTTPProgress signal: message ]! !

!ZnClient methodsFor: 'options' stamp: ''!
signalProgress
	"Return true if I signal HTTPProgress notifications during execution."
	
	^ self optionAt: #signalProgress ifAbsent: [ false ]! !

!ZnClient methodsFor: 'options' stamp: ''!
signalProgress: boolean
	"Set whether I will signal HTTPProgress notifications during execution to boolean."
	
	self optionAt: #signalProgress put: boolean! !

!ZnClient methodsFor: 'options' stamp: ''!
streaming
	"Return whether I will return streaming response objects (see ZnStreamingEntity).
	If true, my users are responsible for actually reading the incoming response entity data stream."
	
	^ self optionAt: #streaming ifAbsent: [ false ]! !

!ZnClient methodsFor: 'options' stamp: ''!
streaming: boolean
	"Set whether I will return streaming response objects (see ZnStreamingEntity) to boolean.
	If true, my users are responsible for actually reading the incoming response entity data stream.
	I cannot do streaming when #isOneShot is true because the stream would be closed too soon."

	(boolean and: [ self isOneShot ])
		ifTrue: [ self error: 'I cannot do streaming when #isOneShot is true.' ].
	self optionAt: #streaming put: boolean! !

!ZnClient methodsFor: 'initialization' stamp: ''!
systemPolicy
	self 
		enforceHttpSuccess: true;
		enforceAcceptContentType: true;
		numberOfRetries: 2! !

!ZnClient methodsFor: 'options' stamp: ''!
timeout
	"Return the timeout in seconds for network socket stream operations
	like connecting, reading and writing. Defaults to a global setting."
	
	^ self optionAt: #timeout ifAbsent: [ ZnNetworkingUtils defaultSocketStreamTimeout ]! !

!ZnClient methodsFor: 'options' stamp: ''!
timeout: seconds
	"Set the timeout for network socket stream operations
	like connecting, reading and writing to seconds."

	self optionAt: #timeout put: seconds! !

!ZnClient methodsFor: 'convenience' stamp: 'SvenVanCaekenberghe 11/1/2018 23:28'!
uploadEntityFrom: path
	"Set up a ZnStreamingEntity from the binary contents of the file at path.
	The path extension is used to guess the content-type,
	can be overridden using #contentType: if necessary.
	Use the file's length to set the content-length.
	You can then use #put or #post to actually do an upload."
	
	| stream type length |
	stream := ZnFileSystemUtils binaryReadStreamFor: path.
	type := ZnMimeType forFilenameExtension: (ZnFileSystemUtils extensionFor: path).
	length := ZnFileSystemUtils fileSizeFor: path.
	self entity: (ZnStreamingEntity readFrom: stream usingType: type andLength: length)! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
url: urlObject
	"Modify the URL of the current request object to urlObject.
	This is done using ZnUrl>>#inContextOf: so that either scheme, 
	host, port or path are set and merged.
	Default the URL using ZnUrl>>#asZnUrlWithDefaults to fall back
	to http or the default http/https ports."
	
	| url |
	url := urlObject asZnUrl.
	request url
		ifNotNil: [ url := url inContextOf: request url ].  
	url := url asZnUrlWithDefaults.  
	self validateConnectionForHost: url host port: url port.   
	request url: url.
	url hasUsername & url hasPassword
		ifTrue: [ self setBasicAuthenticationUsername: url username password: url password ]! !

!ZnClient methodsFor: 'accessing request' stamp: ''!
username: username password: password
	"Set the username/password for basic authenticationfor the current request."
	
	self setBasicAuthenticationUsername: username password: password
! !

!ZnClient methodsFor: 'private' stamp: ''!
validateConnectionForHost: host port: port
	(connection isNil or: [ request url isNil ]) 
		ifTrue: [ ^ self ].
	(request url host = host and: [ request url port = port ])
		ifFalse: [ self close ]  ! !

!ZnClient methodsFor: 'private' stamp: ''!
withDefaultCharacterEncoderDo: block
	^ self defaultEncoder = ZnDefaultCharacterEncoder value
			ifTrue: block
			ifFalse: [ 
				ZnDefaultCharacterEncoder 
					value: self defaultEncoder  
					during: block ]! !

!ZnClient methodsFor: 'private' stamp: ''!
withMaximumEntitySizeDo: block
	^ self maximumEntitySize = ZnMaximumEntitySize value
			ifTrue: block
			ifFalse: [ 
				ZnMaximumEntitySize 
					value: self maximumEntitySize  
					during: block ]! !

!ZnClient methodsFor: 'private' stamp: ''!
withMaximumNumberOfDictionaryEntriesDo: block
	^ self maximumNumberOfDictionaryEntries = ZnMaximumNumberOfDictionaryEntries value
			ifTrue: block
			ifFalse: [ 
				ZnMaximumNumberOfDictionaryEntries 
					value: self maximumNumberOfDictionaryEntries 
					during: block ]! !

!ZnClient methodsFor: 'private' stamp: ''!
withProgressDo: block
	^ ZnSignalProgress value
		ifNil: [
			ZnSignalProgress 
				value: self signalProgress 
				during: [ ^ block value ] ]
		ifNotNil: [ 
			block value ]! !

!ZnClient methodsFor: 'private' stamp: ''!
withTimeoutDo: block
	^ ZnConnectionTimeout value
		ifNil: [
			ZnConnectionTimeout 
				value: self timeout 
				during: [ ^ block value ] ]
		ifNotNil: [ 
			block value ]! !

!ZnClient methodsFor: 'private protocol' stamp: ''!
writeRequest
	self signalHTTPProgress: 'Writing request'.
	self withDefaultCharacterEncoderDo: [ request writeOn: connection ].
	connection flush! !

Object subclass: #ZnCommonLogFormat
	instanceVariableNames: 'combined'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnCommonLogFormat commentStamp: '' prior: 0!
I am ZnCommonLogFormat, I can output ZnServerTransactionEvent objects using Apache Common Log Format (CLF).

https://en.wikipedia.org/wiki/Common_Log_Format
https://httpd.apache.org/docs/trunk/logs.html#common

| formatter |
formatter := ZnCommonLogFormat new.
ZnLogEvent announcer 
  when: ZnServerTransactionEvent 
  do: [ :event | 
    formatter format: event on: Transcript.
    Transcript cr; endEntry ].
!

!ZnCommonLogFormat methodsFor: 'private' stamp: ''!
clientIPAddressFrom: event
	^ event request headers 
		at: ZnConstants remoteAddressHeader 
		ifAbsent: [ '-' ]! !

!ZnCommonLogFormat methodsFor: 'accessing' stamp: ''!
combined
	^ combined! !

!ZnCommonLogFormat methodsFor: 'accessing' stamp: ''!
combined: boolean
	"If boolean is true (the default), use the 'Combined' format, 
	adding Referer and User-Agent fields."
	
	combined := boolean! !

!ZnCommonLogFormat methodsFor: 'public' stamp: ''!
format: event on: stream
	"Output a ZnServerTransactionEvent on a text stream in Common Log Format"
	
	stream << (self clientIPAddressFrom: event); << ' - '.
	stream << (self userFrom: event); space.
	self formatTimestamp: event timestamp on: stream.
	self formatRequest: event on: stream.
	self formatResponse: event on: stream. 
	combined ifTrue: [ self formatCombinedFields: event request on: stream ]! !

!ZnCommonLogFormat methodsFor: 'private' stamp: ''!
formatCombinedFields: request on: stream
	| referer userAgent |
	referer := request headers at: 'Referer' ifAbsent: [ '-' ].
	userAgent := request headers at: 'User-Agent' ifAbsent: [ '-' ].
	stream space.
	stream nextPut: $"; << referer; nextPut: $".
	stream space.
	stream nextPut: $"; << userAgent; nextPut: $"
! !

!ZnCommonLogFormat methodsFor: 'private' stamp: ''!
formatRequest: event on: stream
	| requestLine |
	stream space.
	requestLine := event request requestLine.
	stream << $".
	requestLine printMethodAndUriOn: stream.
	stream space; << requestLine version.
	stream << $"! !

!ZnCommonLogFormat methodsFor: 'private' stamp: ''!
formatResponse: event on: stream
	| size |
	stream space.
	stream print: event response code.
	stream space.
	size := event response hasEntity 
		ifTrue: [ event response contentLength ] 
		ifFalse: [ 0 ].
	stream print: size
! !

!ZnCommonLogFormat methodsFor: 'private' stamp: ''!
formatTimestamp: timestamp on: stream
	stream nextPut: $[.
	timestamp dayOfMonth printOn: stream base: 10 length: 2 padded: true.
	stream nextPut: $/.
	stream << timestamp monthAbbreviation.
	stream nextPut: $/.
	timestamp year printOn: stream.
	stream << $:.
	timestamp printHMSOn: stream.
	stream space.
	stream nextPut: (timestamp offset positive ifTrue: [ $+ ] ifFalse: [ $- ]).
	timestamp offset hours abs printOn: stream base: 10 length: 2 padded: true.
	stream nextPut: $:.
	timestamp offset minutes abs printOn: stream base: 10 length: 2 padded: true.
	stream nextPut: $]! !

!ZnCommonLogFormat methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	combined := true.! !

!ZnCommonLogFormat methodsFor: 'private' stamp: ''!
userFrom: event
	^ [ event request basicAuthentication first ]
		on: Error
		do: [ '-' ]! !

Object subclass: #ZnConstants
	instanceVariableNames: ''
	classVariableNames: 'DefaultMaximumEntitySize HTTPStatusCodes'
	poolDictionaries: ''
	category: 'Zinc-HTTP-Support'!

!ZnConstants commentStamp: '' prior: 0!
I am ZnConstants, holding various constants that we use.

Part of Zinc HTTP Components.!

!ZnConstants class methodsFor: 'accessing' stamp: ''!
defaultHTTPVersion
	^ 'HTTP/1.1'! !

!ZnConstants class methodsFor: 'accessing' stamp: ''!
defaultServerString 
	^ self frameworkNameAndVersion , ' (' , self systemVersion , ')'! !

!ZnConstants class methodsFor: 'accessing' stamp: ''!
defaultUserAgent
	^ self frameworkNameAndVersion , ' (' , self systemVersion , ')'! !

!ZnConstants class methodsFor: 'accessing' stamp: ''!
faviconBytes
	^ #(0 0 1 0 1 0 16 16 16 0 0 0 0 0 40 1 0 0 22 0 0 0 40 0 0 0 16 0 0 0 32 0 0 0 1 0 4 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 255 255 255 0 0 189 255 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1 17 17 17 1 16 1 16 1 17 17 17 1 16 1 16 1 16 0 0 1 16 1 16 0 17 0 0 1 16 1 16 0 1 16 0 1 16 1 16 0 0 17 0 1 17 17 16 0 0 1 16 1 1 17 0 0 0 0 17 0 0 0 0 1 17 17 17 0 0 0 0 1 17 17 17 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0) asByteArray! !

!ZnConstants class methodsFor: 'accessing' stamp: ''!
frameworkLicense
	"http://en.wikipedia.org/wiki/MIT_License"
	
	^ 'MIT License'! !

!ZnConstants class methodsFor: 'accessing' stamp: ''!
frameworkMCVersion
	MCWorkingCopy managersForClass: self class do: [ :each | ^ each ancestors first name ]! !

!ZnConstants class methodsFor: 'accessing' stamp: ''!
frameworkName
	^ 'Zinc HTTP Components'! !

!ZnConstants class methodsFor: 'accessing' stamp: ''!
frameworkNameAndVersion
	^ self frameworkName, ' ', self frameworkVersion ! !

!ZnConstants class methodsFor: 'accessing' stamp: ''!
frameworkVersion
	^ '1.0'! !

!ZnConstants class methodsFor: 'accessing' stamp: ''!
httpStatusCodes
	^ HTTPStatusCodes ! !

!ZnConstants class methodsFor: 'class initialization' stamp: ''!
initialize
	"Forced source code change for automatic initialization: 20170510"
	
	self initializeHTTPStatusCodes! !

ZnConstants initialize!

!ZnConstants class methodsFor: 'private' stamp: ''!
initializeHTTPStatusCodes 
	HTTPStatusCodes := Dictionary newFromPairs: self officialHTTPStatusCodesSpec ! !

!ZnConstants class methodsFor: 'accessing' stamp: ''!
knownHTTPMethods
	^ #(
		"Standard HTTP 1.1 Verbs"
		GET PUT POST DELETE HEAD 
		TRACE OPTIONS CONNECT PATCH
		"WebDAV verbs"
		PROPFIND PROPPATCH MKCOL COPY MOVE LOCK UNLOCK)! !

!ZnConstants class methodsFor: 'accessing' stamp: ''!
knownHTTPVersions
	^ #('HTTP/1.1' 'HTTP/1.0')! !

!ZnConstants class methodsFor: 'resource limits' stamp: ''!
maximumLineLength
	"Return the maximum line length to accept.
	Used by ZnLineReader and thus for reading request/status lines as well as headers.
	This helps to protect us from malicious content."
	
	^ 4096! !

!ZnConstants class methodsFor: 'accessing' stamp: ''!
officialHTTPStatusCodesSpec 
	"This is the official list from http://www.iana.org/assignments/http-status-codes (version 2017-04-14)"

	^ #(
		"1xx: Informational - Request received, continuing process"
		100 'Continue'
		101 'Switching Protocols'
		102 'Processing'
		"2xx: Success - The action was successfully received, understood, and accepted"
		200 'OK'
		201 'Created'
		202 'Accepted'
		203 'Non-Authoritative Information'
		204 'No Content'
		205 'Reset Content'
		206 'Partial Content'
		207 'Multi-Status'
		208 'Already Reported'
		226 'IM Used'
		"3xx: Redirection - Further action must be taken in order to complete the request"
		300 'Multiple Choices'
		301 'Moved Permanently'
		302 'Found'
		303 'See Other'
		304 'Not Modified'
		305 'Use Proxy'
		307 'Temporary Redirect'
		308 'Permanent Redirect'
		"4xx: Client Error - The request contains bad syntax or cannot be fulfilled"
		400 'Bad Request'
		401 'Unauthorized'
		402 'Payment Required'
		403 'Forbidden'
		404 'Not Found'
		405 'Method Not Allowed'
		406 'Not Acceptable'
		407 'Proxy Authentication Required'
		408 'Request Timeout'
		409 'Conflict'
		410 'Gone'
		411 'Length Required'
		412 'Precondition Failed'
		413 'Payload Too Large'
		414 'URI Too Long'
		415 'Unsupported Media Type'
		416 'Requested Range Not Satisfiable'
		417 'Expectation Failed'
		421 'Misdirected Request'
		422 'Unprocessable Entity'
		423 'Locked'
		424 'Failed Dependency'
		426 'Upgrade Required'
		428 'Precondition Required'
		429 'Too Many Requests'
		431 'Request Header Fields Too Large'
		"5xx: Server Error - The server failed to fulfill an apparently valid request"
		500 'Internal Server Error'
		501 'Not Implemented'
		502 'Bad Gateway'
		503 'Service Unavailable'
		504 'Gateway Timeout'
		505 'HTTP Version Not Supported'
		506 'Variant Also Negotiates (Experimental)'
		507 'Insufficient Storage'
		508 'Loop Detected'
		510 'Not Extended'
		511 'Network Authentication Required'
	)! !

!ZnConstants class methodsFor: 'accessing' stamp: ''!
remoteAddressHeader
	^ 'X-Zinc-Remote-Address'! !

!ZnConstants class methodsFor: 'accessing' stamp: ''!
systemVersion
	^ SystemVersion current type , '/' , SystemVersion current dottedMajorMinor ! !

Object subclass: #ZnCookie
	instanceVariableNames: 'attributes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Support'!

!ZnCookie commentStamp: 'SvenVanCaekenberghe 3/22/2018 16:55' prior: 0!
I represent an HTTP Cookie, a small piece of data sent by an HTTP web server to an HTTP user-agent client and by the client to the same server afterwards, to remember stateful information across indepedent request/responses.

References

https://en.wikipedia.org/wiki/HTTP_cookie
https://tools.ietf.org/html/rfc6265

Part of Zinc HTTP Components.!

!ZnCookie class methodsFor: 'instance creation' stamp: ''!
fromString: aString
	^ self new 
		fromString: aString; 
		yourself! !

!ZnCookie class methodsFor: 'instance creation' stamp: ''!
fromString: aString for: urlObject
	| cookie |
	cookie := self fromString: aString.
	"Ensures a domain value if none provided by the cookie"
	cookie defaultDomain: urlObject host.
	^ cookie! !

!ZnCookie class methodsFor: 'instance creation' stamp: ''!
name: nameString value: value
	^ self new
		name: nameString;
		value: value;
		yourself! !

!ZnCookie methodsFor: 'comparing' stamp: ''!
= aCookie
	(aCookie isKindOf: self class) ifFalse: [ ^ false ].
	^ (aCookie name = self name) & (aCookie path = self path) & (aCookie domain = self domain)
	! !

!ZnCookie methodsFor: 'initialize-release' stamp: ''!
defaultDomain: aDomain
	self domain ifNil: [ self domain: aDomain ]! !

!ZnCookie methodsFor: 'accessing' stamp: ''!
domain
	^ attributes at: 'domain' ifAbsent: [ nil ]! !

!ZnCookie methodsFor: 'accessing' stamp: ''!
domain: aDomain
	attributes at: 'domain' put: aDomain! !

!ZnCookie methodsFor: 'accessing' stamp: ''!
expires
	^ attributes at: 'expires' ifAbsent: [ nil ]! !

!ZnCookie methodsFor: 'accessing' stamp: ''!
expires: anExpiration
	attributes at: 'expires' put: anExpiration! !

!ZnCookie methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/22/2018 16:46'!
expiresTimeStamp
	"Return the DateAndTime when I will expire.
	Throw an exception when no expiration is set or the expiration cannot be parsed.
	See #isExpired"
	
	^ ZnUtils parseHttpDate: self expires! !

!ZnCookie methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/22/2018 16:36'!
expiresTimeStamp: dateAndTime
	self expires: (ZnUtils httpDate: dateAndTime)! !

!ZnCookie methodsFor: 'private' stamp: ''!
extraAttributes
	"Answer all attributes except name and value"
	^ attributes associationsSelect: [ :ea | (ea key = 'name' or: [ ea key = 'value' ]) not ]! !

!ZnCookie methodsFor: 'initialize-release' stamp: ''!
fromString: aString
	| tokens val data i |
	tokens := aString substrings: ';'.
	i := tokens first indexOf: $=.
	val := tokens first copyFrom: 1 to: i - 1.
	data := tokens first copyFrom: i + 1 to: tokens first size. 
	self name: val trimBoth; value: data trimBoth.
	2 to: tokens size do: [ :index | 
		self readAttribute: (tokens at: index) ]! !

!ZnCookie methodsFor: 'converting' stamp: ''!
fullString
	^ String streamContents: [ :stream | 
		self writeOn: stream ]! !

!ZnCookie methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/22/2018 16:45'!
hasAttribute: attributeName
	^ attributes includesKey: attributeName asLowercase ! !

!ZnCookie methodsFor: 'comparing' stamp: ''!
hash
	^ (self name hash bitXor: self path hash) bitXor: self domain hash! !

!ZnCookie methodsFor: 'accessing' stamp: ''!
httponly
	"This is not an RFC cookie attribute, but a supplement found in MS Internet Explorer and supported in PHP.
	 See http://msdn.microsoft.com/en-us/library/ms533046.aspx"
	^ attributes at: 'httponly' ifAbsent: [ false ]! !

!ZnCookie methodsFor: 'accessing' stamp: ''!
httponly: aBoolean
	attributes at: 'httponly' put: aBoolean! !

!ZnCookie methodsFor: 'initialization' stamp: ''!
initialize 
	attributes := Dictionary new.! !

!ZnCookie methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/22/2018 16:57'!
isExpired
	| expirationTimeStamp |
	(self hasAttribute: 'expire') ifFalse: [ ^ false ].
	[ expirationTimeStamp := self expiresTimeStamp ] on: Error do: [ ^ false ].
	"note that max-age (#maxage) is not used"
	^ expirationTimeStamp asUTC < DateAndTime now asUTC! !

!ZnCookie methodsFor: 'testing' stamp: ''!
isInDomain: urlObject
	| domain |
	domain := (self domain beginsWith: '.')
		ifTrue: [ self domain copyFrom: 2 to: self domain size ]
		ifFalse: [ self domain ].
	^ urlObject host includesSubstring: domain
	! !

!ZnCookie methodsFor: 'testing' stamp: ''!
isInPath: urlObject
	^ urlObject pathPrintString includesSubstring: self path
	! !

!ZnCookie methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/22/2018 16:37'!
maxage
	^ attributes at: 'max-age' ifAbsent: [ nil ]! !

!ZnCookie methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/22/2018 16:51'!
maxage: numberOfSeconds
	"To be able to use this to validate expiration, a time of creation has to be recorded, so right now, this is not really usable."
	
	^ attributes at: 'max-age' put: numberOfSeconds asString! !

!ZnCookie methodsFor: 'accessing' stamp: ''!
name
	^ attributes at: 'name'! !

!ZnCookie methodsFor: 'accessing' stamp: ''!
name: aName
	attributes at: 'name' put: aName! !

!ZnCookie methodsFor: 'converting' stamp: ''!
nameValueString
	^ String streamContents: [ :stream | 
		self writeNameAndValueOn: stream ]! !

!ZnCookie methodsFor: 'accessing' stamp: ''!
path
	^ attributes at: 'path' ifAbsent: [ '/' ]! !

!ZnCookie methodsFor: 'accessing' stamp: ''!
path: aPath
	attributes at: 'path' put: aPath! !

!ZnCookie methodsFor: 'printing' stamp: ''!
printOn: stream
	self writeOn: stream! !

!ZnCookie methodsFor: 'initialize-release' stamp: ''!
readAttribute: aToken
	| i key data |
	i := aToken indexOf: $=.
	i > 0
		ifTrue: [
			key := (aToken copyFrom: 1 to: i - 1) trimBoth asLowercase.
			data := (aToken copyFrom: i + 1 to: aToken size) trimBoth ]
		ifFalse: [ 
			key := (aToken copyFrom: 1 to: aToken size) trimBoth asLowercase.
			data := true ].
	attributes at: key put: data! !

!ZnCookie methodsFor: 'accessing' stamp: ''!
secure
	^ attributes at: 'secure' ifAbsent: [ false ]! !

!ZnCookie methodsFor: 'accessing' stamp: ''!
secure: aBoolean
	attributes at: 'secure' put: aBoolean! !

!ZnCookie methodsFor: 'accessing' stamp: ''!
value
	^ attributes at: 'value' ifAbsent: [ nil ]! !

!ZnCookie methodsFor: 'accessing' stamp: ''!
value: aValue
	attributes at: 'value' put: aValue! !

!ZnCookie methodsFor: 'printing' stamp: ''!
writeExtraAttributesOn: aStream
	self extraAttributes
		associationsDo: [ :ea | 
			aStream
				nextPutAll: '; ';
				nextPutAll: ea key.
			(ea value isKindOf: Boolean)
				ifFalse: [ 
					aStream
						nextPutAll: '=';
						nextPutAll: ea value ] ]! !

!ZnCookie methodsFor: 'printing' stamp: ''!
writeNameAndValueOn: aStream
	aStream
		nextPutAll: self name;
		nextPutAll: '=';
		nextPutAll: self value! !

!ZnCookie methodsFor: 'printing' stamp: ''!
writeOn: aStream
	self
		writeNameAndValueOn: aStream;
		writeExtraAttributesOn: aStream ! !

Object subclass: #ZnCookieJar
	instanceVariableNames: 'cookies'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Support'!

!ZnCookieJar commentStamp: '' prior: 0!
I represent a container for storing HTTP Cookies for a client session.

Part of Zinc HTTP Components.!

!ZnCookieJar methodsFor: 'accessing' stamp: ''!
add: aCookie
	(self removeIfExpired: aCookie) ifTrue: [ ^ nil ].
	self cookies do: [ :cookie |
		cookie = aCookie ifTrue: [ self cookies remove: cookie ] ].
	self cookies add: aCookie
	! !

!ZnCookieJar methodsFor: 'accessing' stamp: ''!
cookieAt: aName forUrl: urlObject
	(self cookiesForUrl: urlObject) do: [ :cookie | cookie name = aName ifTrue: [ ^ cookie ] ].
	^ nil! !

!ZnCookieJar methodsFor: 'accessing' stamp: ''!
cookies 
	^ cookies ifNil: [ cookies := Bag new ]! !

!ZnCookieJar methodsFor: 'accessing' stamp: ''!
cookiesForUrl: urlObject
	^ self cookies select: [ :cookie | 
		(cookie isExpired not) & (cookie isInDomain: urlObject) & (cookie isInPath: urlObject) ]! !

!ZnCookieJar methodsFor: 'accessing' stamp: ''!
remove: aCookie
	self cookies remove: aCookie ifAbsent: [ nil ]! !

!ZnCookieJar methodsFor: 'operations' stamp: ''!
removeAll
	cookies ifNotNil: [ cookies removeAll ]! !

!ZnCookieJar methodsFor: 'operations' stamp: ''!
removeIfExpired: aCookie
	^ aCookie isExpired 
		ifTrue: [ 
			self remove: aCookie. 
			true ]
		ifFalse: [
			false ]! !

Object subclass: #ZnCredential
	instanceVariableNames: 'domain realm'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Support'!

!ZnCredential commentStamp: '' prior: 0!
I am ZnCredential, an abstract class representing credentials in a domain and realm.

Part of Zinc HTTP Components.!

ZnCredential subclass: #ZnBasicCredential
	instanceVariableNames: 'username password'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Support'!

!ZnBasicCredential commentStamp: '' prior: 0!
I am ZnBasicCredential, a concrete subclass of ZnCredential.
I implement basic HTTP authentication.
I hold a username and password.

Part of Zinc HTTP Components.!

!ZnBasicCredential class methodsFor: 'accessing' stamp: ''!
authName
	^ 'Basic'! !

!ZnBasicCredential methodsFor: 'accessing' stamp: ''!
authTokenForUrl: urlObject method: method
	^ 'Basic ', (ZnUtils encodeBase64: (username, ':', password)) ! !

!ZnBasicCredential methodsFor: 'testing' stamp: ''!
isComplete 
	^ username notNil & password notNil ! !

!ZnBasicCredential methodsFor: 'accessing' stamp: ''!
password
	^ password! !

!ZnBasicCredential methodsFor: 'accessing' stamp: ''!
password: anObject
	password := anObject! !

!ZnBasicCredential methodsFor: 'accessing' stamp: ''!
username
	^ username! !

!ZnBasicCredential methodsFor: 'accessing' stamp: ''!
username: anObject
	username := anObject! !

ZnBasicCredential subclass: #ZnDigestCredential
	instanceVariableNames: 'nonce opaque'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Support'!

!ZnDigestCredential commentStamp: '' prior: 0!
I am ZnDigestCredential, a concrete subclass of ZnCredential and subclass of ZnBasicCredential.
I implement digest HTTP authentication.
I hold a nounce and opaque.

Part of Zinc HTTP Components.!

!ZnDigestCredential class methodsFor: 'accessing' stamp: ''!
authName
	^ 'Digest'! !

!ZnDigestCredential methodsFor: 'computation' stamp: ''!
a1
	^ ZnDigestAuthenticator md5Hash: self username, ':', self realm, ':', self password! !

!ZnDigestCredential methodsFor: 'computation' stamp: ''!
a2ForUrl: urlObject method: method
	"This doesn't support all qop options yet."
	^ ZnDigestAuthenticator md5Hash: method, ':', (self uriPathFrom: urlObject)! !

!ZnDigestCredential methodsFor: 'accessing' stamp: ''!
authTokenForUrl: urlObject method: method
	| result |
	result := WriteStream on: String new.
	result
		nextPutAll: 'Digest username="';
		nextPutAll: self username;
		nextPutAll: '", ';
		nextPutAll: 'nonce="';
		nextPutAll: self nonce;
		nextPutAll: '", ';
		nextPutAll: 'realm="';
		nextPutAll: self realm;
		nextPutAll: '", ';
		nextPutAll: 'uri="';
		nextPutAll: (self uriPathFrom: urlObject);
		nextPutAll: '", ';
		nextPutAll: 'response="';
		nextPutAll: (self responseForUrl: urlObject method: method);
		nextPutAll: '", '.
	self opaque isNil
		ifFalse: [ 
			result
				nextPutAll: 'opaque="';
				nextPutAll: self opaque;
				nextPutAll: '"' ].
	^ result contents! !

!ZnDigestCredential methodsFor: 'testing' stamp: ''!
isComplete 
	^ super isComplete & nonce notNil ! !

!ZnDigestCredential methodsFor: 'operations' stamp: ''!
linkToAuthRequest: aHeader
	| data |
	data := self parseAuthRequest: aHeader.
	self nonce: (data at: 'nonce').
	self realm: (data at: 'realm' ifAbsent: [ '' ])! !

!ZnDigestCredential methodsFor: 'accessing' stamp: ''!
nonce
	^ nonce! !

!ZnDigestCredential methodsFor: 'accessing' stamp: ''!
nonce: anObject
	nonce := anObject! !

!ZnDigestCredential methodsFor: 'accessing' stamp: ''!
opaque
	^ opaque! !

!ZnDigestCredential methodsFor: 'accessing' stamp: ''!
opaque: anObject
	opaque := anObject! !

!ZnDigestCredential methodsFor: 'operations' stamp: ''!
parseAuthRequest: aHeader
	^ ZnDigestAuthenticator parseAuthRequest: aHeader! !

!ZnDigestCredential methodsFor: 'computation' stamp: ''!
responseForUrl: urlObject method: method
	^ ZnDigestAuthenticator md5Hash: self a1, ':', self nonce, ':', (self a2ForUrl: urlObject method: method)! !

!ZnDigestCredential methodsFor: 'operations' stamp: ''!
setAuthorizationHeader: headers for: urlObject method: method
	headers at: 'Authorization' put: (self authTokenForUrl: urlObject method: method)! !

!ZnDigestCredential methodsFor: 'accessing' stamp: ''!
uriPathFrom: urlObject
	^ urlObject pathPrintString ifNil: [ '/' ]! !

!ZnCredential class methodsFor: 'accessing' stamp: ''!
authName
	self subclassResponsibility ! !

!ZnCredential class methodsFor: 'accessing class hierarchy' stamp: ''!
findHandlerFor: anAuthHeader
	| type i |
	i := anAuthHeader indexOf: Character space.
	type := anAuthHeader copyFrom: 1 to: i - 1.
	self allSubclassesDo: [ :subclass |
		subclass authName = type ifTrue: [ ^ subclass ] ].
	^ nil! !

!ZnCredential class methodsFor: 'accessing' stamp: ''!
findRealmFor: anAuthHeader
	(anAuthHeader substrings: ',') do: [ :token |
		"String>>trimBoth: is a grease method, so not sure if we should be using it."
		(token includesSubstring: 'realm=') 
			ifTrue: [ ^ ((token substrings: '=') second) ] ].
	^ nil! !

!ZnCredential methodsFor: 'accessing' stamp: ''!
authToken
	^ self authTokenForUrl: nil method: nil ! !

!ZnCredential methodsFor: 'accessing' stamp: ''!
authTokenForUrl: urlObject method: method
	self subclassResponsibility ! !

!ZnCredential methodsFor: 'accessing' stamp: ''!
domain
	^ domain! !

!ZnCredential methodsFor: 'accessing' stamp: ''!
domain: anObject
	domain := anObject! !

!ZnCredential methodsFor: 'testing' stamp: ''!
isComplete
	^ self subclassResponsibility ! !

!ZnCredential methodsFor: 'operations' stamp: ''!
linkToAuthRequest: aHeader
	"Link this credential to the WWW-Authenticate request in the provided header. 
	Used for authentication methods which require parameters provided by the server
	(e.g. Digest authentication)."
	realm := self class findRealmFor: aHeader! !

!ZnCredential methodsFor: 'accessing' stamp: ''!
realm
	^ realm ifNil: [ realm := '' ]! !

!ZnCredential methodsFor: 'accessing' stamp: ''!
realm: anObject
	realm := anObject! !

!ZnCredential methodsFor: 'accessing' stamp: ''!
setAuthorizationHeader: headers for: urlObject method: method
	headers at: 'Authorization' put: self authToken! !

Object subclass: #ZnDefaultServerDelegate
	instanceVariableNames: 'prefixMap byteArray'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Client-Server'!

!ZnDefaultServerDelegate commentStamp: '' prior: 0!
I am ZnDefaultServerDelegate.
I function as a delegate for ZnServer, implementing #handleRequest:

I implement responses to the following prefixes:

/echo - an echo text of request information for debugging purposes
/dw-bench - a dynamic html page for benchmarking purposes
/unicode - a unicode test page
/random - a random string (/random/32 for a specific size)
/bytes - bytes according to a pattern (/bytes/32 for a specific size)
/favicon.ico - a Zn favicon
/status - a server status page
/error - force server errors
/session - a simpler session based counter
/help - lists all page prefixes
/ - an html welcome page

Without any matches, I respond with a page not found.

I can be reused without any default handlers mapped using #empty and configured using #map:to: where the second argument can be a block.

Part of Zinc HTTP Components.!

!ZnDefaultServerDelegate class methodsFor: 'instance creation' stamp: ''!
empty
	^ self basicNew! !

!ZnDefaultServerDelegate methodsFor: 'responses' stamp: ''!
bytes: request
	"Answer a number of bytes according to a pattern as an application/octet-stream.
	Optionally take the size from the URI, as in /bytes/32 for 32 bytes, default to 64 bytes.
	The main goal is to be efficient, esp. in benchmarks, that's why we cache (non-thread-safe)"
	
	| count |
	count := Integer readFrom: request uri lastPathSegment ifFail: [ 64 ].
	(byteArray isNil or: [ byteArray size ~= count ]) 
		ifTrue: [
			byteArray := ByteArray new: count.
			1 to: count - 1 do: [ :each | 
				byteArray at: each put: (#(65 66 67 68) at: (each \\ 4 + 1)) ].
			byteArray at: count put: 10 ].
	^ ZnResponse ok: (ZnEntity bytes: byteArray)! !

!ZnDefaultServerDelegate methodsFor: 'responses' stamp: ''!
dwbench: request
	"Reply with the dynamic DW-Bench HTML page"

	^ ZnResponse ok: (ZnEntity html: self generateDWBench)! !

!ZnDefaultServerDelegate methodsFor: 'responses' stamp: ''!
echoRequest: request	
	"Echo request by generating a plain text response useful for debugging purposes"
	
	| delay info entity |
	delay := Integer readFrom: (request uri queryAt: 'delay' ifAbsent: [ '' ]) ifFail: [ 0 ].
	delay > 0 ifTrue: [ (Delay forSeconds: delay) wait ].  
	info := self generateEchoRequestString: request.
	entity := ZnEntity textCRLF: info.
	^ ZnResponse ok: entity! !

!ZnDefaultServerDelegate methodsFor: 'responses' stamp: ''!
errorResponse: request
	"Here we artificially signal an Error at the Smalltalk level,
	optionally with a user defined message."
	
	| message |
	message := request uri queryAt: 'message' ifAbsent: [ 'An artifical error' ].
	Error signal: message! !

!ZnDefaultServerDelegate methodsFor: 'responses' stamp: ''!
favicon: request
	"Answer a nice favicon for browsers to display"

	| entity |
	entity := ZnEntity with: ZnConstants faviconBytes type: 'image/vnd.microsoft.icon'. 
	^ ZnResponse ok: entity! !

!ZnDefaultServerDelegate methodsFor: 'responses' stamp: ''!
formTest1: request
	| input page |
	input := request uri queryAt: #input ifAbsent: [ 'input' ].
	page := ZnHtmlOutputStream streamContents: [ :html |
		html page: 'Form Test 1' do: [ 
			html 
				tag: #form 
				attributes: #(action 'form-test-1' 'accept-charset' 'utf-8' method GET) 
				do: [ 
					html 
						str: 'Input'; space;
						tag: #input attributes: { #type. #input. #name. #input. #value. input }; space;
						tag: #input attributes: #(type submit); 
						str: 'for input'; space; tag: #em with: input ] ] ].
	^ ZnResponse ok: (ZnEntity html: page)! !

!ZnDefaultServerDelegate methodsFor: 'responses' stamp: ''!
formTest2: request
	| input page |
	input := 'input'.
	(request hasEntity and: [ request contentType matches: ZnMimeType applicationFormUrlEncoded  ])
		ifTrue: [ input := request entity at: #input ifAbsent: [ 'input' ] ].
	page := ZnHtmlOutputStream streamContents: [ :html |
		html page: 'Form Test 2' do: [ 
			html 
				tag: #form 
				attributes: #(action 'form-test-2' 'accept-charset' 'utf-8' method POST) 
				do: [ 
					html 
						str: 'Input'; space;
						tag: #input attributes: { #type. #input. #name. #input. #value. input }; space;
						tag: #input attributes: #(type submit); 
						str: 'for input'; space; tag: #em with: input ] ] ].
	^ ZnResponse ok: (ZnEntity html: page)! !

!ZnDefaultServerDelegate methodsFor: 'responses' stamp: ''!
formTest3: request
	| contents filename contentType page |
	contents := filename := contentType := ''.
	(request hasEntity and: [ request contentType matches: ZnMimeType multiPartFormData ])
		ifTrue: [ 
			(request entity partNamed: #file ifNone: [ nil ]) 
				ifNotNil: [ :part |
					filename := part fileName.
					contents := part contents.
					contentType := part  contentType.
					contentType isBinary ifTrue: [ contents := contents hex ] ] ].
	page := ZnHtmlOutputStream streamContents: [ :html |
		html page: 'Form Test 3' do: [ 
			html 
				tag: #form 
				attributes: #(action 'form-test-3' 'accept-charset' 'utf-8' 
									enctype 'multipart/form-data' method POST) 
				do: [ 
					html 
						str: 'File'; space;
						tag: #input attributes: #(type file name file); space;
						tag: #input attributes: #(type submit) ];
				tag: #p do: [ html str: 'filename = '; str: filename ];
				tag: #p do: [ html str: 'content-type = '; str: contentType asString ];
				tag: #p do: [ html str: 'contents = '; str: contents asString ] ] ].
	^ ZnResponse ok: (ZnEntity html: page)! !

!ZnDefaultServerDelegate methodsFor: 'private' stamp: ''!
generateDWBench
	"Generate the dynamic DW-Bench HTML page.
	Although the date/time is variable, the page size is constant."
	
	^ ZnHtmlOutputStream streamContents: [ :html |
		html page: 'DW-Bench Dynamic' do: [ 
			html tag: #table attributes: #(border 1) do: [
				1 to: 25 do: [ :row |
					html tag: #tr do: [
						1 to: 25 do: [ :col |
							html tag: #td do: [ html print: row * col ] ] ] ] ].
			html tag: #p do: [ 
				html << Date today yyyymmdd; space; << Time now print24 ] ] ]! !

!ZnDefaultServerDelegate methodsFor: 'private' stamp: ''!
generateEchoRequestString: request
	"Generate a string echoing the request"
	
	^ String streamContents: [ :stream |
		stream nextPutAll: 'This is Zinc HTTP Components echoing your request !!'; crlf.
		stream nextPutAll: 'Running '; print: request server; crlf.
		stream nextPutAll: request method; nextPutAll: ' request for '.
		request uri printPathQueryFragmentOn: stream.
		stream crlf; nextPutAll: 'with headers'; crlf.
		request headersDo: [ :key :value |
			stream space; nextPutAll: key; nextPutAll: ': '; nextPutAll: value; crlf ].
		request hasEntity ifTrue: [
			stream nextPutAll: ' containing '; print: request entity ] ]! !

!ZnDefaultServerDelegate methodsFor: 'private' stamp: ''!
generateHelp
	"Generate an HTML page with links to all pages I support"
	
	^ ZnHtmlOutputStream streamContents: [ :html |
		html page: (self class name, ' Help') do: [  	
			html tag: #h3 with: 'Available Pages'.
			html tag: #ul do: [
				prefixMap keys sorted do: [ :each | 
					html tag: #li do: [ 
						html tag: #a attributes: { #href. each } with: each ] ] ] ] ]! !

!ZnDefaultServerDelegate methodsFor: 'private' stamp: ''!
generateSessionRequest: session
	"Generate an HTML page using a session"
	
	^ ZnHtmlOutputStream streamContents: [ :html |
		html page: 'Session' do: [ 
			html tag: #p do: [ html << 'I am using '; str: session asString ].
			html tag: #p do: [ html << 'Current session id is ' << session id ].
			html tag: #p do: [ html << 'Session hit count ' << (session attributeAt: #hitCount) ] ] ]! !

!ZnDefaultServerDelegate methodsFor: 'private' stamp: ''!
generateStatus
	"Generate an HTML page with the state of the server"
	
	^ ZnHtmlOutputStream streamContents: [ :html |
		html page: 'ZnServer Status' do: [
			html tag: #p do: [ html << 'I am '; print: ZnCurrentServer value ].
			html tag: #p with: self systemVersionInfo.
			html tag: #pre with: self vmStats.
			html tag: #h1 with: 'Processes'.
			html tag: #ol do: [ 
				self processes do: [ :each |
					html tag: #li do: [ 
						html 
							str: each name; space; nextPut: $[; print: each priority; nextPut: $]; 
							tag: #br; str: each printString ] ] ] ] ]! !

!ZnDefaultServerDelegate methodsFor: 'private' stamp: ''!
generateUnicodeTest
	"Generate an HTML page containing all Unicode characters with code points between 0 and 16r024F in a nice table"
	
	^ ZnHtmlOutputStream streamContents: [ :html |
		html page: 'Unicode Test Page' do: [ 
			html tag: #table do: [
				0 to: 16r024F by: 8 do: [ :row |
					html tag: #tr do: [
					row to: row + 7 do: [ :each |
						html tag: #td do: [ 
							html 
								nextPutAll: 'U+'; nextPutAll: (each printPaddedWith: $0 to: 4 base: 16);
								nextPut: $=; nextPut: each asCharacter ] ] ] ] ] ] ]! !

!ZnDefaultServerDelegate methodsFor: 'private' stamp: ''!
generateWelcomePage
	^ ZnHtmlOutputStream streamContents: [ :html |
		html html5; tag: #html do: [
			html tag: #head do: [
				html tag: #title with: 'Zinc HTTP Components'.
				html tag: #style attributes: #(type 'text/css') with: self welcomePageCss ].
		html tag: #body do: [ 
			html tag: #div id: #logo with: #Zn.
			html tag: #h1 with: 'Zinc HTTP Components'.
			html tag: #p with: self welcomePageGreeting.
			self welcomePageLinksOn: html.
			html tag: #h4 with: 'May the Source be with you!!' ] ] ]! !

!ZnDefaultServerDelegate methodsFor: 'public' stamp: ''!
handleRequest: request
	| prefix prefixHandler |
	prefix := request uri isSlash
		ifTrue: [ prefixMap at: '/' ifAbsent: [ nil ] ]
		ifFalse: [ request uri firstPathSegment ].
	prefixHandler := prefixMap at: prefix ifAbsent: [ nil ].
	prefixHandler isNil
		ifFalse: [ 
			^ prefixHandler isSymbol
				ifTrue: [ self perform: prefixHandler with: request ]
				ifFalse: [ prefixHandler value: request ] ].
	^ ZnResponse notFound: request uri! !

!ZnDefaultServerDelegate methodsFor: 'responses' stamp: ''!
help: request
	"Reply with a dynamic HTML page containing links to all pages I support"

	^ ZnResponse ok: (ZnEntity html: self generateHelp)
	! !

!ZnDefaultServerDelegate methodsFor: 'initialization' stamp: ''!
initialize
	self
		map: 'welcome' to: #welcome:;
		map: 'help' to: #help:;
		map: 'echo' to: #echoRequest:;
		map: 'small' to: #small:;
		map: 'dw-bench' to: #dwbench:;
		map: 'favicon.ico' to: #favicon:;
		map: 'random' to: #random:;
		map: 'bytes' to: #bytes:;
		map: 'status' to: #status:;
		map: 'unicode' to: #unicode:;
		map: 'session' to: #sessionRequest:;
		map: 'error' to: #errorResponse:;
		map: 'form-test-1' to: #formTest1:;
		map: 'form-test-2' to: #formTest2:;
		map: 'form-test-3' to: #formTest3:;
		map: '/' to: 'welcome'! !

!ZnDefaultServerDelegate methodsFor: 'accessing' stamp: ''!
map: prefix to: handler
	"Arrange for the receiver to dispatch requests matching prefix to handler.
	Handler can be a symbol naming a method in the receiver, 
	or a block accepting the request and producing the response.
	The special prefix '/' is resolved first and can refer to another prefix."
	
	self prefixMap 
		at: prefix
		put: handler
		! !

!ZnDefaultServerDelegate methodsFor: 'accessing' stamp: ''!
prefixMap
	^ prefixMap ifNil: [ prefixMap := Dictionary new ]! !

!ZnDefaultServerDelegate methodsFor: 'private' stamp: ''!
processes
	^ (Process allSubInstances
		reject: [ :each | each isTerminated or: [ each = Processor activeProcess ] ])
		sorted: [ :a :b | a priority >= b priority ] ! !

!ZnDefaultServerDelegate methodsFor: 'responses' stamp: ''!
random: request
	"Answer a text plain UTF-8 encoded string of random hex characters.
	Optionally take the size from the URI, as in /random/32 for 32 characters.
	Default to 64 characters with a newline at the end"
	
	| count hexString |
	count := Integer readFrom: request uri lastPathSegment ifFail: [ 64 ].
	hexString := String new: count streamContents: [ :stream |
		count - 1 timesRepeat: [ stream nextPut: '0123456789ABCDEF' atRandom ].
		stream lf ].
	^ ZnResponse ok: (ZnEntity text: hexString)! !

!ZnDefaultServerDelegate methodsFor: 'responses' stamp: ''!
sessionRequest: request	
	"An HTML response that uses a session"
	
	| session hitCount |
	session := request session.
	hitCount := session attributeAt: #hitCount ifAbsent: [ 0 ].
	hitCount := hitCount + 1.
	session attributeAt: #hitCount put: hitCount. 
	^ ZnResponse ok: (ZnEntity html: (self generateSessionRequest: session))! !

!ZnDefaultServerDelegate methodsFor: 'responses' stamp: ''!
small: request
	| page |
	page := ZnHtmlOutputStream streamContents: [ :html |
		html page: #Small do: [ 
			html tag: #p with: 'This is a small HTML document' ] ].
	^ ZnResponse ok: (ZnEntity html: page)! !

!ZnDefaultServerDelegate methodsFor: 'responses' stamp: ''!
status: request
	"Reply with a dynamic HTML page describing the state of the server"

	^ ZnResponse ok: (ZnEntity html: self generateStatus)
	! !

!ZnDefaultServerDelegate methodsFor: 'private' stamp: ''!
systemVersionInfo
	^ String streamContents: [ :stream |
			stream 
				print: SystemVersion current; 
				nextPutAll: ' - '; 
				nextPutAll: ZnConstants defaultServerString ]
	! !

!ZnDefaultServerDelegate methodsFor: 'responses' stamp: ''!
unicode: request
	"Answer a Unicode Test HTML page"
	
	^ ZnResponse ok: (ZnEntity html: self generateUnicodeTest)! !

!ZnDefaultServerDelegate methodsFor: 'public' stamp: ''!
value: request
	"I implement the generic #value: message as equivalent to #handleRequest:"
	
	^ self handleRequest: request! !

!ZnDefaultServerDelegate methodsFor: 'private' stamp: ''!
vmStats
	^ SmalltalkImage current vm statisticsReport! !

!ZnDefaultServerDelegate methodsFor: 'responses' stamp: ''!
welcome: request
	"Answer a nice HTML welcome page"
	
	| page entity |
	page := self generateWelcomePage.
	entity := ZnEntity html: page. 
	^ ZnResponse ok: entity! !

!ZnDefaultServerDelegate methodsFor: 'private' stamp: ''!
welcomePageCss
	^ Character space join: 
	'body { color: black; background: white; width: 900px; font-family: Verdana, Arial, Helvetica, sans-serif; font-size: 13px }
p { width: 600px; padding: 0 20px 10px 20px }
ul, ol { width: 600px; padding: 0 5px 5px 30px } 
#logo { color: orange; font-family: Helvetica, sans-serif; font-weight: bold; font-size: 100px }' lines! !

!ZnDefaultServerDelegate methodsFor: 'private' stamp: ''!
welcomePageGreeting
	^ 'Welcome to Zinc HTTP Components, a modern, open-source Smalltalk framework to deal with the HTTP networking protocol.'! !

!ZnDefaultServerDelegate methodsFor: 'private' stamp: ''!
welcomePageLinksOn: html
	html tag: #ul do: [ 
		html tag: #li do: [ 
			html 
				str: 'Project homepage'; space;  
				tag: #a attributes: #(href 'http://zn.stfx.eu') with: 'http://zn.stfx.eu' ].
		html tag: #li do: [ 
			html 
				str: 'Locally available pages'; space;  
				tag: #a attributes: #(href '/help') with: '/help' ] ]! !

Object subclass: #ZnDispatcherDelegate
	instanceVariableNames: 'dispatcherDictionary'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Client-Server'!

!ZnDispatcherDelegate commentStamp: '' prior: 0!
I am ZnDispatcherDelegate, doing a straight-forward dispatching to mapped URLs.

ZnDispatcherDelegate API:

(server := ZnServer startDefaultOn: 9090)
	delegate: (ZnDispatcherDelegate new 
		map: '/hello' to: [ :request :response | response entity: (ZnEntity html: '<h1>hello server</h1>') ];
		map: '/counter' to: [ :request :response | counterApplication handleRequest: request response: response ]).!

!ZnDispatcherDelegate methodsFor: 'accessing' stamp: ''!
dispatcherDictionary
	^ dispatcherDictionary ifNil: [ dispatcherDictionary := Dictionary new ]! !

!ZnDispatcherDelegate methodsFor: 'public' stamp: ''!
handleRequest: request
	| handler |
	handler := self dispatcherDictionary 
		at: request uri pathPrintString 
		ifAbsent: [ ^ ZnResponse notFound: request uri ].	
	^ handler 
		value: request 
		value: (ZnResponse new
			statusLine: (ZnStatusLine ok);
			headers: ZnHeaders defaultResponseHeaders;
			yourself).

! !

!ZnDispatcherDelegate methodsFor: 'public' stamp: ''!
map: relativeUrlString to: handler
	self dispatcherDictionary at: relativeUrlString put: handler! !

Object subclass: #ZnEasy
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Client-Server'!

!ZnEasy commentStamp: '' prior: 0!
I am ZnEasy, a simplified HTTP client for occasional use.

My class side implements a number of convenience methods to do common HTTP client operations.

	ZnEasy get: 'http://zn.stfx.eu/zn/small.html'.

For most requests, I return a response object. 

For a couple of message, I return an image Form.

	ZnEasy getPng: 'http://pharo.org/files/pharo.png'.

For my implementation I use ZnClient, a full featured HTTP client.

Part of Zinc HTTP Components.!

!ZnEasy class methodsFor: 'private' stamp: ''!
client
	^ ZnClient new
		beOneShot;
		yourself! !

!ZnEasy class methodsFor: 'operations' stamp: ''!
delete: urlObject
	^ self client
		url: urlObject;
		delete;
		response! !

!ZnEasy class methodsFor: 'operations' stamp: ''!
delete: urlObject username: username password: password
	^ self client
		url: urlObject;
		setBasicAuthenticationUsername: username password: password;
		delete;
		response
	! !

!ZnEasy class methodsFor: 'operations' stamp: ''!
get: urlObject
	"self get: 'http://stfx.eu/small.html'."
	
	^ self client
		url: urlObject;
		get;
		response
	! !

!ZnEasy class methodsFor: 'operations' stamp: ''!
get: urlObject username: username password: password
	^ self client
		url: urlObject;
		setBasicAuthenticationUsername: username password: password;
		get;
		response
	! !

!ZnEasy class methodsFor: 'operations' stamp: ''!
getGif: urlObject
	"self getGif: 'http://zn.stfx.eu/zn/Hot-Air-Balloon.gif'."
	
	^ self 
		getImageOfType: ZnMimeType imageGif 
		fromUrl: urlObject	! !

!ZnEasy class methodsFor: 'private' stamp: ''!
getImageOfType: mimeType fromUrl: urlObject
	| client |
	(client := self client)
		url: urlObject;
		accept: mimeType;
		enforceHttpSuccess: true;
		enforceAcceptContentType: true;
		get.
	"ImageReadWriter does automatic type detection"
	^ ImageReadWriter formFromStream: client entity readStream! !

!ZnEasy class methodsFor: 'operations' stamp: ''!
getJpeg: urlObject
	"self getJpeg: 'http://stfx.eu/sun-fire-x2100.jpg'."
	
	^ self 
		getImageOfType: ZnMimeType imageJpeg 
		fromUrl: urlObject! !

!ZnEasy class methodsFor: 'operations' stamp: ''!
getPng: urlObject
	"self getPng: 'http://pharo.org/files/pharo.png'."
	"(self getPng: 'http://chart.googleapis.com/chart?cht=tx&chl=',
		'a^2+b^2=c^2' urlEncoded) asMorph openInHand."
	
	^ self 
		getImageOfType: ZnMimeType imagePng 
		fromUrl: urlObject! !

!ZnEasy class methodsFor: 'operations' stamp: ''!
head: urlObject
	^ self client
		url: urlObject;
		head;
		response! !

!ZnEasy class methodsFor: 'operations' stamp: ''!
head: urlObject username: username password: password
	^ self client
		url: urlObject;
		setBasicAuthenticationUsername: username password: password;
		head;
		response! !

!ZnEasy class methodsFor: 'operations' stamp: ''!
post: urlObject data: entity
	^ self client
		url: urlObject;
		entity: entity;
		post;
		response! !

!ZnEasy class methodsFor: 'operations' stamp: ''!
post: urlObject data: entity username: username password: password
	^ self client
		url: urlObject;
		setBasicAuthenticationUsername: username password: password;
		entity: entity;
		post;
		response! !

!ZnEasy class methodsFor: 'operations' stamp: ''!
put: urlObject data: entity
	^ self client
		url: urlObject;
		entity: entity;
		put;
		response	! !

!ZnEasy class methodsFor: 'operations' stamp: ''!
put: urlObject data: entity username: username password: password
	^ self client
		url: urlObject;
		setBasicAuthenticationUsername: username password: password;
		entity: entity;
		put;
		response	! !

Object subclass: #ZnEntity
	instanceVariableNames: 'contentType contentLength'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Core'!

!ZnEntity commentStamp: '' prior: 0!
I am ZnEntity, an abstract HTTP Entity.

An entity has a content type (mime-type), a content length (byte length) and a contents.

Entities can encode and decode themselves to and from binary streams.

Part of Zinc HTTP Components.!

ZnEntity subclass: #ZnApplicationFormUrlEncodedEntity
	instanceVariableNames: 'fields representation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Core'!

!ZnApplicationFormUrlEncodedEntity commentStamp: '' prior: 0!
I am ZnApplicationFormUrlEncodedEntity, a concrete HTTP Entity 
holding a set of fields (key value pairs) that is form URL encoded.
I am a ZnEntity.

Part of Zinc HTTP Components.!

!ZnApplicationFormUrlEncodedEntity class methodsFor: 'testing' stamp: ''!
designatedMimeType
	^ ZnMimeType applicationFormUrlEncoded ! !

!ZnApplicationFormUrlEncodedEntity class methodsFor: 'instance creation' stamp: ''!
withAll: keyedCollection
	^ self new
		addAll: keyedCollection;
		yourself ! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'accessing' stamp: ''!
add: association
	self at: association key put: association value! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'accessing' stamp: ''!
addAll: keyedCollection
	"Note that we use #addAllMulti:"

	self invalidateRepresentation.
	self fields addAllMulti: keyedCollection.
	^ keyedCollection ! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'accessing' stamp: ''!
at: key
	^ self fields at: key! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'accessing' stamp: ''!
at: key add: value
	self invalidateRepresentation.
	^ self fields at: key add: value! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'accessing' stamp: ''!
at: key ifAbsent: block
	^ self fields at: key ifAbsent: block! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'accessing' stamp: ''!
at: key put: value
	self invalidateRepresentation.
	^ self fields at: key put: value! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'private' stamp: ''!
computeRepresentation
	representation := String streamContents: [ :stream | self writeRepresentationOn: stream ].
	contentLength := representation size! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'accessing' stamp: ''!
contentLength
	contentLength isNil ifTrue: [ self computeRepresentation ].
	^ contentLength ! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'accessing' stamp: ''!
contents
	^ self fields! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'accessing' stamp: ''!
fields
	^ fields! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'enumerating' stamp: ''!
fieldsDo: block
	self fields keysAndValuesDo: block! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'testing' stamp: ''!
includesField: field
	^ self fields includesKey: field! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'initialization' stamp: ''!
initialize
	super initialize.
	fields := ZnMultiValueDictionary new.
	self contentType: self class designatedMimeType ! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'private' stamp: ''!
invalidateRepresentation
	representation := nil.
	contentLength := nil! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'testing' stamp: ''!
isEmpty
	^ self fields isEmpty! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	self isEmpty
		ifTrue: [ ^ self ].
	stream space.
	fields printElementsOn: stream! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'initialize-release' stamp: ''!
readFrom: stream
	"We parse from stream, limited to content length if available."
	
	| readStream |
	representation := nil.
	contentLength isNil
		ifTrue: [ readStream := ZnCharacterReadStream on: stream encoding: 'ascii' ]
		ifFalse: [ (readStream := ZnLimitedReadStream on: stream limit: contentLength) ascii ].
	fields := ZnResourceMetaUtils parseQueryFrom: readStream! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'writing' stamp: ''!
writeOn: stream
	"In most use cases, contentLength will have been asked first so that a representation is cached"
	
	representation isNil ifTrue: [ self computeRepresentation ].
	ZnUtils nextPutAll: representation on: stream! !

!ZnApplicationFormUrlEncodedEntity methodsFor: 'writing' stamp: ''!
writeRepresentationOn: stream
	self contentType charSet
		ifNotNil: [:charSet|
			ZnResourceMetaUtils
				writeQueryFields: self fields 
				withEncoder: (ZnCharacterEncoder newForEncoding: charSet)
				on: stream] 
		ifNil: [ 
			ZnResourceMetaUtils 
				writeQueryFields: self fields 
				on: stream ]! !

ZnEntity subclass: #ZnByteArrayEntity
	instanceVariableNames: 'bytes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Core'!

!ZnByteArrayEntity commentStamp: '' prior: 0!
I am ZnByteArrayEntity, a concrete HTTP Entity based on a ByteArray.
It is used to hold binary (non textual) data.
I am a ZnEntity.

Part of Zinc HTTP Components.!

!ZnByteArrayEntity class methodsFor: 'instance creation' stamp: ''!
bytes: byteArray
	^ (self type: ZnMimeType applicationOctetStream length: byteArray size)
		bytes: byteArray;
		yourself ! !

!ZnByteArrayEntity class methodsFor: 'testing' stamp: ''!
designatedMimeType
	"I have no designated mime type, I can handle all binary content.
	See #concreteSubclassForType: "
	
	^ nil! !

!ZnByteArrayEntity class methodsFor: 'testing' stamp: ''!
matches: mimeType
	"I have no designated mime type, I can handle all binary content.
	See #concreteSubclassForType: "
	
	^ false! !

!ZnByteArrayEntity methodsFor: 'comparing' stamp: ''!
= other
	^ super = other and: [ self bytes = other bytes ]! !

!ZnByteArrayEntity methodsFor: 'accessing' stamp: ''!
bytes
	^ bytes! !

!ZnByteArrayEntity methodsFor: 'accessing' stamp: ''!
bytes: anObject
	bytes := anObject! !

!ZnByteArrayEntity methodsFor: 'accessing' stamp: ''!
contents
	^ self bytes! !

!ZnByteArrayEntity methodsFor: 'comparing' stamp: ''!
hash
	^ super hash bitXor: self bytes hash! !

!ZnByteArrayEntity methodsFor: 'testing' stamp: ''!
isEmpty 
	^ self bytes isNil or: [ self bytes isEmpty ]! !

!ZnByteArrayEntity methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	self bytes notNil ifTrue: [ 
		stream space; print: self bytes ]! !

!ZnByteArrayEntity methodsFor: 'initialize-release' stamp: ''!
readFrom: stream
	self contentLength isNil
		ifTrue: [ 
			self bytes: (ZnUtils readUpToEnd: stream limit: ZnMaximumEntitySize value).
			self contentLength: self bytes size ]
		ifFalse: [ | byteArray |
			self contentLength > ZnMaximumEntitySize value
				ifTrue: [ ZnEntityTooLarge signal ].
			byteArray := ByteArray ofSize: self contentLength.
			self contentLength > ZnUtils streamingBufferSize 
				ifTrue: [
					ZnUtils streamFrom: stream to: byteArray writeStream size: self contentLength ]
				ifFalse: [
					stream next: self contentLength into: byteArray ]. 
			self bytes: byteArray ]! !

!ZnByteArrayEntity methodsFor: 'accessing' stamp: ''!
readStream
	^ self bytes readStream! !

!ZnByteArrayEntity methodsFor: 'writing' stamp: ''!
writeOn: stream
	ZnUtils nextPutAll: self bytes on: stream! !

!ZnEntity class methodsFor: 'accessing' stamp: ''!
byteArrayEntityClass
	^ ZnByteArrayEntity ! !

!ZnEntity class methodsFor: 'instance creation' stamp: ''!
bytes: bytes
	^ self byteArrayEntityClass bytes: bytes! !

!ZnEntity class methodsFor: 'private' stamp: ''!
concreteSubclassForType: mimeType binary: forceBinary
	"Answer the concrete ZnEntity subclass that handles mimeType"
	
	^ self allSubclasses 
		detect: [ :each | 
			each matches: mimeType ]
		ifNone: [ 
			(mimeType isBinary or: [ forceBinary ])
				ifTrue: [ self byteArrayEntityClass ] 
				ifFalse: [ self stringEntityClass ] ]
! !

!ZnEntity class methodsFor: 'testing' stamp: ''!
designatedMimeType
	"Answer the mime type we are designated to handle.
	See #concreteSubclassForType: "
	
	self subclassResponsibility ! !

!ZnEntity class methodsFor: 'instance creation' stamp: ''!
html: html
	^ self stringEntityClass html: html! !

!ZnEntity class methodsFor: 'instance creation' stamp: ''!
json: text
	^ self stringEntityClass json: text! !

!ZnEntity class methodsFor: 'testing' stamp: ''!
matches: mimeType
	"Return whether a given mimeType can be handled by instances of the receiver"
	
	^ mimeType matches: self designatedMimeType ! !

!ZnEntity class methodsFor: 'instance creation' stamp: ''!
readBinaryFrom: stream usingType: mimeType andLength: length
	"Create and return a new buffered entity of mimeType and length from stream.
	The concrete subclass depends on mimeType, but textual type will remain binary"
	
	| newEntity |
	newEntity := (self concreteSubclassForType: mimeType binary: true) 
						type: mimeType length: length.
	newEntity readBinaryFrom: stream.
	^ newEntity 
! !

!ZnEntity class methodsFor: 'instance creation' stamp: ''!
readFrom: stream usingType: mimeType andLength: length
	"Create and return a new buffered entity of mimeType and length from stream.
	The concrete subclass depends on mimeType"
	
	| newEntity |
	newEntity := (self concreteSubclassForType: mimeType binary: false) 
						type: mimeType length: length.
	newEntity readFrom: stream.
	^ newEntity 
! !

!ZnEntity class methodsFor: 'accessing' stamp: ''!
stringEntityClass
	^ ZnStringEntity! !

!ZnEntity class methodsFor: 'instance creation' stamp: ''!
text: text
	^ self stringEntityClass text: text! !

!ZnEntity class methodsFor: 'instance creation' stamp: ''!
textCRLF: text
	^ self text: text, String crlf! !

!ZnEntity class methodsFor: 'instance creation' stamp: ''!
type: type
	"Create a new entity with a given mime type and (currently) undetermined length"
	
	^ self new
		contentType: type;
		yourself! !

!ZnEntity class methodsFor: 'instance creation' stamp: ''!
type: type length: length
	"Create a new entity with a given mime type and length"
	
	^ self new
		contentType: type;
		contentLength: length;
		yourself! !

!ZnEntity class methodsFor: 'instance creation' stamp: ''!
with: object
	"Create a new entity holding object.
	Currently works for strings and byte arrays.
	You will probably want to set the mime type to something more specific"
	
	^ object isString 
		ifTrue: [ self text: object ] 
		ifFalse: [ self bytes: object ]! !

!ZnEntity class methodsFor: 'instance creation' stamp: ''!
with: object type: type
	"Create a new entity holding object.
	Currently works for strings and byte arrays.
	The entity is set to mime type"
	
	^ (self with: object)
		contentType: type;
		yourself! !

!ZnEntity class methodsFor: 'instance creation' stamp: ''!
xml: text
	^ self stringEntityClass xml: text! !

!ZnEntity methodsFor: 'comparing' stamp: ''!
= other
	self class = other class ifFalse: [ ^ false ].
	^ self contentType = other contentType and: [ self contentLength = other contentLength ]! !

!ZnEntity methodsFor: 'converting' stamp: ''!
asByteArray
	"Return the ByteArray that is the binary, encoded representation of the receiver"
	
	^ ByteArray streamContents: [ :stream | self writeOn: stream ]! !

!ZnEntity methodsFor: 'initialize-release' stamp: ''!
close! !

!ZnEntity methodsFor: 'accessing' stamp: ''!
contentLength 
	"Content length can be nil when it is (currently) unknown."
	
	^ contentLength! !

!ZnEntity methodsFor: 'accessing' stamp: ''!
contentLength: integer
	contentLength := integer! !

!ZnEntity methodsFor: 'accessing' stamp: ''!
contentType
	^ contentType ! !

!ZnEntity methodsFor: 'accessing' stamp: ''!
contentType: object
	"We only allow assignment compatible with our designated mime type"
	
	| newType |
	newType := object asZnMimeType.
	contentType = newType 
		ifTrue: [ ^ self ]
		ifFalse: [ 
			(self class designatedMimeType isNil or: [ self class matches: newType ])
				ifTrue: [ contentType := newType ] ]! !

!ZnEntity methodsFor: 'accessing' stamp: ''!
contents
	self subclassResponsibility! !

!ZnEntity methodsFor: 'testing' stamp: ''!
hasContentLength
	^ contentLength notNil! !

!ZnEntity methodsFor: 'testing' stamp: ''!
hasContentType
	^ contentType notNil! !

!ZnEntity methodsFor: 'comparing' stamp: ''!
hash
	^ self contentType hash bitXor: self contentLength hash! !

!ZnEntity methodsFor: 'testing' stamp: ''!
isEmpty 
	self subclassResponsibility ! !

!ZnEntity methodsFor: 'printing' stamp: ''!
printContentTypeAndLengthOn: stream
	self hasContentType ifTrue: [ 
		stream print: self contentType; space ].
	self hasContentLength ifTrue: [
		stream print: self contentLength; nextPut: $B ]! !

!ZnEntity methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	self printContentTypeAndLengthOn: stream! !

!ZnEntity methodsFor: 'printing' stamp: ''!
printOn: stream
	super printOn: stream.
	stream nextPut: $(.
	self printContentsOn: stream.
	stream nextPut: $)! !

!ZnEntity methodsFor: 'initialize-release' stamp: ''!
readBinaryFrom: stream
	self readFrom: stream! !

!ZnEntity methodsFor: 'initialize-release' stamp: ''!
readFrom: stream
	self subclassResponsibility ! !

!ZnEntity methodsFor: 'writing' stamp: ''!
writeOn: stream
	self subclassResponsibility ! !

ZnEntity subclass: #ZnMultiPartFormDataEntity
	instanceVariableNames: 'parts representation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Core'!

!ZnMultiPartFormDataEntity commentStamp: '' prior: 0!
I am ZnMultiPartFormDataEntity, a concrete HTTP Entity 
holding multi part form data that is encoded according to specific rules.
I am a ZnEntity.

Acknowledgement: some code borrowed from AJP.

Part of Zinc HTTP Components.!

!ZnMultiPartFormDataEntity class methodsFor: 'testing' stamp: ''!
designatedMimeType
	^ ZnMimeType multiPartFormData  ! !

!ZnMultiPartFormDataEntity class methodsFor: 'testing' stamp: ''!
matches: mimeType
	^ mimeType matches: 'multipart/*' asZnMimeType! !

!ZnMultiPartFormDataEntity methodsFor: 'accessing' stamp: ''!
addPart: mimePart
	self invalidateRepresentation.
	self parts add: mimePart! !

!ZnMultiPartFormDataEntity methodsFor: 'private' stamp: ''!
computeRepresentation
	representation := ByteArray streamContents: [ :stream | 
		self writeRepresentationOn: (ZnBivalentWriteStream on: stream) ].
	contentLength := representation size! !

!ZnMultiPartFormDataEntity methodsFor: 'accessing' stamp: ''!
contentLength
	contentLength isNil ifTrue: [ self computeRepresentation ].
	^ contentLength! !

!ZnMultiPartFormDataEntity methodsFor: 'accessing' stamp: ''!
contentType: object
	"We only allow assignment compatible with our designated mime type.
	The main/sub must be equal but the parameters must be allowed to be different"
	
	| newType |
	newType := object asZnMimeType.
	(self class matches: newType)
		ifTrue: [ contentType := newType ]! !

!ZnMultiPartFormDataEntity methodsFor: 'accessing' stamp: ''!
contents
	^ self parts! !

!ZnMultiPartFormDataEntity methodsFor: 'private' stamp: ''!
generateBoundary
	^ String streamContents: [ :stream | | letters |
			stream nextPutAll: 'Boundary-Zn-'.
			letters := 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'.
			8 timesRepeat: [ stream nextPut: letters atRandom ] ]
	! !

!ZnMultiPartFormDataEntity methodsFor: 'accessing' stamp: ''!
getBoundary
	"Return the multipart/form-data MIME part boundary.
	This should really be there, set externally or internally.
	See #initialize and #mimeTypeWithBoundary"
	
	^ (self contentType parameterAt: 'boundary') withoutQuoting ! !

!ZnMultiPartFormDataEntity methodsFor: 'initialize-release' stamp: ''!
initialize
	super initialize.
	self contentType: self mimeTypeWithBoundary  
	! !

!ZnMultiPartFormDataEntity methodsFor: 'private' stamp: ''!
invalidateRepresentation
	representation := nil.
	contentLength := nil! !

!ZnMultiPartFormDataEntity methodsFor: 'testing' stamp: ''!
isEmpty
	^ parts isNil or: [ parts isEmpty ]! !

!ZnMultiPartFormDataEntity methodsFor: 'private' stamp: ''!
mimeTypeWithBoundary
	| mimeType |
	mimeType := self class designatedMimeType copy.
	mimeType parameterAt: 'boundary' put: self generateBoundary.
	^ mimeType! !

!ZnMultiPartFormDataEntity methodsFor: 'private' stamp: ''!
parse: bytes boundary: boundary binary: binary
	| next start fullBoundary |
	fullBoundary := '--' asByteArray, boundary.
	"Sometimes there is whitespace in front of the first boundary"
	start := (bytes
		indexOfSubCollection: fullBoundary
		startingAt: 1) + 1.
	next := bytes
		indexOfSubCollection: fullBoundary
		startingAt: start 
		ifAbsent: [ bytes size ].
	[ next < (bytes size - 2) ] whileTrue: [ | partReadStream |
		partReadStream := ReadStream on: bytes from: start + 2 to: next - 3.
		self addPart: (ZnMimePart perform: (binary ifTrue: [ #readBinaryFrom: ] ifFalse: [ #readFrom: ]) with: partReadStream).
		start := next + fullBoundary size.
		next := bytes
			indexOfSubCollection: fullBoundary
			startingAt: start
			ifAbsent: [ bytes size ] ]! !

!ZnMultiPartFormDataEntity methodsFor: 'accessing' stamp: ''!
partNamed: fieldName
	^ self partNamed: fieldName ifNone: [ self error: 'Cannot find part named ', fieldName asString ]! !

!ZnMultiPartFormDataEntity methodsFor: 'accessing' stamp: ''!
partNamed: fieldName ifNone: block
	parts isNil ifTrue: [ block value ].
	^ self parts detect: [ :each | each fieldName = fieldName ] ifNone: block! !

!ZnMultiPartFormDataEntity methodsFor: 'accessing' stamp: ''!
parts
	parts isNil ifTrue: [ parts := OrderedCollection new ].
	^ parts! !

!ZnMultiPartFormDataEntity methodsFor: 'enumerating' stamp: ''!
partsDo: block
	parts isNil ifTrue: [ ^ self ].
	parts do: block  ! !

!ZnMultiPartFormDataEntity methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	self isEmpty
		ifTrue: [ ^ self ].
	stream space.
	self parts printElementsOn: stream! !

!ZnMultiPartFormDataEntity methodsFor: 'initialize-release' stamp: ''!
readBinaryFrom: stream
	"Switch to streaming implementation later on"
	
	| bytes |
	"We have to use the contentLength instance variable instead of the accessor else a representation will be computed"
	contentLength isNil
		ifTrue: [ 
			bytes := ZnUtils readUpToEnd: stream limit: ZnMaximumEntitySize value ]
		ifFalse: [
			self contentLength > ZnMaximumEntitySize value ifTrue: [ ZnEntityTooLarge signal ].
			bytes := ByteArray ofSize: contentLength.
			stream next: contentLength into: bytes ].
	self parse: bytes boundary: self getBoundary asByteArray binary: true! !

!ZnMultiPartFormDataEntity methodsFor: 'initialize-release' stamp: ''!
readFrom: stream
	"Switch to streaming implementation later on"
	
	| bytes |
	"We have to use the contentLength instance variable instead of the accessor else a representation will be computed"
	contentLength isNil
		ifTrue: [ 
			bytes := ZnUtils readUpToEnd: stream limit: ZnMaximumEntitySize value ]
		ifFalse: [
			self contentLength > ZnMaximumEntitySize value ifTrue: [ ZnEntityTooLarge signal ].
			bytes := ByteArray ofSize: contentLength.
			stream next: contentLength into: bytes ].
	self parse: bytes boundary: self getBoundary asByteArray binary: false! !

!ZnMultiPartFormDataEntity methodsFor: 'writing' stamp: ''!
writeOn: stream
	"We always go via our representation because we need a bivalent stream"
	
	representation isNil ifTrue: [ self computeRepresentation ].
	ZnUtils nextPutAll: representation on: stream! !

!ZnMultiPartFormDataEntity methodsFor: 'writing' stamp: ''!
writeRepresentationOn: stream
	| boundary |
	boundary := self getBoundary.
	self parts do: [ :each |
		stream nextPutAll: '--'; nextPutAll: boundary; nextPutAll: String crlf.
		each writeOn: stream.
		stream nextPutAll: String crlf ].
	stream nextPutAll: '--'; nextPutAll: boundary; nextPutAll: '--'; nextPutAll: String crlf
! !

ZnEntity subclass: #ZnStreamingEntity
	instanceVariableNames: 'stream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Core'!

!ZnStreamingEntity commentStamp: '' prior: 0!
I am ZnStreamingEntity, a concrete HTTP Entity based on a stream.
I am a ZnEntity.

When a ZnStreamingEntity is created using #readFrom: 
its content type and length are set but no body is actually read.
Instead, stream is set to a ZnLimitedReadStream that clients can/must
use to read from.

When the client is done, either all data must be read or #consumeContent must be called.
A client should normally not close the stream.

A ZnStreamingEntity can also be instanciated with an existing read stream.
During #writeOn: this stream is copied to the output stream.

Part of Zinc HTTP Components.!

!ZnStreamingEntity class methodsFor: 'testing' stamp: ''!
designatedMimeType
	"I have no designated mime type, I can handle all content.
	See #concreteSubclassForType: "
	
	^ nil! !

!ZnStreamingEntity class methodsFor: 'testing' stamp: ''!
matches: mimeType
	"I have no designated mime type, I can handle all content.
	See #concreteSubclassForType: "
	
	^ false! !

!ZnStreamingEntity class methodsFor: 'instance creation' stamp: ''!
readFrom: stream usingType: mimeType andLength: length
	"Create and return a new streaming entity of mimeType and length from stream.
	Clients can/must use the embedded ZnLimitedReadStream to read all content, 
	or call #consumeContent when they are done. They should not close the stream."
	
	^ (self type: mimeType length: length)
		readFrom: stream;
		yourself
! !

!ZnStreamingEntity methodsFor: 'initialize-release' stamp: ''!
close
	stream ifNotNil: [ 
		stream close. 
		stream := nil ]! !

!ZnStreamingEntity methodsFor: 'accessing' stamp: ''!
consumeContent
	"To be improved later on using some kind of larger buffer"
	
	[ self stream atEnd ] 
		whileFalse: [ self stream next ]! !

!ZnStreamingEntity methodsFor: 'accessing' stamp: ''!
contents
	"Since this is a non-repeatable entity, 
	#contents will #consumeContent, 
	this can thus only be done once"
	
	stream ifNil: [ ^ nil ].
	^ ZnUtils readUpToEnd: self stream limit: ZnMaximumEntitySize value! !

!ZnStreamingEntity methodsFor: 'testing' stamp: ''!
isEmpty 
	"We actually don't know"
	
	^ false! !

!ZnStreamingEntity methodsFor: 'initialize-release' stamp: ''!
readFrom: readStream
	self contentLength isNil
		ifTrue: [ 
			self stream: readStream ]
		ifFalse: [ | limitedStream |
			limitedStream := ZnLimitedReadStream on: readStream limit: self contentLength.
			self stream: limitedStream ]! !

!ZnStreamingEntity methodsFor: 'accessing' stamp: ''!
readStream 
	^ self stream! !

!ZnStreamingEntity methodsFor: 'accessing' stamp: ''!
stream 
	^ stream! !

!ZnStreamingEntity methodsFor: 'accessing' stamp: ''!
stream: readStream
	stream := readStream! !

!ZnStreamingEntity methodsFor: 'writing' stamp: ''!
writeOn: outputStream
	(self contentLength isNil or: [ self contentLength = 0 ])
		ifTrue: [ 
			ZnUtils 
				streamFrom: self stream 
				to: outputStream ]
		ifFalse: [
			ZnUtils 
				streamFrom: self stream 
				to: outputStream 
				size: self contentLength ].
	self close! !

ZnEntity subclass: #ZnStringEntity
	instanceVariableNames: 'string encoder'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Core'!

!ZnStringEntity commentStamp: '' prior: 0!
I am ZnStringEntity, a concrete HTTP Entity based on a String.
It is used to hold textual (non-binary) data.
I am a ZnEntity.

Optionally, an encoding is used to convert to and from bytes.
The default encoding it UTF-8.

Note that content length is the encoded byte count, not the number of characters in the string.

Part of Zinc HTTP Components.!

!ZnStringEntity class methodsFor: 'testing' stamp: ''!
designatedMimeType
	"I have no designated mime type, I can handle all non-binary content.
	See #concreteSubclassForType: "
	
	^ nil! !

!ZnStringEntity class methodsFor: 'instance creation' stamp: ''!
html: string
	^ (self type: ZnMimeType textHtml)
		string: string;
		yourself! !

!ZnStringEntity class methodsFor: 'instance creation' stamp: ''!
json: string
	^ (self type: ZnMimeType applicationJson)
		string: string;
		yourself! !

!ZnStringEntity class methodsFor: 'testing' stamp: ''!
matches: mimeType
	"I have no designated mime type, I can handle all non-binary content.
	See #concreteSubclassForType: "
	
	^ false! !

!ZnStringEntity class methodsFor: 'instance creation' stamp: ''!
text: string
	^ (self type: ZnMimeType textPlain)
		string: string;
		yourself! !

!ZnStringEntity class methodsFor: 'instance creation' stamp: ''!
xml: string
	^ (self type: ZnMimeType applicationXml)
		string: string;
		yourself! !

!ZnStringEntity methodsFor: 'comparing' stamp: ''!
= other
	^ super = other and: [ self string = other string ]! !

!ZnStringEntity methodsFor: 'private' stamp: ''!
computeContentLength
	self string isNil ifTrue: [ ^ self ].
	contentLength := self encoder encodedByteCountForString: string! !

!ZnStringEntity methodsFor: 'accessing' stamp: ''!
contentLength
	contentLength isNil ifTrue: [ self computeContentLength ].
	^ contentLength! !

!ZnStringEntity methodsFor: 'accessing' stamp: ''!
contentType: object
	"We only allow assignment compatible with our designated mime type"
	
	| newType |
	newType := object asZnMimeType.
	contentType = newType 
		ifTrue: [ ^ self ]
		ifFalse: [ 
			newType isBinary
				ifFalse: [ 
					contentType := newType.
					self invalidateEncoder ] ]! !

!ZnStringEntity methodsFor: 'accessing' stamp: ''!
contents
	^ self string! !

!ZnStringEntity methodsFor: 'accessing' stamp: ''!
encoder
	self initializeEncoder.
	^ encoder! !

!ZnStringEntity methodsFor: 'accessing' stamp: ''!
encoder: anEncoding
	(encoder notNil and: [ anEncoding ~= encoder ]) ifTrue: [ self invalidateContentLength ].
	encoder := anEncoding! !

!ZnStringEntity methodsFor: 'private' stamp: ''!
hasEncoder
	^ encoder notNil! !

!ZnStringEntity methodsFor: 'comparing' stamp: ''!
hash
	^ super hash bitXor: self string hash! !

!ZnStringEntity methodsFor: 'private' stamp: ''!
initializeEncoder
	"If the content-type specifies an encoding, use that.
	If not, use the value of ZnDefaultCharacterEncoder,
	if that is not set, fall back to UTF-8"
	
	| charSet newEncoder |
	self hasEncoder
		ifTrue: [ ^ self ].
	newEncoder := (charSet := contentType charSet) isNil
		ifTrue: [ ZnDefaultCharacterEncoder value ]
		ifFalse: [ 
			(ZnCharacterEncoder newForEncoding: charSet)
				beLenient;
				yourself ].
	self encoder: newEncoder! !

!ZnStringEntity methodsFor: 'private' stamp: ''!
invalidateContentLength
	self contentLength: nil! !

!ZnStringEntity methodsFor: 'private' stamp: ''!
invalidateEncoder
	self encoder: nil! !

!ZnStringEntity methodsFor: 'testing' stamp: ''!
isEmpty 
	^ self string isNil or: [ self string isEmpty ] ! !

!ZnStringEntity methodsFor: 'printing' stamp: ''!
printContentsOn: stream
	super printContentsOn: stream.
	self string notNil ifTrue: [ 
		stream space; nextPutAll: self string ]! !

!ZnStringEntity methodsFor: 'initialize-release' stamp: ''!
readFrom: stream
	| buffer totalRead read readStream stringStream total |
	total := self contentLength.
	readStream := total ifNotNil: [ ZnLimitedReadStream on: stream limit: total ] ifNil: [ stream ].
	buffer := String new: (ZnUtils streamingBufferSize min: (total ifNil: [ ZnUtils streamingBufferSize ])).
	stringStream := nil.
	totalRead := 0.
	self initializeEncoder.
	[ readStream atEnd ] whileFalse: [
		[ read := encoder readInto: buffer startingAt: 1 count: buffer size fromStream: readStream ]
			on: ZnByteStringBecameWideString 
			do: [ :notification | 
					buffer := notification wideString.
					stringStream ifNotNil: [ | wideString position |
						position := stringStream position.
						wideString := WideString from: stringStream originalContents.
						stringStream on: wideString; setFrom: position + 1 to: position ].
					notification resume ].
		totalRead := totalRead + read.
		totalRead > ZnMaximumEntitySize value
			ifTrue: [ ZnEntityTooLarge signal ].
		stringStream ifNil: [ 
			readStream atEnd
				ifTrue: [ ^ self string: (buffer copyFrom: 1 to: read) ]
				ifFalse: [ stringStream := (total ifNil: [ buffer species new ] ifNotNil: [ buffer species new: total ]) writeStream ] ].
		stringStream next: read putAll: buffer startingAt: 1.
		ZnUtils signalProgress: totalRead total: total ].
	self string: (stringStream ifNil: [ String new ] ifNotNil: [ stringStream contents ])! !

!ZnStringEntity methodsFor: 'accessing' stamp: ''!
readStream
	^ self string readStream! !

!ZnStringEntity methodsFor: 'accessing' stamp: ''!
string
	^ string! !

!ZnStringEntity methodsFor: 'accessing' stamp: ''!
string: aString
	(string notNil and: [ aString ~= string ]) ifTrue: [ self invalidateContentLength ].
	string := aString! !

!ZnStringEntity methodsFor: 'writing' stamp: ''!
writeOn: stream
	| totalWritten toWrite total  |
	self initializeEncoder.
	totalWritten := 0.
	total := string size.
	[ totalWritten < total ] whileTrue: [ 
		toWrite := ZnUtils streamingBufferSize min: (total - totalWritten).
		encoder next: toWrite putAll: string startingAt: totalWritten + 1 toStream: stream.
		totalWritten := totalWritten + toWrite.
		ZnUtils signalProgress: totalWritten total: total ]! !

Object subclass: #ZnEntityReader
	instanceVariableNames: 'headers stream streaming binary allowReadingUpToEnd'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Streaming'!

!ZnEntityReader commentStamp: '' prior: 0!
I am ZnEntityReader, I help reading ZnEntities from a stream given meta data in headers.

I deal with chunking and gzip decoding.

I have several options:
	- to read streaming entities, where the client has to do the actual reading
	- to read binary entities, where textual content is not interpreted
	- to allow reading up to end, when there is no content length set

Part of Zinc HTTP Components.!

!ZnEntityReader methodsFor: 'accessing' stamp: ''!
allowReadingUpToEnd
	allowReadingUpToEnd := true! !

!ZnEntityReader methodsFor: 'testing' stamp: ''!
allowsReadingUpToEnd
	^ allowReadingUpToEnd notNil and: [ allowReadingUpToEnd ]! !

!ZnEntityReader methodsFor: 'accessing' stamp: ''!
binary
	binary := true! !

!ZnEntityReader methodsFor: 'testing' stamp: ''!
canReadContent
	^ self hasContentLength or: [ self isChunked or: [ self allowsReadingUpToEnd ] ]! !

!ZnEntityReader methodsFor: 'accessing' stamp: ''!
contentLength
	^ (self headers includesKey: 'Content-Length')
		ifTrue: [ self headers contentLength ]
		ifFalse: [ nil ]! !

!ZnEntityReader methodsFor: 'accessing' stamp: ''!
contentType
	^ (self headers includesKey: 'Content-Type') 
		ifTrue: [ self headers contentType ] 
		ifFalse: [ ZnMimeType default ]! !

!ZnEntityReader methodsFor: 'testing' stamp: ''!
hasContentLength
	^ self headers hasContentLength and: [ self headers contentLength > 0 ]! !

!ZnEntityReader methodsFor: 'accessing' stamp: ''!
headers
	^ headers! !

!ZnEntityReader methodsFor: 'accessing' stamp: ''!
headers: znHeaders
	headers := znHeaders! !

!ZnEntityReader methodsFor: 'testing' stamp: ''!
isBinary
	^ binary notNil and: [ binary ]! !

!ZnEntityReader methodsFor: 'testing' stamp: ''!
isChunked
	| transferEncoding |
	transferEncoding := self headers at: 'Transfer-Encoding' ifAbsent: [ ^ false ].
	^ transferEncoding = 'chunked' or: [ transferEncoding beginsWith: 'chunked' ]! !

!ZnEntityReader methodsFor: 'testing' stamp: ''!
isGzipped
	| contentEncoding |
	contentEncoding := self headers at: 'Content-Encoding' ifAbsent: [ ^ false ].
	^ contentEncoding = 'gzip'! !

!ZnEntityReader methodsFor: 'testing' stamp: ''!
isIdentityEncoded
	^ self isGzipped not! !

!ZnEntityReader methodsFor: 'testing' stamp: ''!
isStreaming
	^ streaming notNil and: [ streaming ]! !

!ZnEntityReader methodsFor: 'public' stamp: ''!
readEntity
	| entity |
	self canReadContent ifFalse: [ ^ nil ].
	entity := self readEntityFromStream.
	^ entity isEmpty
		ifTrue: [ nil ]
		ifFalse: [ entity ]    ! !

!ZnEntityReader methodsFor: 'private' stamp: ''!
readEntityFromStream
	| entity decodedEntityLength chunkedStream isChunked |
	chunkedStream := (isChunked := self isChunked) ifTrue: [ stream := ZnChunkedReadStream on: stream ] ifFalse: [ nil ].
	self isGzipped 
		ifTrue: [ 
			self hasContentLength ifTrue: [ stream := ZnLimitedReadStream on: stream limit: self contentLength ].
			decodedEntityLength := nil.
			stream := GZipReadStream on: stream ]
		ifFalse: [ decodedEntityLength := self contentLength ].
	entity := self readFrom: stream usingType: self contentType andLength: decodedEntityLength.
	isChunked ifTrue: [ 
		self isIdentityEncoded ifTrue: [ entity contentLength: stream totalSize ].
		chunkedStream extraHeaders ifNotNil: [ :extraHeaders | self headers addAll: extraHeaders ] ].
	^ entity! !

!ZnEntityReader methodsFor: 'private' stamp: ''!
readFrom: readStream usingType: contentType andLength: length
	| entityClass selector |
	entityClass := self isStreaming 
		ifTrue: [ ZnStreamingEntity ]
		ifFalse: [ ZnEntity ].
	selector := self isBinary 
		ifTrue: [ #readBinaryFrom:usingType:andLength: ]
		ifFalse: [ #readFrom:usingType:andLength: ].
	^ entityClass perform: selector 
		with: readStream
		with: contentType 
		with: length! !

!ZnEntityReader methodsFor: 'accessing' stamp: ''!
stream
	^ stream! !

!ZnEntityReader methodsFor: 'accessing' stamp: ''!
stream: readStream
	stream := readStream ! !

!ZnEntityReader methodsFor: 'accessing' stamp: ''!
streaming
	streaming := true! !

Object subclass: #ZnEntityWriter
	instanceVariableNames: 'headers stream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Streaming'!

!ZnEntityWriter commentStamp: '' prior: 0!
I am ZnEntityWriter, I help writing ZnEntities to a stream given meta data in headers.

I deal with chunking and gzip encoding.

Part of Zinc HTTP Components.!

!ZnEntityWriter methodsFor: 'accessing' stamp: ''!
contentType
	^ (self headers includesKey: 'Content-Type') 
		ifTrue: [ self headers contentType ] 
		ifFalse: [ ZnMimeType default ]! !

!ZnEntityWriter methodsFor: 'accessing' stamp: ''!
headers
	^ headers! !

!ZnEntityWriter methodsFor: 'accessing' stamp: ''!
headers: znHeaders
	headers := znHeaders! !

!ZnEntityWriter methodsFor: 'testing' stamp: ''!
isChunked
	| transferEncoding |
	transferEncoding := self headers at: 'Transfer-Encoding' ifAbsent: [ ^ false ].
	^ transferEncoding = 'chunked' or: [ transferEncoding beginsWith: 'chunked' ]! !

!ZnEntityWriter methodsFor: 'testing' stamp: ''!
isGzipped
	| contentEncoding |
	contentEncoding := self headers at: 'Content-Encoding' ifAbsent: [ ^ false ].
	^ contentEncoding = 'gzip'! !

!ZnEntityWriter methodsFor: 'accessing' stamp: ''!
stream
	^ stream! !

!ZnEntityWriter methodsFor: 'accessing' stamp: ''!
stream: writeStream
	stream := writeStream! !

!ZnEntityWriter methodsFor: 'public' stamp: ''!
writeEntity: entity
	| actualStream gzipStream bufferedStream chunkedStream isChunked |
	actualStream := stream.
	chunkedStream := gzipStream := bufferedStream := nil.
	(isChunked := self isChunked)
		ifTrue: [ 
			chunkedStream := actualStream := ZnChunkedWriteStream on: actualStream ].
	self isGzipped 
		ifTrue: [ 
			gzipStream := actualStream := GZipWriteStream on: actualStream ]
		ifFalse: [ 
			(isChunked and: [ self contentType isBinary not ])
				ifTrue: [ bufferedStream := actualStream := ZnUtils bufferedWriteStreamOn: actualStream ] ].
	entity writeOn: actualStream.
	gzipStream 
		ifNotNil: [ gzipStream finish ].
	bufferedStream 
		ifNotNil: [ bufferedStream finish ].
	chunkedStream
		ifNotNil: [ chunkedStream finish ]
	! !

Object subclass: #ZnFileSystemUtils
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Support'!

!ZnFileSystemUtils commentStamp: 'SvenVanCaekenberghe 7/11/2018 16:29' prior: 0!
I am ZnFileSystemUtils.
I help is abstracting over API differences pre and post the introduction of FileSystem in Pharo 2.0 as well as pre and post the important stream and file changes midway in Pharo 7.0

Part of Zinc HTTP Components.!

!ZnFileSystemUtils class methodsFor: 'paths' stamp: ''!
baseNameFor: fileName
	^ fileName asFileReference basename ! !

!ZnFileSystemUtils class methodsFor: 'streams' stamp: 'SvenVanCaekenberghe 9/27/2018 14:42'!
binaryReadStreamFor: fileName
	| fileReference |
	fileReference := fileName asFileReference.
	^ (fileReference respondsTo: #binaryReadStream)
		ifTrue: [ fileReference binaryReadStream ]
		ifFalse: [ fileReference readStream binary ]! !

!ZnFileSystemUtils class methodsFor: 'streams' stamp: 'SvenVanCaekenberghe 11/1/2018 23:54'!
binaryWriteStreamFor: fileName
	| fileReference |
	fileReference := fileName asFileReference.
	^ (fileReference respondsTo: #binaryWriteStream)
		ifTrue: [ fileReference binaryWriteStream ]
		ifFalse: [ fileReference writeStream binary ]! !

!ZnFileSystemUtils class methodsFor: 'streams' stamp: 'SvenVanCaekenberghe 11/1/2018 23:54'!
binaryWriteStreamFor: fileName do: block
	| fileReference |
	fileReference := fileName asFileReference.
	^ (fileReference respondsTo: #binaryWriteStreamDo:ifPresent:)
		ifTrue: [
			fileName asFileReference 
				binaryWriteStreamDo: block 
				ifPresent: [ FileExists signalWith: fileReference ] ]
		ifFalse: [ 
			fileReference isFile 
				ifTrue: [ FileExists signalWith: fileReference ] 
				ifFalse: [ | binaryStream |
					binaryStream := self binaryWriteStreamFor: fileName.
					[ block value: binaryStream ] ensure: [ binaryStream close ] ] ]! !

!ZnFileSystemUtils class methodsFor: 'defaults' stamp: ''!
defaultDirectory
	^ FileSystem disk workingDirectory! !

!ZnFileSystemUtils class methodsFor: 'defaults' stamp: ''!
defaultDirectoryPath
	^ self defaultDirectory fullName! !

!ZnFileSystemUtils class methodsFor: 'paths' stamp: ''!
delete: path
	^ path asFileReference delete! !

!ZnFileSystemUtils class methodsFor: 'paths' stamp: ''!
deleteIfExists: path
	^ (self exists: path)
		ifTrue: [ self delete: path. true ]
		ifFalse: [ false ]! !

!ZnFileSystemUtils class methodsFor: 'paths' stamp: ''!
directory: path
	^ path asFileReference! !

!ZnFileSystemUtils class methodsFor: 'paths' stamp: ''!
downloadPathTo: directoryOrFileName for: baseName
	| downloadPath |
	downloadPath := directoryOrFileName asFileReference.
	^ downloadPath isDirectory
		ifTrue: [
			baseName ifNil: [ 
				self error: 'Not enough information to construct a download path' ]. 
			(downloadPath / baseName) fullName ]
		ifFalse: [ 
			directoryOrFileName ]
	! !

!ZnFileSystemUtils class methodsFor: 'paths' stamp: ''!
exists: path
	^ path asFileReference exists! !

!ZnFileSystemUtils class methodsFor: 'paths' stamp: ''!
extensionFor: fileName
	^ fileName asFileReference extension! !

!ZnFileSystemUtils class methodsFor: 'queries' stamp: ''!
fileSizeFor: fileName
	^ fileName asFileReference entry size! !

!ZnFileSystemUtils class methodsFor: 'paths' stamp: ''!
fullNameFor: fileName
	^ fileName asFileReference fullName! !

!ZnFileSystemUtils class methodsFor: 'queries' stamp: ''!
modificationTimeFor: path
	^ path asFileReference entry modificationTime ! !

!ZnFileSystemUtils class methodsFor: 'streams' stamp: 'SvenVanCaekenberghe 9/27/2018 14:43'!
readStreamFor: fileName
	^ fileName asFileReference readStream! !

!ZnFileSystemUtils class methodsFor: 'streams' stamp: 'SvenVanCaekenberghe 11/1/2018 23:49'!
readStreamFor: fileName do: block
	^ fileName asFileReference 
		readStreamDo: block
		ifAbsent: [ FileDoesNotExist signalWith: fileName asFileReference ]! !

!ZnFileSystemUtils class methodsFor: 'streams' stamp: 'SvenVanCaekenberghe 11/1/2018 23:51'!
writeStreamFor: fileName do: block
	^ fileName asFileReference 
		writeStreamDo: block 
		ifPresent: [ FileExists signalWith: fileName asFileReference ]! !

Object subclass: #ZnHeaders
	instanceVariableNames: 'headers'
	classVariableNames: 'CommonHeaders'
	poolDictionaries: ''
	category: 'Zinc-HTTP-Core'!

!ZnHeaders commentStamp: '' prior: 0!
I am ZnHeaders.
I am a collection of HTTP headers.
I can be used for generating and parsing.

Header names are normalized and used case insensitive.
Header values can be multi-valued.

Part of Zinc HTTP Components.!

!ZnHeaders class methodsFor: 'instance creation' stamp: ''!
defaultRequestHeaders
	^ self new
		at: 'User-Agent' put: ZnConstants defaultUserAgent;
		at: 'Accept' put: '*/*';
		yourself! !

!ZnHeaders class methodsFor: 'instance creation' stamp: ''!
defaultResponseHeaders
	^ self new
		at: 'Server' put: ZnConstants defaultServerString;
		at: 'Date' put: ZnUtils httpDate;
		yourself! !

!ZnHeaders class methodsFor: 'class initialization' stamp: ''!
initialize
	CommonHeaders := (
		#(
			'Content-Length' 'Content-Type' 'Date' 'Server' 'Connection' 'User-Agent' 'Host'
			'Accept' 'Accept-Language' 'Accept-Encoding'
			'Referer' 'Dnt'
			'Cookie' 'Set-Cookie' 'Authorization' 'WWW-Authenticate' 'Proxy-Authorization'
			'Content-Encoding' 'Transfer-Encoding' 'Location'
			'If-Modified-Since' 'Content-Disposition' 
			'X-Zinc-Remote-Address' 'X-Forwarded-Server' 'X-Forwarded-For' 'X-Forwarded-Host' ) 
			collect: [ :each | ZnUtils capitalizeString: each ]) asSet! !

ZnHeaders initialize!

!ZnHeaders class methodsFor: 'instance creation' stamp: ''!
readFrom: stream
	^ self new
		readFrom: stream;
		yourself! !

!ZnHeaders class methodsFor: 'instance creation' stamp: ''!
requestHeadersFor: url
	| headers |
	headers := self defaultRequestHeaders.
	headers request: url.
	^ headers! !

!ZnHeaders class methodsFor: 'instance creation' stamp: ''!
withAll: keyedCollection
	^ self new
		addAll: keyedCollection;
		yourself ! !

!ZnHeaders methodsFor: 'comparing' stamp: ''!
= other
	self class = other class ifFalse: [ ^ false ].
	^ self headers = other headers! !

!ZnHeaders methodsFor: 'accessing' stamp: ''!
acceptEntityDescription: entity
	"Take over the content type and length from entity"
	
	entity 
		ifNotNil: [
			self
				contentType: entity contentType;
				contentLength: entity contentLength ]
		ifNil: [
			self 
				clearContentType; 
				clearContentLength ] ! !

!ZnHeaders methodsFor: 'accessing' stamp: ''!
addAll: keyedCollection
	"Note that we use #addAllMulti:"

	keyedCollection isEmpty
		ifFalse: [ self headers addAllMulti: keyedCollection ].
	^ keyedCollection! !

!ZnHeaders methodsFor: 'accessing' stamp: ''!
at: headerName
	"Return the value stored under headerName,
	a String for single-valued headers, an Array of Strings for multi-valued headers"
	
	^ self headers at: (self normalizeHeaderKey: headerName)! !

!ZnHeaders methodsFor: 'accessing' stamp: ''!
at: headerName add: value
	"Store value under headerName, optionally turning it into a multi-valued header 
	when a value was already present"
	
	^ self headers at: (self normalizeHeaderKey: headerName) add: value 
		
! !

!ZnHeaders methodsFor: 'accessing' stamp: ''!
at: headerName ifAbsent: block
	"Like #at: but executes block when the headerName is not found"
	
	self isEmpty ifTrue: [ ^ block value ].
	^ self headers at: (self normalizeHeaderKey: headerName) ifAbsent: block! !

!ZnHeaders methodsFor: 'accessing' stamp: ''!
at: headerName ifPresent: block
	"Like #at: but executes block when the headerName is found"

	^ self headers at: (self normalizeHeaderKey: headerName) ifPresent: block! !

!ZnHeaders methodsFor: 'accessing' stamp: ''!
at: headerName put: value
	"Store value under headerName, replace existing entries"
	
	^ self headers at: (self normalizeHeaderKey: headerName) put: value! !

!ZnHeaders methodsFor: 'accessing' stamp: ''!
at: headerName put: value ifPresentMerge: binaryBlock
	"Store value under headerName, when there is an existing entry,
	stored the result of evaluating binaryBlock with old and new value"
	
	| normalizedKey existingValue newValue |
	normalizedKey := self normalizeHeaderKey: headerName.
	existingValue := self headers at: normalizedKey ifAbsent: [ nil ].
	newValue := existingValue isNil
		ifTrue: [ value ] 
		ifFalse: [ binaryBlock value: existingValue value: value ].
	^ self headers at: normalizedKey put: newValue 
		

! !

!ZnHeaders methodsFor: 'accessing' stamp: ''!
clearContentLength
	self removeKey: 'Content-Length' ifAbsent: [] ! !

!ZnHeaders methodsFor: 'accessing' stamp: ''!
clearContentType
	self removeKey: 'Content-Type' ifAbsent: [] ! !

!ZnHeaders methodsFor: 'accessing' stamp: ''!
contentLength
	"We allow multiple content-length headers provided they are identical."
	
	| value |
	(value := self headers at: 'Content-Length') isString
		ifFalse: [
			value asSet size = 1
				ifTrue: [ value := value first ]
				ifFalse: [ self error: 'Multiple, different Content-Length headers are not allowed' ] ].
	^ Integer readFrom: value ifFail: [ self error: 'Illegal HTTP Content Length' ] ! !

!ZnHeaders methodsFor: 'accessing' stamp: ''!
contentLength: object
	self at: 'Content-Length' put: object asString! !

!ZnHeaders methodsFor: 'accessing' stamp: ''!
contentType
	| headerValue |
	headerValue := (self headers at: 'Content-Type') trimBoth.
	headerValue isEmpty ifTrue: [ ^ ZnMimeType default ].
	^ ZnMimeType fromString: headerValue! !

!ZnHeaders methodsFor: 'accessing' stamp: ''!
contentType: object
	self at: 'Content-Type' put: object asString! !

!ZnHeaders methodsFor: 'private' stamp: ''!
extendHeaderAt: key from: line
	"The value of a continuation header line is concatenated, 
	keeping the whitespace, but without the CRLF"

	| existingValue newValue |
	existingValue := self at: key.
	newValue := existingValue isArray
		ifTrue: [ | last |
			last := existingValue size.
			existingValue at: last put: ((existingValue at: last) , line).
			existingValue ]
		ifFalse: [ existingValue , line ].
	self at: key put: newValue! !

!ZnHeaders methodsFor: 'testing' stamp: ''!
hasContentLength
	^ self includesKey: 'Content-Length'! !

!ZnHeaders methodsFor: 'testing' stamp: ''!
hasContentType
	^ self includesKey: 'Content-Type'! !

!ZnHeaders methodsFor: 'comparing' stamp: ''!
hash
	^ self headers hash! !

!ZnHeaders methodsFor: 'private' stamp: ''!
headers
	headers isNil ifTrue: [ headers := ZnMultiValueDictionary new ].
	^ headers! !

!ZnHeaders methodsFor: 'enumerating' stamp: ''!
headersDo: block
	"Execute a two argument (key, value) block for each header.
	Multi-valued headers are handled transparently."

	self isEmpty ifTrue: [ ^ self ].
	self headers keysAndValuesDo: [ :headerKey :headerValue |
		headerValue isArray
			ifTrue: [ 
				headerValue do: [ :each | 
					block value: headerKey value: each ] ]
			ifFalse: [
				block value: headerKey value: headerValue ] ]! !

!ZnHeaders methodsFor: 'testing' stamp: ''!
includesKey: headerName
	^ self isEmpty not and: [ self headers includesKey: (self normalizeHeaderKey: headerName) ]
	! !

!ZnHeaders methodsFor: 'testing' stamp: ''!
isDescribingEntity
	"Do I include enough information to describe an entity (i.e. content length and type) ?"
	
	^ (self headers includesKey: 'Content-Type') 
		and: [ self headers includesKey: 'Content-Length' ]
! !

!ZnHeaders methodsFor: 'testing' stamp: ''!
isEmpty 
	^ headers isNil or: [ self headers isEmpty ]! !

!ZnHeaders methodsFor: 'enumerating' stamp: ''!
keysAndValuesDo: block
	"Execute a two argument (key, value) block for each header.
	Multi-valued headers are handled transparently."

	self headersDo: block! !

!ZnHeaders methodsFor: 'private' stamp: ''!
normalizeHeaderKey: string
	"Test string to see if it has proper header key capitalization,
	if true, return string, if not, return a properly capitalized copy"
	
	"Optimization: if string is a know,, properly capitalized header, use it as is"
	(CommonHeaders includes: string) 
		ifTrue: [ ^ string ].
	^ (ZnUtils isCapitalizedString: string)
		ifTrue: [ string ] 
		ifFalse: [ ZnUtils capitalizeString: string ] ! !

!ZnHeaders methodsFor: 'copying' stamp: ''!
postCopy
	headers := headers copy! !

!ZnHeaders methodsFor: 'printing' stamp: ''!
printOn: stream
	super printOn: stream.
	self isEmpty ifFalse: [ self headers printElementsOn: stream ]
	! !

!ZnHeaders methodsFor: 'initialize-release' stamp: ''!
readFrom: stream
	| line reader |
	reader := ZnLineReader on: stream.
	[ (line := reader nextLine) isEmpty ] whileFalse: [ | key |
		key := self readOneHeaderFrom: line readStream.
		"Continuation header lines start with a space or tab"
		[ stream atEnd not and: [ #[ 32 9 ] includes: stream peek asInteger ] ]
			whileTrue: [ 
				self extendHeaderAt: key from: reader nextLine ] ] ! !

!ZnHeaders methodsFor: 'private' stamp: ''!
readOneHeaderFrom: stream
	| key value |
	key := stream upTo: $:.
	[ stream peek == Character space ] whileTrue: [ stream next ].
	value := stream upToEnd.
	self at: key add: value.
	^ key! !

!ZnHeaders methodsFor: 'accessing' stamp: ''!
removeKey: headerName	
	^ self headers removeKey: (self normalizeHeaderKey: headerName)! !

!ZnHeaders methodsFor: 'accessing' stamp: ''!
removeKey: headerName ifAbsent: block
	self isEmpty ifTrue: [ ^ block value ].
	^ self headers removeKey: (self normalizeHeaderKey: headerName) ifAbsent: block! !

!ZnHeaders methodsFor: 'accessing' stamp: ''!
request: url
	"Setup the receiver to request url"
	
	(url isNil or: [ url hasHost not ]) ifTrue: [ ^ self ].  
	self at: 'Host' put: url authority.
	(ZnNetworkingUtils proxyAuthorizationHeaderValueToUrl: url)
		ifNotNil: [ :value | self at: 'Proxy-Authorization' put: value ].    
  ! !

!ZnHeaders methodsFor: 'initialize-release' stamp: ''!
unlimited
	self headers unlimited! !

!ZnHeaders methodsFor: 'writing' stamp: ''!
writeOn: stream
	| writeStream |
	writeStream := ZnBivalentWriteStream on: stream.
	self headersDo: [ :key :value |
		writeStream nextPutAll: key; nextPut: $:; space; nextPutAll: value; nextPutAll: String crlf ]! !

Object subclass: #ZnHtmlOutputStream
	instanceVariableNames: 'stream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Streaming'!

!ZnHtmlOutputStream commentStamp: '' prior: 0!
I am ZnHtmlOutputStream. I wrap another character write stream to offer a richer API for generating correct HTML markup.

See https://en.wikipedia.org/wiki/HTML

My streaming protocol contains the traditional write stream operations. These are raw and do not do any conversions/escaping.

My html protocols contains a rich API for generating correct HTML. 

String streamContents: [ :out | | html |
	html := ZnHtmlOutputStream on: out.
	html html5.
	html tag: #html do: [ 
		html tag: #body do: [
			html tag: #div class: #main do: [
				html tag: #p with: 'Hello World & Universe !!'.
				html tag: #hr.
				html 
					tag: #em 
					attributes: #(class big id 1 disable nil) 
					with: 'The END' ] ] ] ].
	
ZnHtmlOutputStream streamContents: [ :html |
	html page: 'Hello World' do: [ 
		html tag: #div class: #main do: [
			html tag: #p with: 'Hello World & Universe !!' ] ] ]

Part of Zinc HTTP Components.!

!ZnHtmlOutputStream class methodsFor: 'instance creation' stamp: ''!
on: writeStream
	^ self new
		on: writeStream;
		yourself! !

!ZnHtmlOutputStream class methodsFor: 'convenience' stamp: ''!
streamContents: block
	"Execute block with a ZnHtmlOutputStream as argument to generate HTML, returning the resulting string"
	
	"ZnHtmlOutputStream streamContents: [ :html |
		html page: 'Hello World' do: [ 
			html tag: #p with: 'Hello World & Universe !!' ] ]"
	
	^ String streamContents: [ :out |
			block value: (self on: out) ]! !

!ZnHtmlOutputStream methodsFor: 'streaming' stamp: ''!
<< collection
	^ self nextPutAll: collection asString! !

!ZnHtmlOutputStream methodsFor: 'initialize-release' stamp: ''!
close
	stream close! !

!ZnHtmlOutputStream methodsFor: 'private html' stamp: ''!
closeTag: tag
	stream
		nextPut: $<;
		nextPut: $/;
		nextPutAll: tag;
		nextPut: $>! !

!ZnHtmlOutputStream methodsFor: 'html' stamp: ''!
comment: string
	"Write an HTML/XML comment consisting of string, as in <!!-- string -->"
	
	stream
		nextPutAll: '<!!-- ';
		nextPutAll: string;
		nextPutAll: ' -->'! !

!ZnHtmlOutputStream methodsFor: 'html' stamp: ''!
escape: string
	"Write string, escaping characters as needed for regular text"
	
	string do: [ :each | self escapeCharacter: each ]! !

!ZnHtmlOutputStream methodsFor: 'html' stamp: ''!
escapeAttributeValue: string
	"Write string, escaping characters as needed for the value of an attribute"

	string do: [ :each |
		each == $"
			ifTrue: [ stream nextPutAll: '&quot;' ]
			ifFalse: [ stream nextPut: each ] ] ! !

!ZnHtmlOutputStream methodsFor: 'html' stamp: ''!
escapeCharacter: char
	"Write char, escaping it as needed for regular text"

	char == $<
		ifTrue: [ stream nextPutAll: '&lt;' ]
		ifFalse: [ 
			char == $&
				ifTrue: [ stream nextPutAll: '&amp;' ]
				ifFalse: [ stream nextPut: char ] ]! !

!ZnHtmlOutputStream methodsFor: 'streaming' stamp: ''!
flush
	stream flush! !

!ZnHtmlOutputStream methodsFor: 'html' stamp: ''!
format: string with: args
	"Like String>>#format: format the string template using the args given, esacaping characters when needed for regular text"

	| input currentChar |
	input := string readStream.
	[ input atEnd ] whileFalse: [
		(currentChar := input next) == ${
			ifTrue: [ | expression index | 
				expression := input upTo: $}.
				index := Integer readFrom: expression ifFail: [ expression ].
				self escape: (args at: index) asString ]
			ifFalse: [
				currentChar == $\
					ifTrue: [ input atEnd ifFalse: [ self escapeCharacter: stream next ] ]
					ifFalse: [ self escapeCharacter: currentChar ] ] ]! !

!ZnHtmlOutputStream methodsFor: 'html' stamp: ''!
html5
	"Write the standard HTML5 DOCTYPE tag"
	
	stream nextPutAll: '<!!DOCTYPE html>'! !

!ZnHtmlOutputStream methodsFor: 'streaming' stamp: ''!
next: count putAll: collection	
	self 
		next: count 
		putAll: collection 
		startingAt: 1! !

!ZnHtmlOutputStream methodsFor: 'streaming' stamp: ''!
next: count putAll: collection startingAt: offset
	"Write count characters from collection starting at offset."
	
	stream 
		next: count 
		putAll: collection 
		startingAt: offset! !

!ZnHtmlOutputStream methodsFor: 'streaming' stamp: ''!
nextPut: object
	^ stream nextPut: object! !

!ZnHtmlOutputStream methodsFor: 'streaming' stamp: ''!
nextPutAll: collection
	self 
		next: collection size 
		putAll: collection 
		startingAt: 1! !

!ZnHtmlOutputStream methodsFor: 'initialize-release' stamp: ''!
on: writeStream
	stream := writeStream! !

!ZnHtmlOutputStream methodsFor: 'private html' stamp: ''!
openTag: tag
	stream
		nextPut: $<;
		nextPutAll: tag;
		nextPut: $>! !

!ZnHtmlOutputStream methodsFor: 'private html' stamp: ''!
openTag: tag attribute: name value: value
	stream
		nextPut: $<;
		nextPutAll: tag;
		space; 
		nextPutAll: name.
	value ifNotNil: [
		stream nextPutAll: '="'.
		self escapeAttributeValue: value asString.
		stream nextPut: $" ].
	stream nextPut: $>! !

!ZnHtmlOutputStream methodsFor: 'private html' stamp: ''!
openTag: tag attributes: attributes
	stream
		nextPut: $<;
		nextPutAll: tag.
	self renderAttributes: attributes.
	stream nextPut: $>! !

!ZnHtmlOutputStream methodsFor: 'html' stamp: ''!
page: title do: block
	"Write out a standard page using title, then execute block"
	
	"ZnHtmlOutputStream streamContents: [ :html | html page: 'Hello' do: [ html tag: #p with: 'World' ] ]"

	self html5; tag: #html do: [ 
		self
			tag: #head do: [ self tag: #title with: title ];
			tag: #body do: [ 
				self tag: #h1 with: title.
				block value ] ]! !

!ZnHtmlOutputStream methodsFor: 'streaming' stamp: ''!
print: object
	object printOn: self! !

!ZnHtmlOutputStream methodsFor: 'private html' stamp: ''!
renderAttributes: attributes
	attributes pairsDo: [ :name :value | 
		stream space; nextPutAll: name.
		value ifNotNil: [
			stream nextPutAll: '="'.
			self escapeAttributeValue: value asString.
			stream nextPut: $" ] ]! !

!ZnHtmlOutputStream methodsFor: 'streaming' stamp: ''!
space
	self nextPut: Character space! !

!ZnHtmlOutputStream methodsFor: 'html' stamp: ''!
str: string
	"Write string, escaping characters as needed for regular text"

	self escape: string! !

!ZnHtmlOutputStream methodsFor: 'html tags' stamp: ''!
tag: tag
	"Write a standalone, self closing HTML tag"
	
	"<tag/>"
	
	stream
		nextPut: $<;
		nextPutAll: tag;
		nextPut: $/;
		nextPut: $>! !

!ZnHtmlOutputStream methodsFor: 'html tags' stamp: ''!
tag: tag attributes: attributes
	"Write a standalone, self closing HTML tag with attributes, a collection of alternating keys and values"
	
	"<tag attr1=""value1"" ... attrN=""valueN""/>"

	stream
		nextPut: $<;
		nextPutAll: tag.
	self renderAttributes: attributes.
	stream
		nextPut: $/;
		nextPut: $>! !

!ZnHtmlOutputStream methodsFor: 'html tags' stamp: ''!
tag: tag attributes: attributePairs do: block
	"Write an HTML tag with attributes, a collection of alternating keys and values, executing block to generate enclosed content" 
	
	"<tag attr1=""value1"" ... attrN=""valueN""> ... </tag>"

	self openTag: tag attributes: attributePairs.
	block value.
	self closeTag: tag! !

!ZnHtmlOutputStream methodsFor: 'html tags' stamp: ''!
tag: tag attributes: attributePairs with: string
	"Write an HTML tag with attributes, a collection of alternating keys and values, using the escaped string as content" 
	
	"<tag attr1=""value1"" ... attrN=""valueN"">string</tag>"

	self openTag: tag attributes: attributePairs.
	self escape: string.
	self closeTag: tag ! !

!ZnHtmlOutputStream methodsFor: 'html tags' stamp: ''!
tag: tag class: cssClass do: block
	"Write an HTML tag with class cssClass, executing block to generate enclosed content" 
	
	"<tag class=""cssClass""> ... </tag>"

	self openTag: tag attribute: #class value: cssClass.
	block value.
	self closeTag: tag! !

!ZnHtmlOutputStream methodsFor: 'html tags' stamp: ''!
tag: tag class: cssClass with: string
	"Write an HTML tag with class cssClass, using the escaped string as content" 
	
	"<tag class=""cssClass"">string</tag>"

	self openTag: tag attribute: #class value: cssClass.
	self escape: string.
	self closeTag: tag ! !

!ZnHtmlOutputStream methodsFor: 'html tags' stamp: ''!
tag: tag do: block
	"Write an HTML tag, executing block to generate enclosed content" 
	
	"<tag>...</tag>"

	self openTag: tag.
	block value.
	self closeTag: tag! !

!ZnHtmlOutputStream methodsFor: 'html tags' stamp: ''!
tag: tag id: cssId do: block
	"Write an HTML tag with id cssId, executing block to generate enclosed content" 
	
	"<tag id=""cssId""> ... </tag>"

	self openTag: tag attribute: #id value: cssId.
	block value.
	self closeTag: tag! !

!ZnHtmlOutputStream methodsFor: 'html tags' stamp: ''!
tag: tag id: cssId with: string
	"Write an HTML tag with id cssId, using the escaped string as content" 
	
	"<tag id=""cssId"">string</tag>"

	self openTag: tag attribute: #id value: cssId.
	self escape: string.
	self closeTag: tag ! !

!ZnHtmlOutputStream methodsFor: 'html tags' stamp: ''!
tag: tag with: string
	"Write an HTML tag using the escaped string as content" 
	
	"<tag>string</tag>"

	self openTag: tag.
	self escape: string.
	self closeTag: tag ! !

!ZnHtmlOutputStream methodsFor: 'accessing' stamp: ''!
wrappedStream
	^ stream! !

Object subclass: #ZnLineReader
	instanceVariableNames: 'stream buffer position limit'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Streaming'!

!ZnLineReader commentStamp: '' prior: 0!
I am ZnLineReader and I help reading CRLF terminated lines from a stream.
I can work on ascii and binary streams doing conversions as necessary.
I limit lines to a maximum length.
I work with a reusable buffer.

Part of Zinc HTTP Components.!

!ZnLineReader class methodsFor: 'instance creation' stamp: ''!
on: stream
	^ self new
		on: stream;
		yourself! !

!ZnLineReader methodsFor: 'private' stamp: ''!
growBuffer
	| newBuffer |
	buffer size >= limit 
		ifTrue: [
			(ZnLineTooLong limit: limit) signal ].
	newBuffer := String new: buffer size * 2.
	newBuffer replaceFrom: 1 to: buffer size with: buffer.
	buffer := newBuffer ! !

!ZnLineReader methodsFor: 'accessing' stamp: ''!
limit: size
	limit := size! !

!ZnLineReader methodsFor: 'accessing' stamp: ''!
nextLine
	self reset.
	[ self processNext isNil ] whileFalse.
	^ buffer copyFrom: 1 to: position ! !

!ZnLineReader methodsFor: 'initialize-release' stamp: ''!
on: readStream
	stream := readStream.
	buffer := String new: 64.
	self 
		limit: ZnConstants maximumLineLength; 
		reset! !

!ZnLineReader methodsFor: 'private' stamp: ''!
processNext
	| item |
	stream atEnd ifTrue: [ ^ nil ].
	(item := stream next) ifNil: [ ^ nil ].
	stream isBinary ifTrue: [ item := item asCharacter ].
	item == Character lf
		ifTrue: [
			(position > 0 and: [ (buffer at: position) == Character cr ])
				ifTrue: [
					position := position - 1.
					^ nil ] ].
	self store: item.
	^ item
			
	! !

!ZnLineReader methodsFor: 'private' stamp: ''!
reset
	position := 0! !

!ZnLineReader methodsFor: 'private' stamp: ''!
store: item
	position >= buffer size ifTrue: [ self growBuffer ].
	position := position + 1.
	buffer at: position put: item		
	! !

Object subclass: #ZnMessage
	instanceVariableNames: 'headers entity'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Core'!

!ZnMessage commentStamp: '' prior: 0!
I am ZnMessage, an abstract class representing an HTTP Message.

Part of Zinc HTTP Components.!

!ZnMessage class methodsFor: 'instance creation' stamp: ''!
readBinaryFrom: stream
	^ self new
		readBinaryFrom: stream;
		yourself! !

!ZnMessage class methodsFor: 'instance creation' stamp: ''!
readFrom: stream
	^ self new
		readFrom: stream;
		yourself! !

!ZnMessage class methodsFor: 'instance creation' stamp: ''!
readHeaderFrom: stream
	^ self new
		readHeaderFrom: stream;
		yourself! !

!ZnMessage class methodsFor: 'instance creation' stamp: ''!
readStreamingFrom: stream
	^ self new
		readStreamingFrom: stream;
		yourself! !

!ZnMessage methodsFor: 'comparing' stamp: ''!
= other
	self class = other class ifFalse: [ ^ false ].
	^ self headers = other headers and: [ self entity = other entity ]! !

!ZnMessage methodsFor: 'accessing' stamp: ''!
clearEntity
	"Set my entity to nil. 
	Unlike #resetEntity:, my content type and length remain unchanged."

	entity ifNotNil: [
		entity close.
		entity := nil ]! !

!ZnMessage methodsFor: 'accessing' stamp: ''!
contentLength
	self hasEntity ifTrue: [ ^ self entity contentLength ].
	(self hasHeaders and: [ self headers hasContentLength ]) ifTrue: [ ^ self headers contentLength ].
	^ nil! !

!ZnMessage methodsFor: 'accessing' stamp: ''!
contentType
	self hasEntity ifTrue: [ ^ self entity contentType ].
	(self hasHeaders and: [ self headers hasContentType ]) ifTrue: [ ^ self headers contentType ].
	^ nil! !

!ZnMessage methodsFor: 'accessing' stamp: ''!
contents
	^ entity isNil ifTrue: [ nil ] ifFalse: [ entity contents ]! !

!ZnMessage methodsFor: 'accessing' stamp: ''!
cookies
	self subclassResponsibility ! !

!ZnMessage methodsFor: 'accessing' stamp: ''!
entity
	^ entity! !

!ZnMessage methodsFor: 'accessing' stamp: ''!
entity: object
	"Set my entity to object. Unless I already am describing an entity using my content type and length
	set my content type and length as defined by object."

	entity := object.
	self headers isDescribingEntity ifFalse: [
		self headers acceptEntityDescription: object ]! !

!ZnMessage methodsFor: 'private' stamp: ''!
entityReaderOn: stream
	^ ZnEntityReader new
		headers: self headers;
		stream: stream;
		yourself! !

!ZnMessage methodsFor: 'private' stamp: ''!
entityWriterOn: stream
	^ ZnEntityWriter new
		headers: self headers;
		stream: stream;
		yourself! !

!ZnMessage methodsFor: 'testing' stamp: ''!
hasEntity
	^ self entity notNil! !

!ZnMessage methodsFor: 'testing' stamp: ''!
hasHeaders
	^ headers notNil and: [ self headers isEmpty not ]! !

!ZnMessage methodsFor: 'testing' stamp: ''!
hasSession
	"Return if there currently is a server session.
	This only returns a value during #handleRequest:"

	^ ZnCurrentServerSession value notNil! !

!ZnMessage methodsFor: 'comparing' stamp: ''!
hash
	^ self headers hash bitXor: self entity hash! !

!ZnMessage methodsFor: 'accessing' stamp: ''!
headers
	headers isNil ifTrue: [ headers := ZnHeaders new ].
	^ headers! !

!ZnMessage methodsFor: 'accessing' stamp: ''!
headers: object
	headers := object! !

!ZnMessage methodsFor: 'enumerating' stamp: ''!
headersDo: twoArgumentBlock
	self hasHeaders
		ifTrue: [ self headers headersDo: twoArgumentBlock ]! !

!ZnMessage methodsFor: 'testing' stamp: ''!
isConnectionClose
	| value |
	value := self headers at: 'Connection' ifAbsent: [ ^ false ].
	^ value sameAs: 'close'! !

!ZnMessage methodsFor: 'testing' stamp: ''!
isConnectionKeepAlive
	| value |
	value := self headers at: 'Connection' ifAbsent: [ ^ false ].
	^ value sameAs: 'keep-alive'! !

!ZnMessage methodsFor: 'copying' stamp: ''!
postCopy
	headers := headers copy.
	"Note that we don't copy the entity, see also #resetEntity: and ZnClient>>#resetEntity"! !

!ZnMessage methodsFor: 'initialize-release' stamp: ''!
readBinaryFrom: stream
	| entityReader |
	self readHeaderFrom: stream.
	(entityReader := self entityReaderOn: stream)
		binary.
	self entity: entityReader readEntity! !

!ZnMessage methodsFor: 'initialize-release' stamp: ''!
readEntityFrom: stream
	self entity: (self entityReaderOn: stream) readEntity
	! !

!ZnMessage methodsFor: 'initialize-release' stamp: ''!
readFrom: stream
	self readHeaderFrom: stream.
	self readEntityFrom: stream
	! !

!ZnMessage methodsFor: 'initialize-release' stamp: ''!
readHeaderFrom: stream
	self headers: (ZnHeaders readFrom: stream)
	! !

!ZnMessage methodsFor: 'initialize-release' stamp: ''!
readStreamingFrom: stream
	| entityReader |
	self readHeaderFrom: stream.
	(entityReader := self entityReaderOn: stream)
		streaming.
	self entity: entityReader readEntity
	! !

!ZnMessage methodsFor: 'accessing' stamp: ''!
resetEntity: object
	"Set my entity to object. Always set my content type and length as defined by object,
	even if I am already describing an entity using my content type and length.
	See also: #clearEntity"

	(entity ~= object and: [ entity notNil ]) 
		ifTrue: [ entity close ].
	entity := object.
	self headers acceptEntityDescription: object! !

!ZnMessage methodsFor: 'accessing' stamp: ''!
server
	"Return the current server.
	This only returns a value during #handleRequest:"

	^ ZnCurrentServer value! !

!ZnMessage methodsFor: 'accessing' stamp: ''!
session
	"Return the current server session.
	This only returns a value during #handleRequest:"

	^ ZnCurrentServerSession value! !

!ZnMessage methodsFor: 'accessing' stamp: ''!
setConnectionClose
	self headers at: 'Connection' put: 'close'! !

!ZnMessage methodsFor: 'accessing' stamp: ''!
setConnectionKeepAlive
	self headers at: 'Connection' put: 'keep-alive'! !

!ZnMessage methodsFor: 'accessing' stamp: ''!
setContentType: mimeType
	"Explicitly set my Content-Type to mimeType.
	I normally do this automatically by taking the #contentType from my entity."
	
	self headers contentType: mimeType! !

!ZnMessage methodsFor: 'testing' stamp: ''!
wantsConnectionClose
	"Return if the HTTP protocol should close the connection after processing the receiver"
	
	^ false! !

!ZnMessage methodsFor: 'writing' stamp: ''!
writeOn: stream
	| bivalentWriteStream |
	bivalentWriteStream := ZnBivalentWriteStream on: stream.
	self headers writeOn: bivalentWriteStream.
	bivalentWriteStream nextPutAll: String crlf.
	self hasEntity ifTrue: [ 
		(self entityWriterOn: bivalentWriteStream) writeEntity: self entity ]! !

!ZnMessage methodsFor: 'writing' stamp: ''!
writeToTranscript
	"Write the receiver on the Transcript like when sent or received over the network.
	Useful for debugging, obviously"
	
	self writeOn: Transcript.
	Transcript flush! !

ZnMessage subclass: #ZnRequest
	instanceVariableNames: 'requestLine'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Core'!

!ZnRequest commentStamp: '' prior: 0!
I am ZnRequest, representing an HTTP Request, 
consisting of a request line, headers and an optional entity (body).
I am a ZnMessage.
I can be used for generating and parsing.

Part of Zinc HTTP Components.!

!ZnRequest class methodsFor: 'instance creation' stamp: ''!
delete: urlObject
	^ self method: #DELETE url: urlObject! !

!ZnRequest class methodsFor: 'instance creation' stamp: ''!
empty
	^ self new
		requestLine: ZnRequestLine empty;
		headers: ZnHeaders defaultRequestHeaders;
		yourself! !

!ZnRequest class methodsFor: 'instance creation' stamp: ''!
get: urlObject
	^ self method: #GET url: urlObject! !

!ZnRequest class methodsFor: 'instance creation' stamp: ''!
head: urlObject
	^ self method: #HEAD url: urlObject! !

!ZnRequest class methodsFor: 'instance creation' stamp: ''!
method: method url: urlObject
	| url |
	url := urlObject asZnUrl.
	^ self new
		requestLine: (ZnRequestLine method: method uri: url);
		headers: (ZnHeaders requestHeadersFor: url);
		yourself! !

!ZnRequest class methodsFor: 'instance creation' stamp: ''!
options: urlObject
	^ self method: #OPTIONS url: urlObject! !

!ZnRequest class methodsFor: 'instance creation' stamp: ''!
patch: urlObject
	^ self method: #PATCH url: urlObject! !

!ZnRequest class methodsFor: 'instance creation' stamp: ''!
post: urlObject
	^ self method: #POST url: urlObject! !

!ZnRequest class methodsFor: 'instance creation' stamp: ''!
put: urlObject
	^ self method: #PUT url: urlObject! !

!ZnRequest methodsFor: 'comparing' stamp: ''!
= other
	^ super = other and: [ self requestLine = other requestLine ]! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
absoluteUrl
	^ self host 
		ifNil: [ self url ]
		ifNotNil: [ :hostUrl | self url inContextOf: hostUrl ]! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
accept
	^ self headers at: 'Accept'! !

!ZnRequest methodsFor: 'testing' stamp: ''!
acceptsEncodingGzip
	| value |
	value := self headers at: 'Accept-Encoding' ifAbsent: [ ^ false ].
	^ value includesSubstring: 'gzip'! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
authorization
	^ self headers at: 'Authorization'! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
basicAuthentication
	| authorization separator |
	authorization := self authorization findTokens: ' '.
	(authorization size = 2 and: [ authorization first = 'Basic' ])
		ifFalse: [ ^ NotFound signal: 'Incomplete Basic Authentication' ].
	authorization := ZnUtils decodeBase64: authorization second.
	separator := authorization 
		indexOf: $: 
		ifAbsent: [ ^ NotFound signal: 'Incomplete Basic Authentication' ].
	^ { authorization first: separator - 1. authorization allButFirst: separator }! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
cookies
	| value cookies |
	value := self headers at: 'Cookie' ifAbsent: [ ^ #() ].
	cookies := value isString
		ifTrue: [ value findTokens: #($;) ]
		ifFalse: [ value ].
	^ cookies collect: [ :each | ZnCookie fromString: each ]! !

!ZnRequest methodsFor: 'comparing' stamp: ''!
hash
	^ super hash bitXor: self requestLine hash! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
host
	| host scheme |
	host := self headers at: 'Host' ifAbsent: [ ^ nil ].
	scheme :=ZnCurrentServer value ifNil: [ #http ] ifNotNil: [ :server | server scheme ].
	^ ZnUrl fromString: host defaultScheme: scheme! !

!ZnRequest methodsFor: 'testing' stamp: ''!
isHttp10
	^ self requestLine isHttp10 ! !

!ZnRequest methodsFor: 'testing' stamp: ''!
isHttp11
	^ self requestLine isHttp11! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
mergedFields
	"Return a new ZnMultiValueDictionary containing all query fields, if any,
	merged with all ZnApplicationFormUrlEncodedEntity fields, if any,
	preserving multi values for identical keys"

	| fields |
	fields := ZnMultiValueDictionary new.
	self uri hasQuery
		ifTrue: [ fields addAllMulti: self uri query ].
	(self contentType = ZnMimeType applicationFormUrlEncoded and: [ self hasEntity ])
		ifTrue: [ fields addAllMulti: self entity fields ].
	^ fields! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
method
	^ self requestLine method ! !

!ZnRequest methodsFor: 'initialize-release' stamp: ''!
method: method
	self requestLine method: method! !

!ZnRequest methodsFor: 'copying' stamp: ''!
postCopy
	super postCopy.
	requestLine := requestLine copy! !

!ZnRequest methodsFor: 'printing' stamp: ''!
printOn: stream
	super printOn: stream.
	stream nextPut: $(.
	self requestLine printMethodAndUriOn: stream.
	stream nextPut: $)! !

!ZnRequest methodsFor: 'initialize-release' stamp: ''!
readHeaderFrom: stream
	self requestLine: (ZnRequestLine readFrom: stream).
	super readHeaderFrom: stream
	! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
relativeUrl
	^ self url asRelativeUrl ! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
requestLine
	^ requestLine ! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
requestLine: object
	requestLine := object! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
session
	"Return the current server session.
	If necessary a new session is created.
	This only returns a value during #handleRequest:"

	^ super session
		ifNil: [ 
			| session |
			session := self server sessionFor: self.
			ZnCurrentServerSession value: session.
			session ]! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
setAccept: object
	self headers at: 'Accept' put: object asString! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
setAcceptEncodingGzip
	self headers at: 'Accept-Encoding' put: 'gzip'! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
setAuthorization: authorization
	self headers at: 'Authorization'  put: authorization! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
setBasicAuthenticationUsername: username password: password
	(username isNil | password isNil)
		ifTrue: [
			self headers removeKey: 'Authorization' ifAbsent: []]
		ifFalse: [
			self setAuthorization: 'Basic ', (ZnUtils encodeBase64: (username, ':', password)) ]
! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
setCookie: cookie
	self headers at: 'Cookie' put: cookie
! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
setIfModifiedSince: reference
	self headers at: 'If-Modified-Since' put: (ZnUtils httpDate: reference)! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
uri
	^ self requestLine uri ! !

!ZnRequest methodsFor: 'accessing' stamp: ''!
url
	^ self uri ! !

!ZnRequest methodsFor: 'initialize-release' stamp: ''!
url: url
	self requestLine uri: url.
	self headers request: self url! !

!ZnRequest methodsFor: 'testing' stamp: ''!
wantsConnectionClose
	"Return if the HTTP protocol should close the connection after processing the receiver.
	Overwritten to return true in case of the presense of an explicit connection close request header
	or the absense of a keep alive header when using the old HTTP 1.0 protocol."

	^ self isConnectionClose 
		or: [ self isHttp10 and: [ self isConnectionKeepAlive not ] ]! !

!ZnRequest methodsFor: 'writing' stamp: ''!
writeOn: stream
	| bivalentWriteStream |
	bivalentWriteStream := ZnBivalentWriteStream on: stream.
	self requestLine writeOn: bivalentWriteStream.
	super writeOn: bivalentWriteStream! !

ZnMessage subclass: #ZnResponse
	instanceVariableNames: 'statusLine'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Core'!

!ZnResponse commentStamp: '' prior: 0!
I am ZnResponse, representing an HTTP Response 
consisting of a status line, headers and an optional entity (body).
I am a ZnMessage.
I can be used for generating and parsing.

Part of Zinc HTTP Components.!

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
accepted
	^ self statusLine: ZnStatusLine accepted! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
badRequest: request
	| message |
	message := String streamContents: [ :out |
		out << 'Bad Request '; << request method; space.
		request uri printPathQueryFragmentOn: out ].
	^ self 
		badRequest: request 
		entity: (ZnEntity textCRLF: message)! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
badRequest: request entity: entity
	^ (self statusLine: ZnStatusLine badRequest)
		entity: entity;
		yourself! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
created: url
	^ self 
		created: url 
		entity: (ZnEntity textCRLF: 'Created ' , url asString)! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
created: url entity: entity
	^ (self statusLine: ZnStatusLine created)
		setLocation: url asString;
		entity: entity;
		yourself! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
methodNotAllowed: request
	| message |
	message := String streamContents: [ :out |
		out << 'Method Not Allowed '; << request method; space.
		request uri printPathQueryFragmentOn: out ].
	^ self 
		methodNotAllowed: request
		entity: (ZnEntity textCRLF: message)! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
methodNotAllowed: request entity: entity
	^ (self statusLine: ZnStatusLine methodNotAllowed)
		entity: entity;
		yourself! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
noContent
	^ self statusLine: ZnStatusLine noContent! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
notFound: url
	^ self
		notFound: url
		entity: (ZnEntity textCRLF: 'Not Found ' , url asZnUrl pathQueryFragmentPrintString)! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
notFound: url entity: entity
	^ (self statusLine: ZnStatusLine notFound)
		entity: entity;
		yourself! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
notModified
	^ self statusLine: ZnStatusLine notModified! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
ok: entity
	^ (self statusLine: ZnStatusLine ok)
		entity: entity;
		yourself! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
redirect: url
	^ self
		redirect: url
		entity: (ZnEntity textCRLF: 'Redirect ' , url asString)! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
redirect: url entity: entity
	^ (self statusLine: ZnStatusLine redirect)
		setLocation: url asString;
		entity: entity;
		yourself! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
serverError: string
	^ self serverErrorWithEntity: (ZnEntity textCRLF: string)! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
serverErrorWithEntity: entity
	^ (self statusLine: ZnStatusLine internalServerError)
		entity: entity;
		yourself! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
statusCode: httpResponeCode
	^ self statusLine: (ZnStatusLine code: httpResponeCode)! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
statusLine: statusLine
	^ self new
		statusLine: statusLine;
		headers: ZnHeaders defaultResponseHeaders;
		yourself! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
unauthorized
	^ self unauthorized: 'Basic realm=ZincHTTPComponents'! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
unauthorized: authString
	^ self
		unauthorized: authString
		entity: (ZnEntity textCRLF: 'Unauthorized')! !

!ZnResponse class methodsFor: 'instance creation' stamp: ''!
unauthorized: authString entity: entity
	^ (self statusLine: ZnStatusLine unauthorized)
		setWWWAuthenticate: authString;
		entity: entity;
		yourself! !

!ZnResponse methodsFor: 'comparing' stamp: ''!
= other
	^ super = other and: [ self statusLine = other statusLine ]! !

!ZnResponse methodsFor: 'accessing' stamp: ''!
addCookie: cookie
	self headers at: 'Set-Cookie' add: cookie fullString
! !

!ZnResponse methodsFor: 'accessing' stamp: ''!
code
	^ self statusLine code! !

!ZnResponse methodsFor: 'accessing' stamp: ''!
cookies
	| value cookies |
	value := self headers at: 'Set-Cookie' ifAbsent: [ ^ #() ].
	cookies := value isString
		ifTrue: [ Array with: value ]
		ifFalse: [ value ].
	^ cookies collect: [ :each | ZnCookie fromString: each ]! !

!ZnResponse methodsFor: 'private' stamp: ''!
entityReaderOn: stream
	"Reading undefined content is allowed for responses"
	
	^ (super entityReaderOn: stream)  
		allowReadingUpToEnd;
		yourself! !

!ZnResponse methodsFor: 'testing' stamp: ''!
hasContentEncoding
	^ self headers includesKey: 'Content-Encoding'! !

!ZnResponse methodsFor: 'testing' stamp: ''!
hasTransferEncoding
	^ self headers includesKey: 'Transfer-Encoding'! !

!ZnResponse methodsFor: 'comparing' stamp: ''!
hash
	^ super hash bitXor: self status hash! !

!ZnResponse methodsFor: 'testing' stamp: ''!
isAuthenticationRequired
	^ self code = 401 ! !

!ZnResponse methodsFor: 'testing' stamp: ''!
isBadRequest
	^ self code = 400! !

!ZnResponse methodsFor: 'testing' stamp: ''!
isCreated
	^ self code = 201! !

!ZnResponse methodsFor: 'testing' stamp: ''!
isError
	^self code > 399! !

!ZnResponse methodsFor: 'testing' stamp: ''!
isInformational
	^ self code between: 100 and: 199! !

!ZnResponse methodsFor: 'testing' stamp: ''!
isNoContent
	^ self code = 204! !

!ZnResponse methodsFor: 'testing' stamp: ''!
isNotFound
	^ self code = 404 ! !

!ZnResponse methodsFor: 'testing' stamp: ''!
isNotModified
	^ self code = 304! !

!ZnResponse methodsFor: 'testing' stamp: ''!
isRedirect
	^ #(301 302 303 307) includes: self code! !

!ZnResponse methodsFor: 'testing' stamp: ''!
isSuccess
	^ #(200 201 202 204) includes: self code! !

!ZnResponse methodsFor: 'accessing' stamp: ''!
location
	^ self headers at: 'Location'! !

!ZnResponse methodsFor: 'copying' stamp: ''!
postCopy
	super postCopy.
	statusLine := statusLine copy! !

!ZnResponse methodsFor: 'printing' stamp: ''!
printOn: stream
	super printOn: stream.
	stream nextPut: $(.
	self statusLine printCodeAndReasonOn: stream.
	self hasEntity ifTrue: [ 
		stream space.
		self entity printContentTypeAndLengthOn: stream ].
	stream nextPut: $)
	! !

!ZnResponse methodsFor: 'initialize-release' stamp: ''!
readEntityFrom: stream
	(self isInformational or: [ self isNoContent or: [ self isNotModified ] ])
		ifFalse: [ super readEntityFrom: stream ]! !

!ZnResponse methodsFor: 'initialize-release' stamp: ''!
readHeaderFrom: stream
	self statusLine: (ZnStatusLine readFrom: stream).
	super readHeaderFrom: stream
	! !

!ZnResponse methodsFor: 'accessing' stamp: ''!
setConnectionCloseFor: request
	request wantsConnectionClose 
		ifTrue: [ self setConnectionClose ]! !

!ZnResponse methodsFor: 'accessing' stamp: ''!
setContentEncodingGzip
	self headers at: 'Content-Encoding' put: 'gzip'! !

!ZnResponse methodsFor: 'accessing' stamp: ''!
setKeepAliveFor: request
	(request isHttp10 and: [ request isConnectionKeepAlive ])
		ifTrue: [ self setConnectionKeepAlive ]
	! !

!ZnResponse methodsFor: 'accessing' stamp: ''!
setLocation: uriString
	self headers at: 'Location' put: uriString! !

!ZnResponse methodsFor: 'accessing' stamp: ''!
setTransferEncodingChunked
	self headers 
		at: 'Transfer-Encoding' put: 'chunked';
		clearContentLength! !

!ZnResponse methodsFor: 'accessing' stamp: ''!
setWWWAuthenticate: string
	self headers at: 'WWW-Authenticate' put: string! !

!ZnResponse methodsFor: 'accessing' stamp: ''!
status
	^ self statusLine code! !

!ZnResponse methodsFor: 'accessing' stamp: ''!
statusLine
	^ statusLine! !

!ZnResponse methodsFor: 'accessing' stamp: ''!
statusLine: object
	statusLine := object! !

!ZnResponse methodsFor: 'accessing' stamp: ''!
useConnection: connection
	"Hook method that can be overwritten to give subclasses the chance to 
	keep using connection in the current thread/process after the server wrote the response"! !

!ZnResponse methodsFor: 'writing' stamp: ''!
writeOn: stream
	| bivalentWriteStream |
	bivalentWriteStream := ZnBivalentWriteStream on: stream.
	self statusLine writeOn: bivalentWriteStream.
	super writeOn: bivalentWriteStream! !

!ZnResponse methodsFor: 'accessing' stamp: ''!
wwwAuthenticate
	^ self headers at: 'WWW-Authenticate'! !

Object subclass: #ZnMimePart
	instanceVariableNames: 'headers entity'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Core'!

!ZnMimePart commentStamp: '' prior: 0!
I am ZnMimePart.
I hold headers and an optional entity.
I am used in MultiPartMimeEncoding.

Acknowledgement: some code borrowed from AJP.

Part of Zinc HTTP Components.!

!ZnMimePart class methodsFor: 'instance creation' stamp: ''!
fieldName: fieldName entity: entity
	^ self new
		setContentDisposition: 'form-data;name="', fieldName, '"';
		entity: entity;
		yourself! !

!ZnMimePart class methodsFor: 'instance creation' stamp: ''!
fieldName: fieldName fileName: fileName entity: entity
	"Pathnames are often silenty encoded using UTF-8,
	this is a no-op for ASCII, but will fail on Latin-1 and others"

	| encodedFileName |
	encodedFileName := fileName utf8Encoded asString.
	^ self new
		setContentDisposition: 'form-data;name="', fieldName, '";filename="', encodedFileName, '"';
		entity: entity;
		yourself! !

!ZnMimePart class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 9/27/2018 14:43'!
fieldName: fieldName fileNamed: fileName
	| mimeType size fileStream entity baseName |
	mimeType := ZnMimeType forFilenameExtension: (ZnFileSystemUtils extensionFor: fileName).
	fileStream := mimeType isBinary 
		ifTrue: [ ZnFileSystemUtils binaryReadStreamFor: fileName ] 
		ifFalse: [ ZnFileSystemUtils readStreamFor: fileName ].
	size := ZnFileSystemUtils fileSizeFor: fileName.
	(entity := ZnStreamingEntity type: mimeType length: size)
		stream: fileStream.
	baseName := ZnFileSystemUtils baseNameFor: fileName.
	^ self fieldName: fieldName fileName: baseName entity: entity! !

!ZnMimePart class methodsFor: 'instance creation' stamp: ''!
fieldName: fieldName value: fieldValue
	^ self
		fieldName: fieldName
		entity: (ZnEntity with: fieldValue)! !

!ZnMimePart class methodsFor: 'instance creation' stamp: ''!
readBinaryFrom: stream
	^ self new
		readBinaryFrom: stream;
		yourself! !

!ZnMimePart class methodsFor: 'instance creation' stamp: ''!
readFrom: stream
	^ self new
		readFrom: stream;
		yourself! !

!ZnMimePart methodsFor: 'comparing' stamp: ''!
= other
	self class = other class ifFalse: [ ^ false ].
	^ self headers = other headers and: [ self entity = other entity ]! !

!ZnMimePart methodsFor: 'accessing' stamp: ''!
contentDisposition
	^ self headers at: 'Content-Disposition' ifAbsent: [ nil ]! !

!ZnMimePart methodsFor: 'private' stamp: ''!
contentDispositionValues
	| contentDisposition values start semiColonIndex |
	(contentDisposition := self contentDisposition) isNil ifTrue: [ ^ #() ].
	values := OrderedCollection new: 3.
	start := 1.
	semiColonIndex := contentDisposition indexOf: $; startingAt: start ifAbsent: [ 0 ].
	[ semiColonIndex isZero ] whileFalse: [
		| value |
		value := contentDisposition copyFrom: start to: semiColonIndex - 1.
		values add: (ZnUtils trimString: value).
		start := semiColonIndex + 1.
		semiColonIndex := contentDisposition indexOf: $; startingAt: start ifAbsent: [ 0 ] ].
	values add: (ZnUtils trimString: (contentDisposition copyFrom: start to: contentDisposition size)).
	^ values! !

!ZnMimePart methodsFor: 'accessing' stamp: ''!
contentLength
	self hasEntity ifTrue: [ ^ self entity contentLength ].
	(self hasHeaders and: [ self headers hasContentLength ]) ifTrue: [ ^ self headers contentLength ].
	^ nil! !

!ZnMimePart methodsFor: 'accessing' stamp: ''!
contentType
	self hasEntity ifTrue: [ ^ self entity contentType ].
	(self hasHeaders and: [ self headers hasContentType ]) ifTrue: [ ^ self headers contentType ].
	^ nil! !

!ZnMimePart methodsFor: 'accessing' stamp: ''!
contents
	^ entity isNil ifTrue: [ nil ] ifFalse: [ entity contents ]! !

!ZnMimePart methodsFor: 'private' stamp: ''!
detectContentDispositionValue: aString
	self contentDispositionValues do: [ :each | 
		((each beginsWith: aString)
			and: [ each size > (aString size + 1)
			and: [ (each at: aString size + 1) = $= ] ]) ifTrue: [
				^ (each copyAfter: $=) withoutQuoting ] ].
	^ nil! !

!ZnMimePart methodsFor: 'accessing' stamp: ''!
entity
	^ entity! !

!ZnMimePart methodsFor: 'accessing' stamp: ''!
entity: object
	entity := object.
	self headers isDescribingEntity ifFalse: [
		self headers acceptEntityDescription: object ]! !

!ZnMimePart methodsFor: 'private' stamp: ''!
entityReaderOn: stream
	^ ZnEntityReader new
		headers: self headers;
		stream: stream;
		allowReadingUpToEnd;
		yourself! !

!ZnMimePart methodsFor: 'private' stamp: ''!
entityWriterOn: stream
	^ ZnEntityWriter new
		headers: self headers;
		stream: stream;
		yourself! !

!ZnMimePart methodsFor: 'accessing' stamp: ''!
fieldName
	^ self detectContentDispositionValue: 'name'! !

!ZnMimePart methodsFor: 'accessing' stamp: ''!
fieldValue
	^ self hasEntity 
		ifTrue: [ self entity contents ] 
		ifFalse: [ nil ] ! !

!ZnMimePart methodsFor: 'accessing' stamp: ''!
fieldValueString
	^ self fieldValue
		ifNil: [ String empty ]
		ifNotNil: [ :value | value asString ]! !

!ZnMimePart methodsFor: 'accessing' stamp: ''!
fileName
	"Pathnames are often silenty encoded using UTF-8,
	this is a no-op for ASCII, but will fail on Latin-1 and others"

	^ (self detectContentDispositionValue: 'filename')
		ifNotNil: [ :encodedFileName | encodedFileName asByteArray utf8Decoded ]! !

!ZnMimePart methodsFor: 'testing' stamp: ''!
hasEntity
	^ self entity notNil! !

!ZnMimePart methodsFor: 'testing' stamp: ''!
hasHeaders
	^ headers notNil and: [ self headers isEmpty not ]! !

!ZnMimePart methodsFor: 'comparing' stamp: ''!
hash
	^ self headers hash bitXor: self entity hash! !

!ZnMimePart methodsFor: 'accessing' stamp: ''!
headers
	headers isNil ifTrue: [ headers := ZnHeaders new ].
	^ headers! !

!ZnMimePart methodsFor: 'accessing' stamp: ''!
headers: object
	headers := object! !

!ZnMimePart methodsFor: 'enumerating' stamp: ''!
headersDo: twoArgumentBlock
	self hasHeaders
		ifTrue: [ self headers headersDo: twoArgumentBlock ]! !

!ZnMimePart methodsFor: 'copying' stamp: ''!
postCopy
	headers := headers copy.
	"Note that we don't copy the entity, see also #resetEntity: and ZnClient>>#resetEntity"! !

!ZnMimePart methodsFor: 'printing' stamp: ''!
printOn: stream
	super printOn: stream.
	stream nextPut: $(.
	self hasEntity ifTrue: [ 
		self entity printContentTypeAndLengthOn: stream ].
	stream nextPut: $)
	! !

!ZnMimePart methodsFor: 'initialize-release' stamp: ''!
readBinaryFrom: stream
	| entityReader |
	self headers: (ZnHeaders readFrom: stream).
	(entityReader := self entityReaderOn: stream)
		binary.
	self entity: entityReader readEntity! !

!ZnMimePart methodsFor: 'initialize-release' stamp: ''!
readFrom: stream
	self headers: (ZnHeaders readFrom: stream).
	self entity: (self entityReaderOn: stream) readEntity
	! !

!ZnMimePart methodsFor: 'accessing' stamp: ''!
setContentDisposition: value
	^ self headers at: 'Content-Disposition' put: value ! !

!ZnMimePart methodsFor: 'writing' stamp: ''!
writeOn: stream
	self headers writeOn: stream.
	stream nextPutAll: String crlf.
	self hasEntity ifTrue: [ 
		(self entityWriterOn: stream) writeEntity: self entity ]! !

Object subclass: #ZnNetworkingUtils
	instanceVariableNames: 'socketStreamClass secureSocketStreamClass sslSessionClass'
	classVariableNames: 'Default DefaultSocketStreamTimeout'
	poolDictionaries: ''
	category: 'Zinc-HTTP-Support'!

!ZnNetworkingUtils commentStamp: '' prior: 0!
I am ZnNetworkingUtils holding various utilities related to HTTP networking.

My class side is a general API.

My instance side is a factory.

ZnNetworkingUtils default socketStreamClass: SocketStream.
ZnNetworkingUtils default socketStreamClass: ZdcSocketStream.

Part of Zinc HTTP Components.!

!ZnNetworkingUtils class methodsFor: 'accessing' stamp: ''!
default
	^ Default ifNil: [ Default := self new ]! !

!ZnNetworkingUtils class methodsFor: 'accessing' stamp: ''!
default: instance
	Default := instance ! !

!ZnNetworkingUtils class methodsFor: 'constants' stamp: ''!
defaultSocketStreamTimeout
	"Global default timeout in seconds for SocketStream IO"
	
	^ DefaultSocketStreamTimeout! !

!ZnNetworkingUtils class methodsFor: 'constants' stamp: ''!
defaultSocketStreamTimeout: seconds
	"Global default timeout in seconds for SocketStream IO"
	
	^ DefaultSocketStreamTimeout := seconds! !

!ZnNetworkingUtils class methodsFor: 'class initialization' stamp: ''!
initialize
	"The default will try to use ZdcSocketStream and ZdcSecureSocketStream"
	 
	Default := self new.
	DefaultSocketStreamTimeout := 30! !

ZnNetworkingUtils initialize!

!ZnNetworkingUtils class methodsFor: 'converting' stamp: ''!
ipAddressToString: byteArray
	^ String streamContents: [ :stream |
			byteArray 
				do: [ :each | stream print: each ] 
				separatedBy: [ stream nextPut: $. ] ]! !

!ZnNetworkingUtils class methodsFor: 'constants' stamp: ''!
listenBacklogSize
	"Server socket backlog size (number of pending connection waiting to be accepted)"
	
	^ 32! !

!ZnNetworkingUtils class methodsFor: 'proxy' stamp: ''!
proxyAuthorizationHeaderValueToUrl: url
	"Answer the Proxy-Authorization header value for HTTP access to url, or nil if none"
	
	^ self default proxyAuthorizationHeaderValueToUrl: url  ! !

!ZnNetworkingUtils class methodsFor: 'proxy' stamp: ''!
proxyUrl
	"Answer the host:port URL the HTTP/S proxy"
	
	^ self default proxyUrl! !

!ZnNetworkingUtils class methodsFor: 'networking' stamp: ''!
secureSocketStreamOn: socket
	^ self default 
		secureSocketStreamOn: socket  ! !

!ZnNetworkingUtils class methodsFor: 'networking' stamp: ''!
serverSocketOn: port
	^ self default 
		serverSocketOn: port  ! !

!ZnNetworkingUtils class methodsFor: 'networking' stamp: ''!
serverSocketOn: port interface: address
	^ self default 
		serverSocketOn: port interface: address ! !

!ZnNetworkingUtils class methodsFor: 'proxy' stamp: ''!
shouldProxyUrl: url
	"Answer if we should proxy HTTP access to url."
	
	^ self default shouldProxyUrl: url  ! !

!ZnNetworkingUtils class methodsFor: 'constants' stamp: ''!
socketBufferSize
	"Size in bytes for Sockets and SocketStream IO buffers"

	^ 4096! !

!ZnNetworkingUtils class methodsFor: 'networking' stamp: ''!
socketStreamOn: socket
	^ self default 
		socketStreamOn: socket  ! !

!ZnNetworkingUtils class methodsFor: 'constants' stamp: ''!
socketStreamTimeout
	"Access the current timeout in seconds for SocketStream IO"
	
	^ ZnConnectionTimeout value 
		ifNil: [ self defaultSocketStreamTimeout ]! !

!ZnNetworkingUtils class methodsFor: 'networking' stamp: ''!
socketStreamToUrl: url
	^ self default
		socketStreamToUrl: url  ! !

!ZnNetworkingUtils methodsFor: 'accessing' stamp: ''!
bufferSize
	^ self class socketBufferSize! !

!ZnNetworkingUtils methodsFor: 'proxy' stamp: ''!
httpProxyPassword
	"Return the password for proxy authorization"
	
	^ NetworkSystemSettings proxyPassword ! !

!ZnNetworkingUtils methodsFor: 'proxy' stamp: ''!
httpProxyPort
	"Return the port of the proxy server to use"
	
	^ NetworkSystemSettings httpProxyPort! !

!ZnNetworkingUtils methodsFor: 'proxy' stamp: ''!
httpProxyServer
	"Return the hostname of the proxy server to use"
	
	^ NetworkSystemSettings httpProxyServer! !

!ZnNetworkingUtils methodsFor: 'proxy' stamp: ''!
httpProxyUser
	"Return the user for proxy authorization"
	
	^ NetworkSystemSettings proxyUser! !

!ZnNetworkingUtils methodsFor: 'initialization' stamp: 'SvenVanCaekenberghe 5/26/2018 17:18'!
initialize
	super initialize.
	self socketStreamClass: (Smalltalk globals at: #ZdcSocketStream ifAbsent: [ SocketStream ]).
	self secureSocketStreamClass: (Smalltalk globals at: #ZdcSecureSocketStream ifAbsent: [ nil ]).	
	self sslSessionClass: (Smalltalk globals at: #ZdcPluginSSLSession ifAbsent: [ nil ]).! !

!ZnNetworkingUtils methodsFor: 'proxy' stamp: ''!
isProxyAuthorizationRequired
	"Does the proxy require authorization ?"
	
	^ self isProxySet and: [ NetworkSystemSettings useNetworkAuthentification ]! !

!ZnNetworkingUtils methodsFor: 'proxy' stamp: ''!
isProxySet
	"Should a proxy be used ?"
	
	^ NetworkSystemSettings useHTTPProxy ! !

!ZnNetworkingUtils methodsFor: 'proxy' stamp: ''!
proxyAuthorizationHeaderValueToUrl: url
	"Answer the Proxy-Authorization header value for HTTP access to url, or nil if none"
	
	^ (self isProxyAuthorizationRequired and: [ self shouldProxyUrl: url ])
		ifTrue: [ 'Basic ', (ZnUtils encodeBase64: (self httpProxyUser, ':', self httpProxyPassword)) ] 
		ifFalse: [ nil ]  ! !

!ZnNetworkingUtils methodsFor: 'proxy' stamp: ''!
proxyUrl
	"System settings do currently not support HTTPS proxies."

	^ ZnUrl new
		scheme: #http;
		host: self httpProxyServer;
		port: self httpProxyPort;
		yourself! !

!ZnNetworkingUtils methodsFor: 'accessing' stamp: ''!
secureSocketStreamClass
	^ secureSocketStreamClass ifNil: [ 
		secureSocketStreamClass := Smalltalk globals at: #ZdcSecureSocketStream ifAbsent: [ nil ] ]! !

!ZnNetworkingUtils methodsFor: 'initialize-release' stamp: ''!
secureSocketStreamClass: anObject
	secureSocketStreamClass := anObject! !

!ZnNetworkingUtils methodsFor: 'public' stamp: ''!
secureSocketStreamOn: socket
	| stream |
	stream := self secureSocketStreamClass on: socket.
	self setSocketStreamParameters: stream.  
	^ stream! !

!ZnNetworkingUtils methodsFor: 'public' stamp: ''!
serverSocketOn: port
	| socket |
	socket := Socket newTCP.
	self setServerSocketOptions: socket.
	socket listenOn: port backlogSize: self class listenBacklogSize.
	socket isValid
		ifFalse: [ self error: 'Cannot create socket on port ', port printString ].
	^ socket! !

!ZnNetworkingUtils methodsFor: 'public' stamp: ''!
serverSocketOn: port interface: address
	| socket |
	socket := Socket newTCP.
	self setServerSocketOptions: socket.
	socket 
		listenOn: port 
		backlogSize: self class listenBacklogSize 
		interface: address.
	socket isValid
		ifFalse: [ self error: 'Cannot create socket on port ', port printString ].
	^ socket! !

!ZnNetworkingUtils methodsFor: 'private' stamp: ''!
setServerSocketOptions: socket
	socket
		setOption: 'TCP_NODELAY' value: 1;
		setOption: 'SO_SNDBUF' value: self class socketBufferSize;
		setOption: 'SO_RCVBUF' value: self class socketBufferSize ! !

!ZnNetworkingUtils methodsFor: 'private' stamp: ''!
setSocketStreamParameters: stream
	stream 
		binary;
		shouldSignal: true;
		autoFlush: false;
		bufferSize: self bufferSize;
		timeout: self timeout.
! !

!ZnNetworkingUtils methodsFor: 'proxy' stamp: ''!
shouldProxyUrl: url
	"Even when a proxy is set, some URLs should not be proxied.
	There is no setting for this in the image.
	We exclude localhost and explicit exceptions from being proxied."
	
	self isProxySet ifFalse: [ ^ false ].
	^ url isLocalHost not
		and: [ 
			[ (NetworkSystemSettings isAnExceptionFor: url) not ]
				on: MessageNotUnderstood
				do: [ true ] ]! !

!ZnNetworkingUtils methodsFor: 'accessing' stamp: ''!
socketStreamClass
	^ socketStreamClass ifNil: [ socketStreamClass := SocketStream ]! !

!ZnNetworkingUtils methodsFor: 'initialize-release' stamp: ''!
socketStreamClass: anObject
	socketStreamClass := anObject! !

!ZnNetworkingUtils methodsFor: 'public' stamp: ''!
socketStreamOn: socket
	| stream |
	stream := self socketStreamClass on: socket.
	self setSocketStreamParameters: stream.  
	^ stream! !

!ZnNetworkingUtils methodsFor: 'proxy' stamp: ''!
socketStreamToProxy
	^ self socketStreamToUrlDirectly: self proxyUrl! !

!ZnNetworkingUtils methodsFor: 'public' stamp: ''!
socketStreamToUrl: url
	url hasHost ifFalse: [ ZnMissingHost signal ].  
	^ (self shouldProxyUrl: url)
		ifTrue: [ self socketStreamToProxy ] 
		ifFalse: [ self socketStreamToUrlDirectly: url ]! !

!ZnNetworkingUtils methodsFor: 'public' stamp: ''!
socketStreamToUrlDirectly: url
	| stream address |
	address := NetNameResolver addressForName: url host timeout: self timeout. 
	stream := (self streamClassForScheme: url scheme) 
		openConnectionToHost: address
		port: url portOrDefault
		timeout: self timeout.  
	self setSocketStreamParameters: stream. 
	"note that for TLS/SSL, #connect should be called"
	^ stream! !

!ZnNetworkingUtils methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 5/26/2018 17:19'!
sslSessionClass
	^ sslSessionClass ifNil: [ 
		sslSessionClass := Smalltalk globals at: #ZdcPluginSSLSession ifAbsent: [ nil ] ]! !

!ZnNetworkingUtils methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 5/26/2018 17:17'!
sslSessionClass: anObject
	sslSessionClass := anObject! !

!ZnNetworkingUtils methodsFor: 'accessing' stamp: ''!
streamClassForScheme: scheme
	(#(http ws) includes: scheme) ifTrue: [ 
		^ self socketStreamClass ].
	(#(https wss) includes: scheme) ifTrue: [ 
		^ self secureSocketStreamClass ifNil: [
			self error: 'No secure socket stream class set or available' ] ].
	(ZnUnknownScheme scheme: scheme) signal! !

!ZnNetworkingUtils methodsFor: 'accessing' stamp: ''!
timeout
	^ self class socketStreamTimeout! !

Object subclass: #ZnRequestLine
	instanceVariableNames: 'method uri version'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Core'!

!ZnRequestLine commentStamp: '' prior: 0!
I am an HTTP Request Line.
I consist of an HTTP method, an URI and an HTTP protocol version.
I can be used for generating and parsing.

Part of Zinc HTTP Components.!

!ZnRequestLine class methodsFor: 'instance creation' stamp: ''!
empty
	^ self new
		version: ZnConstants defaultHTTPVersion;
		yourself ! !

!ZnRequestLine class methodsFor: 'instance creation' stamp: ''!
method: method uri: uri
	^ self new
		method: method;
		uri: uri;
		version: ZnConstants defaultHTTPVersion;
		yourself ! !

!ZnRequestLine class methodsFor: 'instance creation' stamp: ''!
readFrom: stream
	^ self new
		readFrom: stream;
		yourself! !

!ZnRequestLine methodsFor: 'comparing' stamp: ''!
= other
	self class = other class ifFalse: [ ^ false ].
	^ self method = other method
		and: [ self uri asRelativeUrl = other uri asRelativeUrl
			and: [ self version = other version ] ]! !

!ZnRequestLine methodsFor: 'comparing' stamp: ''!
hash
	^ (self method hash bitXor: self uri hash) bitXor: self version hash! !

!ZnRequestLine methodsFor: 'testing' stamp: ''!
isHttp10
	^ self version = 'HTTP/1.0'! !

!ZnRequestLine methodsFor: 'testing' stamp: ''!
isHttp11
	^ self version = 'HTTP/1.1'! !

!ZnRequestLine methodsFor: 'accessing' stamp: ''!
method
	^ method! !

!ZnRequestLine methodsFor: 'accessing' stamp: ''!
method: string
	method := ZnConstants knownHTTPMethods 
		detect: [ :each | each = string ] 
		ifNone: [ (ZnUnknownHttpMethod method: string) signal ] ! !

!ZnRequestLine methodsFor: 'copying' stamp: ''!
postCopy
	uri := uri copy! !

!ZnRequestLine methodsFor: 'printing' stamp: ''!
printMethodAndUriOn: stream
	(self method notNil and: [ self uri notNil ])
		ifFalse: [ ^ self ].
	stream
		nextPutAll: self method;
		space.
	self uri printPathQueryFragmentOn: stream! !

!ZnRequestLine methodsFor: 'printing' stamp: ''!
printOn: stream
	super printOn: stream.
	(self method notNil and: [ self uri notNil ])
		ifFalse: [ ^ self ].
	stream nextPut: $(.
	self printMethodAndUriOn: stream.
	stream nextPut: $)! !

!ZnRequestLine methodsFor: 'initialize-release' stamp: ''!
readFrom: stream
	| line lineStream |
	line := (ZnLineReader on: stream) nextLine.
	lineStream := line readStream.
	self method: (lineStream upTo: Character space).
	self uri: (lineStream upTo: Character space).
	self version: (lineStream upToEnd)! !

!ZnRequestLine methodsFor: 'accessing' stamp: ''!
uri
	^ uri! !

!ZnRequestLine methodsFor: 'accessing' stamp: ''!
uri: anURI
	uri := anURI asZnUrl.
	uri enforceKnownScheme! !

!ZnRequestLine methodsFor: 'accessing' stamp: ''!
uriPathQueryFragment
	^ String streamContents: [ :stream | 
			self uri printPathQueryFragmentOn: stream ]! !

!ZnRequestLine methodsFor: 'accessing' stamp: ''!
version
	^ version! !

!ZnRequestLine methodsFor: 'accessing' stamp: ''!
version: string
	version := ZnConstants knownHTTPVersions 
		detect: [ :each | each = string ] 
		ifNone: [ (ZnUnknownHttpVersion version: string) signal ] ! !

!ZnRequestLine methodsFor: 'writing' stamp: ''!
writeOn: stream
	| writeStream |
	writeStream := ZnBivalentWriteStream on: stream.
	writeStream nextPutAll: self method; space.
	(ZnNetworkingUtils shouldProxyUrl: self uri)
		ifTrue: [ 
			self uri hasSecureScheme 
				ifTrue: [
					"See also ZnClient>>#setupTLSTo:"
					method = #CONNECT
						ifTrue: [ writeStream nextPutAll: uri authorityWithPort ]
						ifFalse: [ self uri printPathQueryFragmentOn: writeStream ] ]
				ifFalse: [ self uri printOn: writeStream ] ]
		ifFalse: [ self uri printPathQueryFragmentOn: writeStream ].
	writeStream space; nextPutAll: self version; nextPutAll: String crlf ! !

Object subclass: #ZnServer
	instanceVariableNames: 'options sessionManager'
	classVariableNames: 'AlwaysRestart ManagedServers'
	poolDictionaries: ''
	category: 'Zinc-HTTP-Client-Server'!

!ZnServer commentStamp: '' prior: 0!
I am ZnServer, an abstract superclass of HTTP Servers.
I am a facade for controlling a default server implementation.

I delegate my public class protocol methods to #defaultServerClass.

  ZnServer startDefaultOn: 1701.
  ZnClient new get: 'http://localhost:1701'.

Subclasses can register with me to have start/stop sent to them on System startUp/shutDown.
The default server instance will be registered automatically when it is started.

Part of Zinc HTTP Components.!

!ZnServer class methodsFor: 'public' stamp: ''!
adoptAsDefault: server
	"Adopt server as the default instance that we manage.
	If there was a previous default, stop it.
	Delegate to the default server class."
	
	self defaultServerClass adoptAsDefault: server! !

!ZnServer class methodsFor: 'accessing' stamp: ''!
alwaysRestart
	^ AlwaysRestart ifNil: [ AlwaysRestart := true ]! !

!ZnServer class methodsFor: 'accessing' stamp: ''!
alwaysRestart: boolean
	"Set if managed servers should always restart on image save, as opposed to only when quiting."

	^ AlwaysRestart := boolean! !

!ZnServer class methodsFor: 'public' stamp: ''!
default
	"Return the default instance that we manage.
	Delegate to the default server class."
	
	^ self defaultServerClass default! !

!ZnServer class methodsFor: 'public' stamp: ''!
defaultOn: portNumber
	"Return the default instance on a given port,
	Keep a reference to it in a class instance variable.
	If there was no previous default instance, create a new one.
	If there was a previous default instance, reuse it:
	if it was running stop it, change the port if necessary.
	Delegate to the default server class."

	^ self defaultServerClass defaultOn: portNumber ! !

!ZnServer class methodsFor: 'accessing' stamp: ''!
defaultServerClass
	"Return the default ZnServer subclass to use"
	
	^ ZnManagingMultiThreadedServer ! !

!ZnServer class methodsFor: 'class initialization' stamp: ''!
initialize
	ManagedServers := IdentitySet new.
	AlwaysRestart := true.
	self environment 
		at: #SessionManager 
		ifPresent: [ :manager | manager default registerNetworkClassNamed: self name ]
		ifAbsent: [ Smalltalk addToStartUpList: self; addToShutDownList: self ]
	! !

ZnServer initialize!

!ZnServer class methodsFor: 'accessing' stamp: ''!
managedServers
	^ ManagedServers ifNil: [ ManagedServers := IdentitySet new ]! !

!ZnServer class methodsFor: 'public' stamp: ''!
on: portNumber
	"Instantiate a new listener on a given port,
	send #start to it to start listening.
	Delegate to the default server class."

	^ self defaultServerClass on: portNumber ! !

!ZnServer class methodsFor: 'system startup' stamp: ''!
register: server
	"Arrange for server to be sent start/stop on system startUp/shutDown"
	
	self managedServers add: server! !

!ZnServer class methodsFor: 'system startup' stamp: ''!
shutDown: quiting
	"Our system shutDown hook: stop all servers we manage"
	
	(quiting or: [ self alwaysRestart ])
		ifTrue: [
			self managedServers do: [ :each | each stop: false ] ]! !

!ZnServer class methodsFor: 'public' stamp: ''!
startDefaultOn: portNumber
	"Start and return the default instance on a given port.
	Keep a reference to it in a class instance variable.
	If there was no previous default instance, create a new one.
	If there was a previous default instance, reuse it:
	if it was running stop and start it, effectively restarting it.
	Change the port if necessary.
	Starting the default server will register it automatically.
	Delegate to the default server class."

	^ self defaultServerClass startDefaultOn: portNumber ! !

!ZnServer class methodsFor: 'public' stamp: ''!
startOn: portNumber
	"Instanciate and return a new listener on a given port and start listening.
	Delegate to the default server class."

	^ self defaultServerClass startOn: portNumber ! !

!ZnServer class methodsFor: 'system startup' stamp: ''!
startUp: resuming
	"Our system startUp hook: start all servers we manage.
	We do this using deferred startup actions to allow normal error handling."

	(resuming or: [ self alwaysRestart ])
		ifTrue: [
			self environment 
				at: #SessionManager 
				ifPresent: [ :manager | 
					manager default currentSession addDeferredStartupAction: [ 
						self managedServers do: [ :each | each start ] ] ]
				ifAbsent: [ 
					Smalltalk addDeferredStartupAction: [ 
						self managedServers do: [ :each | each start ] ] ] ]! !

!ZnServer class methodsFor: 'public' stamp: ''!
stopDefault
	"Stop the default instance, if any, and clear it.
	Return the stopped instance, if any.
	Delegate to the default server class."

	^ self defaultServerClass stopDefault ! !

!ZnServer class methodsFor: 'system startup' stamp: ''!
unregister: server
	"No longer send server start/stop on system startUp/shutDown"
	
	self managedServers remove: server ifAbsent: [ ]! !

!ZnServer methodsFor: 'options' stamp: ''!
authenticator
	"Return the optional authenticator,
	the object that will be sent #authenticateRequest:do:
	to authenticate or refuse the requests. 
	When authentication succeeds, the block should be executed,
	when authentication fails, a appropriate response should be returned.
	If there is no authenticator, all requests will pass"
	
	^ self optionAt: #authenticator ifAbsent: [ nil ]! !

!ZnServer methodsFor: 'options' stamp: ''!
authenticator: object
	"Set the object that will be sent #authenticateRequest:do:
	to authenticate or refuse the requests. Can be nil.
	When authentication succeeds, the block should be executed,
	when authentication fails, a appropriate response should be returned"

	self optionAt: #authenticator put: object! !

!ZnServer methodsFor: 'options' stamp: ''!
bindingAddress
	"Return the interface address we are (or will be) listening on.
	Nil means that we are (or will be) listening on all/any interfaces."
	
	^ self optionAt: #bindAddress ifAbsent: [ nil ] ! !

!ZnServer methodsFor: 'options' stamp: ''!
bindingAddress: address
	"Set the interface address we will be listening on.
	Specify nil to listen on all/any interfaces, the default.
	Address must be a 4 element ByteArray, like #[127 0 0 1].
	Cannot be changed after the server is already running."
	
	self optionAt: #bindAddress put: address! !

!ZnServer methodsFor: 'options' stamp: ''!
debugMode
	"Return whether we are in debug mode, the default is false."
	
	^ self optionAt: #debugMode ifAbsent: [ false ] ! !

!ZnServer methodsFor: 'options' stamp: ''!
debugMode: boolean
	"Set my debug mode, the default being false.
	In debug mode, Smalltalk Error during #handleRequest: will raise a Debugger.
	When not in debug mode, a Smalltalk Error during #handleRequest: will result in an HTTP Server Error response."
	
	^ self optionAt: #debugMode put: boolean! !

!ZnServer methodsFor: 'options' stamp: ''!
defaultEncoder
	"The default character encoder to use when none is set in a mime-type"
	
	^ self optionAt: #defaultEncoder ifAbsent: [ ZnDefaultCharacterEncoder value ] ! !

!ZnServer methodsFor: 'options' stamp: ''!
defaultEncoder: encoder
	"Set the default character encoder to use when none is set in a mime-type"
	
	^ self optionAt: #defaultEncoder put: encoder asZnCharacterEncoder! !

!ZnServer methodsFor: 'options' stamp: ''!
delegate
	"Return the optional delegate, 
	the object that will be sent #handleRequest: to handle a request and produce a response.
	The default delegate is ZnDefaultServerDelegate"
	
	^ self optionAt: #delegate ifAbsentPut: [ ZnDefaultServerDelegate new ]! !

!ZnServer methodsFor: 'options' stamp: ''!
delegate: object
	"Set the delegate to object. Can be nil.
	This will be sent #handleRequest: to handle a request and produce a response"
	
	self optionAt: #delegate put: object! !

!ZnServer methodsFor: 'testing' stamp: ''!
isListening
	"Return true when I have a valid server socket listening at the correct port"
	
	self subclassResponsibility ! !

!ZnServer methodsFor: 'testing' stamp: ''!
isRunning
	"Return true when I am running"
	
	self subclassResponsibility ! !

!ZnServer methodsFor: 'accessing' stamp: ''!
localUrl
	"Return a ZnUrl to access me."
	
	^ ZnUrl new
		scheme: self scheme;
		host: NetNameResolver loopBackName;
		port: self port;
		yourself! !

!ZnServer methodsFor: 'options' stamp: ''!
logServerErrorDetails
	"Return whether we log server error details including a short strack trace, the default is true."
	
	^ self optionAt: #debugMode ifAbsent: [ true ] ! !

!ZnServer methodsFor: 'options' stamp: ''!
logServerErrorDetails: boolean
	"Set whether we log server error details including a short strack trace, the default is true."
	
	self optionAt: #debugMode put: boolean! !

!ZnServer methodsFor: 'options' stamp: ''!
maximumEntitySize
	"Return the maximum entity size in bytes that I will read from a stream before signalling ZnEntityTooLarge"
	
	^ self optionAt: #maximumEntitySize ifAbsent: [ ZnMaximumEntitySize value ]! !

!ZnServer methodsFor: 'options' stamp: ''!
maximumEntitySize: integer
	"Set the maximum entity size in bytes that I will read from a stream before signalling ZnEntityTooLarge"
	
	^ self optionAt: #maximumEntitySize put: integer! !

!ZnServer methodsFor: 'options' stamp: ''!
maximumNumberOfDictionaryEntries
	"Return the maximum number of entries allowed in ZnMutliValueDictionaries before signalling ZnTooManyDictionaryEntries. This protects us from overflow attacks."
		
	^ self
		optionAt: #maximumNumberOfDictionaryEntries
		ifAbsent: [ ZnMaximumNumberOfDictionaryEntries value ]! !

!ZnServer methodsFor: 'options' stamp: ''!
maximumNumberOfDictionaryEntries: anInteger
	"Set the maximum number of entries allowed in ZnMutliValueDictionaries before signalling ZnTooManyDictionaryEntries. This protects us from overflow attacks."
	
	^ self
		optionAt: #maximumNumberOfDictionaryEntries
		put: anInteger! !

!ZnServer methodsFor: 'accessing' stamp: ''!
optionAt: key ifAbsent: block
	"Return my option/settings stored under key.
	Execute block if I have no such option/setting.
	This is a generic interface, see my options protocol for specific usages."

	options ifNil: [ ^ block value ].
	^ options at: key ifAbsent: block! !

!ZnServer methodsFor: 'accessing' stamp: ''!
optionAt: key ifAbsentPut: block
	"Return my option/settings stored under key.
	If I have no such option/setting, store the result of evaluating block as new value and return it.
	This is a generic interface, see my options protocol for specific usages."

	^ options at: key ifAbsentPut: block! !

!ZnServer methodsFor: 'accessing' stamp: ''!
optionAt: key put: value
	"Set my option/setting identified by key to be value.
	This is a generic interface, see my options protocol for specific usages."
	
	options ifNil: [ options := Dictionary new ].
	options at: key put: value! !

!ZnServer methodsFor: 'options' stamp: ''!
port
	"Return the integer port number we are (or will be) listening on"
	
	^ self optionAt: #port ifAbsent: [ 1701 ] ! !

!ZnServer methodsFor: 'options' stamp: ''!
port: integer
	"Set the port number we will be listening on.
	Cannot be changed after the server is already running."
	
	(self isRunning and: [ self port ~= integer ]) 
		ifTrue: [ self error: 'Stop me before changing my port' ].
	self optionAt: #port put: integer! !

!ZnServer methodsFor: 'accessing' stamp: ''!
process
	"Return the process that is running my main listening loop.
	Will be nil when I am not running"
	
	self subclassResponsibility ! !

!ZnServer methodsFor: 'options' stamp: ''!
reader
	"Return a block that when given a stream reads an entity from it."
	
	^ self optionAt: #reader ifAbsentPut: [ [ :stream | ZnRequest readFrom: stream ] ]! !

!ZnServer methodsFor: 'options' stamp: ''!
reader: block
	"Customize how entities are read from a stream, see #reader"
	
	self optionAt: #reader put: block! !

!ZnServer methodsFor: 'public' stamp: ''!
register
	"Ask for start/stop to be sent to me on System startUp/shutDown"
	
	self class register: self! !

!ZnServer methodsFor: 'options' stamp: ''!
route
	"Return the route of the server.
	This is a short identification string to be appended at the end of server session ids, separated by a dot.
	Routes are used by load balancers and proxies to correctly implement session affiinity or stickyness.
	The default is nil, meaning that no route has to be appended."
	
	^ self optionAt: #route ifAbsent: [ nil ]! !

!ZnServer methodsFor: 'options' stamp: ''!
route: object
	"Set the route of the server.
	This is a short identification string to be appended at the end of server session ids, separated by a dot.
	Routes are used by load balancers and proxies to correctly implement session affiinity or stickyness.
	The default is nil, meaning that no route has to be appended."

	self optionAt: #route put: object! !

!ZnServer methodsFor: 'accessing' stamp: ''!
scheme
	^ #http! !

!ZnServer methodsFor: 'accessing' stamp: ''!
serverSocket
	"Return the server socket that I am using.
	Will be nil when I am not running"
	
	self subclassResponsibility ! !

!ZnServer methodsFor: 'options' stamp: ''!
serverUrl
	"Return the explicitely set external server URL, if any. Defaults to nil."
	
	^ self optionAt: #serverUrl ifAbsent: [ nil ]! !

!ZnServer methodsFor: 'options' stamp: ''!
serverUrl: urlObject
	"Set the explicit external server URL to urlObject. Defaults to nil.
	urlObject should be a ZnUrl or a String that parses correctly to one.
	See also #url."
	
	^ self optionAt: #serverUrl put: urlObject asZnUrl! !

!ZnServer methodsFor: 'accessing' stamp: ''!
sessionFor: request
	"Bind an existing session to request or create a new session"
	
	^ self sessionManager sessionFor: request! !

!ZnServer methodsFor: 'accessing' stamp: ''!
sessionManager
	"Return my session manager"

	^ sessionManager ifNil: [ sessionManager := ZnServerSessionManager new ]! !

!ZnServer methodsFor: 'public' stamp: ''!
start
	"Start me. I will start listening on my port for incoming HTTP connections.
	If I am running, I will first stop and thus effectively restart"
	
	self subclassResponsibility! !

!ZnServer methodsFor: 'public' stamp: ''!
stop
	"Stop me. I will stop listening on my port for incoming HTTP connections.
	Does nothing when I am not running"
	
	self stop: true ! !

!ZnServer methodsFor: 'public' stamp: ''!
stop: unregister
	"Stop me. I will stop listening on my port for incoming HTTP connections.
	If unregister is true, unregister me from the list of managed instances.
	Does nothing when I am not running"
	
	self subclassResponsibility ! !

!ZnServer methodsFor: 'public' stamp: ''!
unregister
	"Ask for start/stop to no longer be sent to me on System startUp/shutDown"

	self class unregister: self! !

!ZnServer methodsFor: 'accessing' stamp: ''!
url
	"Return the base external URL (a new ZnUrl instance) to access me.
	This defaults to #localUrl but can be set explicitely using the #serverUrl option.
	Missing elements from #serverUrl are merged in from #localUrl."

	^ self serverUrl 
		ifNil: [ self localUrl ] 
		ifNotNil: [ :serverUrl | serverUrl inContextOf: self localUrl ]! !

!ZnServer methodsFor: 'options' stamp: ''!
useGzipCompressionAndChunking
	"Return whether we should try to use gzip content encoding and chunked transfer encoding, the default is false."
	
	^ self optionAt: #useGzipCompressionAndChunking ifAbsent: [ false ] ! !

!ZnServer methodsFor: 'options' stamp: ''!
useGzipCompressionAndChunking: boolean
	"Set whether we should try to use gzip content encoding and chunked transfer encoding, the default being false."
	
	self optionAt: #useGzipCompressionAndChunking put: boolean! !

ZnServer subclass: #ZnSingleThreadedServer
	instanceVariableNames: 'process serverSocket logLevel'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Client-Server'!

ZnSingleThreadedServer class
	instanceVariableNames: 'default'!

!ZnSingleThreadedServer commentStamp: '' prior: 0!
I am ZnSingleThreadedServer.
I can be customized with a delegate (#handleRequest:) and an authenticator (#authenticateRequest:do:).

  ZnSingleThreadedServer startDefaultOn: 1701.
  ZnSingleThreadedServer default authenticator: (ZnBasicAuthenticator username: 'foo' password: 'secret').
  ZnClient new username: 'foo' password: 'secret'; get: 'http://localhost:1701'.

I use ZnDefaultServerDelegate when no other delegate is set.
I am single threaded, I run in a single process.
I close connections after each request/response cycle.

Part of Zinc HTTP Components.!

ZnSingleThreadedServer class
	instanceVariableNames: 'default'!

ZnSingleThreadedServer subclass: #ZnMultiThreadedServer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Client-Server'!

!ZnMultiThreadedServer commentStamp: '' prior: 0!
I am ZnMultiThreadedServer.
I inherit most features from ZnSingleThreadedServer.

  ZnMultiThreadedServer startDefaultOn: 1701.
  ZnMultiThreadedServer default authenticator: (ZnBasicAuthenticator username: 'foo' password: 'secret').
  ZnClient new username: 'foo' password: 'secret'; get: 'http://localhost:1701'.

I am multi threaded, I fork a new process for each incoming request.
I try to keep connections alive in each process.

Part of Zinc HTTP Components.!

ZnMultiThreadedServer subclass: #ZnManagingMultiThreadedServer
	instanceVariableNames: 'lock connections'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Client-Server'!

!ZnManagingMultiThreadedServer commentStamp: '' prior: 0!
ZnManagingMultiThreadedServer is a ZnMultiThreadedServer that manages the socket stream connections used by the worker threads that are spawned.

  ZnManagingMultiThreadedServer startDefaultOn: 1701.

I keep track of all the connections that I spawn so that I can close them when I am stopped.

Part of Zinc HTTP Components.!

!ZnManagingMultiThreadedServer methodsFor: 'private' stamp: ''!
closeConnections
	self connections isEmpty ifTrue: [ ^ self ].
	self lock critical: [
		self connections do: [ :each | 
			self logConnectionClosed: each.
			[ each close ] 
				on: Exception 
				do: [ ] ].
		self connections removeAll ]! !

!ZnManagingMultiThreadedServer methodsFor: 'private' stamp: ''!
closeSocketStream: socketStream
	super closeSocketStream: socketStream.
	self lock critical: [ 
		self connections remove: socketStream ifAbsent: [ ] ]! !

!ZnManagingMultiThreadedServer methodsFor: 'accessing' stamp: ''!
connections
	^ connections ifNil: [ connections := OrderedCollection new ]! !

!ZnManagingMultiThreadedServer methodsFor: 'accessing' stamp: ''!
lock
	^ lock ifNil: [ lock := Mutex new ]! !

!ZnManagingMultiThreadedServer methodsFor: 'private' stamp: ''!
socketStreamOn: socket
	| socketStream |
	socketStream := super socketStreamOn: socket.
	self lock critical: [ 
		self connections add: socketStream ].
	^ socketStream! !

!ZnManagingMultiThreadedServer methodsFor: 'public' stamp: ''!
stop: unregister
	super stop: unregister.
	self closeConnections! !

!ZnMultiThreadedServer methodsFor: 'request handling' stamp: ''!
augmentResponse: response forRequest: request
	"Our handler has produced response for request, manipulate the response before writing it"

	super augmentResponse: response forRequest: request.
	response setKeepAliveFor: request.
	response setConnectionCloseFor: request! !

!ZnMultiThreadedServer methodsFor: 'private' stamp: ''!
closeSocketStream: socketStream
	[ socketStream close ] 
		on: Exception 
		do: [ ]! !

!ZnMultiThreadedServer methodsFor: 'private' stamp: ''!
exceptionSet: classNames
	^ classNames
		inject: ExceptionSet new
		into: [ :set :each |
			(Smalltalk includesKey: each)
				ifTrue: [ set add: (Smalltalk at: each); yourself ]
				ifFalse: [ set ] ]! !

!ZnMultiThreadedServer methodsFor: 'request handling' stamp: ''!
executeOneRequestResponseOn: stream
	"Execute one HTTP request / response cycle on stream in 3 steps
	#readRequest: #handleRequest: and #writeResponse:on:
	Return true when the outer loop we are in should stop."

	| request response timing|
	timing := ZnServerTransactionTiming new.
	(request := self readRequestSafely: stream timing: timing)
		ifNil: [ ^ true ]
		ifNotNil: [ 
			response := self handleRequest: request timing: timing.
			self augmentResponse: response forRequest: request.
			self writeResponseSafely: response on: stream timing: timing.
			self logServerTransactionRequest: request response: response timing: timing.
			response useConnection: stream ].
	^ request wantsConnectionClose or: [ response wantsConnectionClose ]! !

!ZnMultiThreadedServer methodsFor: 'request handling' stamp: ''!
executeRequestResponseLoopOn: stream
	"Execute the HTTP request / response loop on stream one cycle at a time
	until the other end indicates it wants to stop or times out"

	ZnCurrentServer 
		value: self 
		during: [ 
			[ self executeOneRequestResponseOn: stream ] whileFalse ]! !

!ZnMultiThreadedServer methodsFor: 'request handling' stamp: ''!
listenLoop
	"We create a listening Socket, then wait for a connection.
	After each connection we also check that the listening Socket is still valid 
	- if not we just make a recursive call to this method to start over."

	self initializeServerSocket.
	[ [ 
		serverSocket isValid 
			ifFalse: [
				"will trigger #ifCurtailed: block and destroy socket"
				^ self listenLoop ].
		self serveConnectionsOn: serverSocket ] repeat ]

		ifCurtailed: [ self releaseServerSocket ]! !

!ZnMultiThreadedServer methodsFor: 'private' stamp: ''!
readRequestBadExceptionSet
	"Return the set of exceptions which, when they occur while reading a request,
	are interpreted as equivalent to a request parse error or bad request."

	^ self
		exceptionSet:
			#(#ZnParseError #ZnCharacterEncodingError #ZnUnknownScheme #ZnPortNotANumber #ZnTooManyDictionaryEntries #ZnEntityTooLarge)! !

!ZnMultiThreadedServer methodsFor: 'request handling' stamp: ''!
readRequestSafely: stream timing: timing
	"Read request from stream, returning nil when the connection is closed or times out"

	^ [ 
		[ self readRequest: stream timing: timing ]
			on: self readRequestBadExceptionSet
			do: [ :exception | 
				self logServerReadError: exception.
				self writeResponseBad: exception on: stream timing: timing.
				nil ] ]
			on: self readRequestTerminationExceptionSet
			do: [ :exception | 
				self logServerReadError: exception.
				nil ]! !

!ZnMultiThreadedServer methodsFor: 'private' stamp: ''!
readRequestTerminationExceptionSet
	"Return the set of exceptions which, when they occur while reading a request,
	are interpreted as equivalent to a timeout or connection close."

	^ self
		exceptionSet:
			#(#ConnectionClosed #ConnectionTimedOut #PrimitiveFailed)! !

!ZnMultiThreadedServer methodsFor: 'request handling' stamp: ''!
serveConnectionsOn: listeningSocket
	"We wait up to acceptWaitTimeout seconds for an incoming connection.
	If we get one we wrap it in a SocketStream and #executeRequestResponseLoopOn: on it"
		 
	| stream socket |
	socket := listeningSocket waitForAcceptFor: self acceptWaitTimeout.
	socket ifNil: [ ^ self noteAcceptWaitTimedOut ].
	stream := self socketStreamOn: socket.
	[ [ [ self executeRequestResponseLoopOn: stream ]
		ensure: [ self logConnectionClosed: stream. self closeSocketStream: stream ] ]
			ifCurtailed: [ socket destroy ] ]
				forkAt: Processor lowIOPriority
				named: self workerProcessName ! !

!ZnMultiThreadedServer methodsFor: 'private' stamp: ''!
workerProcessName
	^ String streamContents: [ :stream |
		stream nextPutAll: self class name; nextPutAll: ' HTTP worker' ]! !

!ZnMultiThreadedServer methodsFor: 'private' stamp: ''!
writeResponseBad: exception on: stream timing: timing
	"When we failed to parse a request, write a bad request response before closing."

	| response |
	response := ZnResponse new
		statusLine: ZnStatusLine badRequest;
		headers: ZnHeaders defaultResponseHeaders;
		entity: (ZnEntity textCRLF: 'Bad Request ' , exception printString);
		yourself.
	response setConnectionClose.
	self writeResponseSafely: response on: stream timing: timing! !

!ZnMultiThreadedServer methodsFor: 'request handling' stamp: ''!
writeResponseSafely: response on: stream timing: timing
	"Write response to stream, when the connection is closed or times out we ignore this"

	[ self writeResponse: response on: stream timing: timing ]
		on: self writeResponseTerminationExceptionSet
		do: [ :exception | self logServerWriteError: exception ]! !

!ZnMultiThreadedServer methodsFor: 'private' stamp: ''!
writeResponseTerminationExceptionSet
	"Return the set of exceptions which, when they occur while writing a response,
	are interpreted as equivalent to a timeout or connection close."

	^ self exceptionSet: #(#ConnectionClosed #ConnectionTimedOut #PrimitiveFailed)! !

!ZnSingleThreadedServer class methodsFor: 'public' stamp: ''!
adoptAsDefault: server
	"Adopt server as the default instance that we manage.
	If there was a previous default, stop it.
	Note that this is a class instance variable."
	
	self stopDefault.
	default := server! !

!ZnSingleThreadedServer class methodsFor: 'public' stamp: ''!
default
	"Return the default instance that we manage.
	Note that this is a class instance variable."
	
	^ default! !

!ZnSingleThreadedServer class methodsFor: 'public' stamp: ''!
defaultOn: portNumber
	"Return the default instance on a given port,
	Keep a reference to it in a class instance variable.
	If there was no previous default instance, create a new one.
	If there was a previous default instance, reuse it:
	if it was running stop it, change the port if necessary."

	^ default 
		ifNil: [
			default := self on: portNumber ] 
		ifNotNil: [ 
			default stop; port: portNumber; yourself ]! !

!ZnSingleThreadedServer class methodsFor: 'class initialization' stamp: ''!
initialize
	default := nil! !

ZnSingleThreadedServer initialize!

!ZnSingleThreadedServer class methodsFor: 'public' stamp: ''!
on: aNumber
	"Instantiate and return a new listener on a given port,
	send #start to it to start listening."

	^ self new 
		port: aNumber; 
		yourself! !

!ZnSingleThreadedServer class methodsFor: 'public' stamp: ''!
startDefaultOn: portNumber
	"Start and return the default instance on a given port.
	Keep a reference to it in a class instance variable.
	If there was no previous default instance, create a new one.
	If there was a previous default instance, reuse it:
	if it was running stop and start it, effectively restarting it.
	Change the port if necessary.
	Starting the default server will register it automatically."

	^ (self defaultOn: portNumber) 
		start; 
		yourself! !

!ZnSingleThreadedServer class methodsFor: 'public' stamp: ''!
startOn: portNumber
	"Instanciate and return a new listener on a given port and start listening."

	^ (self on: portNumber) 
		start;
		yourself! !

!ZnSingleThreadedServer class methodsFor: 'public' stamp: ''!
stopDefault
	"Stop and unregister the Default instance, if any, and clear it.
	Return the stopped instance, if any."
	
	| server |
	(server := default) ifNotNil: [ 
		default stop.  
		default := nil ].
	^ server! !

!ZnSingleThreadedServer methodsFor: 'constants' stamp: ''!
acceptWaitTimeout
	"How many seconds to wait for a server socket listening for an accept ?"
	
	^ 300! !

!ZnSingleThreadedServer methodsFor: 'request handling' stamp: ''!
augmentResponse: response forRequest: request
	"Our handler has produced response for request, manipulate the response before writing it"
	
	(self useGzipCompressionAndChunking and: [ response contentType isBinary not ])
		ifTrue: [ 
			(request acceptsEncodingGzip and: [ response hasContentEncoding not ]) 
				ifTrue: [ response setContentEncodingGzip ].
			response hasTransferEncoding
				ifFalse: [ response setTransferEncodingChunked ] ]! !

!ZnSingleThreadedServer methodsFor: 'request handling' stamp: ''!
authenticateAndDelegateRequest: request
	"Handle request and return a response.
	If we have a delegate, pass the responsability.
	If we have no delegate, we return a page not found.
	Make sure to pass via our authenticator."

	^ self
		authenticateRequest: request
		do: [ 
			self delegate
				ifNil: [ ZnResponse notFound: request uri ]
				ifNotNil: [ :delegate | delegate handleRequest: request ] ]! !

!ZnSingleThreadedServer methodsFor: 'request handling' stamp: ''!
authenticateRequest: request do: block
	"Validate request and execute block.
	When our authenticator is not nil, pass the responsability"

	^ self authenticator
		ifNil: [ block value ]
		ifNotNil: [ :authenticator | authenticator authenticateRequest: request do: block ]! !

!ZnSingleThreadedServer methodsFor: 'private' stamp: ''!
closeDelegate
	self delegate
		ifNotNil: [ :delegate | 
			(delegate respondsTo: #close)
				ifTrue: [ delegate close ] ]! !

!ZnSingleThreadedServer methodsFor: 'request handling' stamp: ''!
executeOneRequestResponseOn: stream
	"Execute one single HTTP request / response cycle on stream in 3 steps:
	#readRequest: #handleRequest: and #writeResponse:on:
	Indicate that we intend to close the connection afterwards"

	| request response timing|
	ZnCurrentServer
		value: self
		during: [ 
			timing := ZnServerTransactionTiming new.
			request := self readRequest: stream timing: timing.
			response := self handleRequest: request timing: timing.
			response setConnectionClose.
			self writeResponse: response on: stream timing: timing.
			self logServerTransactionRequest: request response: response timing: timing ]! !

!ZnSingleThreadedServer methodsFor: 'request handling' stamp: ''!
handleRequest: request timing: timing
	"Handle request and return a response.
	Do logging and timing. Set a dynamic variable referencing the server.
	If necessary do session management in the response."

	| response initialMilliseconds |
	initialMilliseconds := Time millisecondClockValue.
	ZnCurrentServerSession value: nil.
	response := self handleRequestProtected: request.
	request method = #HEAD
		ifTrue: [ response clearEntity ].
	timing handlerDuration: (self logRequest: request response: response handledStarted: initialMilliseconds).
	ZnCurrentServerSession value 
		ifNotNil: [ :session | self sessionManager setSession: session in: response ].
	^ response! !

!ZnSingleThreadedServer methodsFor: 'request handling' stamp: ''!
handleRequestProtected: request
	"Handle request and return a response.
	If a ZnRespond notification is signaled, return its response directly.
	If an Error is thrown, return a HTTP Server Error response."

	^ [ [ self authenticateAndDelegateRequest: request ]
			on: ZnRespond 
			do: [ :notification | notification response ] ]
		on: Error
		do: [ :exception | 
			self debugMode
				ifTrue: [ exception pass ]
				ifFalse: [
					self logServerHandlerError: exception.
					ZnResponse serverError: exception printString ] ]! !

!ZnSingleThreadedServer methodsFor: 'initialization' stamp: ''!
initialize
	self loggingOn ! !

!ZnSingleThreadedServer methodsFor: 'private' stamp: ''!
initializeServerSocket
	"Set up a new server socket and listen on it"
	
	serverSocket := self bindingAddress
		ifNil: [ ZnNetworkingUtils serverSocketOn: self port ]
		ifNotNil: [ ZnNetworkingUtils serverSocketOn: self port interface: self bindingAddress ].
	"Binding a server socket to port 0 means letting the OS assign a port, fetch it"
	self port isZero 
		ifTrue: [ self optionAt: #port put: serverSocket port ].
	self logServerSocketBound! !

!ZnSingleThreadedServer methodsFor: 'testing' stamp: ''!
isListening
	"Return true when I have a valid server socket listening at the correct port"
	
	^ self serverSocket notNil
		and: [ self serverSocket isValid and: [ self serverSocket localPort = self port ] ]! !

!ZnSingleThreadedServer methodsFor: 'testing' stamp: ''!
isRunning
	"Return true when I am running"
	
	^ self process notNil and: [ self serverSocket notNil ]! !

!ZnSingleThreadedServer methodsFor: 'request handling' stamp: ''!
listenLoop
	"We create a listening Socket, then wait for a connection.
	After each connection we also check that the listening Socket is still valid 
	- if not we just make a recursive call to this method to start over."

	self initializeServerSocket.
	[ [ 
		serverSocket isValid
			 ifFalse: [
				"will trigger #ifCurtailed: block and destroy socket"
				^ self listenLoop ].
		self serveConnectionOn: serverSocket ] repeat ]

		ifCurtailed: [ self releaseServerSocket ]! !

!ZnSingleThreadedServer methodsFor: 'private logging' stamp: ''!
logConnectionAccepted: socket
	logLevel < 3 ifTrue: [ ^ nil ].
	^ (self newLogEvent: ZnConnectionAcceptedEvent)
		address: ([ socket remoteAddress ] on: Error do: [ nil ]);
		emit! !

!ZnSingleThreadedServer methodsFor: 'private logging' stamp: ''!
logConnectionClosed: socketStream
	logLevel < 3 ifTrue: [ ^ nil ].
	socketStream socket ifNil: [ ^ nil ].
	^ (self newLogEvent: ZnServerConnectionClosedEvent)
		address: ([ socketStream socket remoteAddress] on: Error do: [ nil ]);
		emit! !

!ZnSingleThreadedServer methodsFor: 'logging' stamp: ''!
logLevel: integer
	"Set the log level to integer.
	0 - no logging
	1 - simplified transaction logging
	2 - detailed transaction logging
	3 - log everything"
	
	logLevel := integer! !

!ZnSingleThreadedServer methodsFor: 'private logging' stamp: ''!
logRequest: request response: response handledStarted: initialMilliseconds
	^ logLevel < 3
		ifTrue: [ Time millisecondsSince: initialMilliseconds ]
		ifFalse: [ 
			(self newLogEvent: ZnRequestResponseHandledEvent)
				request: request;
				response: response;
				duration: (Time millisecondsSince: initialMilliseconds);
				emit;
				duration ]! !

!ZnSingleThreadedServer methodsFor: 'private logging' stamp: ''!
logRequestRead: request started: initialMilliseconds
	^ logLevel < 3
		ifTrue: [ Time millisecondsSince: initialMilliseconds ]
		ifFalse: [ 
			(self newLogEvent: ZnRequestReadEvent)
				request: request;
				duration: (Time millisecondsSince: initialMilliseconds);
				emit;
				duration ]! !

!ZnSingleThreadedServer methodsFor: 'private logging' stamp: ''!
logResponseWritten: response started: initialMilliseconds
	^ logLevel < 3
		ifTrue: [ Time millisecondsSince: initialMilliseconds ]
		ifFalse: [ 
			(self newLogEvent: ZnResponseWrittenEvent)
				response: response;
				duration: (Time millisecondsSince: initialMilliseconds);
				emit;
				duration ]! !

!ZnSingleThreadedServer methodsFor: 'private logging' stamp: ''!
logServerGeneric: subject
	^ (self newLogEvent: ZnServerGenericLogEvent)
		subject: subject;
		emit! !

!ZnSingleThreadedServer methodsFor: 'private logging' stamp: ''!
logServerHandlerError: exception
	^ (self newLogEvent: ZnServerHandlerErrorEvent)
		exception: exception;
		emit! !

!ZnSingleThreadedServer methodsFor: 'private logging' stamp: ''!
logServerReadError: exception
	logLevel < 3 ifTrue: [ ^ nil ].
	^ (self newLogEvent: ZnServerReadErrorEvent)
		exception: exception;
		emit! !

!ZnSingleThreadedServer methodsFor: 'private logging' stamp: ''!
logServerSocketBound
	logLevel < 1 ifTrue: [ ^ nil ].
	^ (self newLogEvent: ZnServerSocketBoundEvent)
		address: ([ serverSocket address ] on: Error do: [ nil ]);
		port: ([ serverSocket port ] on: Error do: [ nil ]);
		emit! !

!ZnSingleThreadedServer methodsFor: 'private logging' stamp: ''!
logServerSocketReleasedAddress: address port: port
	logLevel < 1 ifTrue: [ ^ nil ].
	^ (self newLogEvent: ZnServerSocketReleasedEvent)
		address: address;
		port: port;
		emit! !

!ZnSingleThreadedServer methodsFor: 'private logging' stamp: ''!
logServerTransactionRequest: request response: response timing: timing
	| logEventClass |
	logLevel < 1 ifTrue: [ ^ nil ].
	logEventClass := logLevel = 1
		ifTrue: [ ZnSimplifiedServerTransactionEvent ]
		ifFalse: [ ZnServerTransactionEvent ].
	^ (self newLogEvent: logEventClass)
		request: request;
		response: response;
		timing: timing;
		emit! !

!ZnSingleThreadedServer methodsFor: 'private logging' stamp: ''!
logServerWriteError: exception
	logLevel < 3 ifTrue: [ ^ nil ].
	^ (self newLogEvent: ZnServerWriteErrorEvent)
		exception: exception;
		emit! !

!ZnSingleThreadedServer methodsFor: 'private logging' stamp: ''!
logStarted
	logLevel < 1 ifTrue: [ ^ nil ].
	^ (self newLogEvent: ZnServerStartedEvent)
		description: self serverProcessName;
		emit! !

!ZnSingleThreadedServer methodsFor: 'private logging' stamp: ''!
logStopped
	logLevel < 1 ifTrue: [ ^ nil ].
	^ (self newLogEvent: ZnServerStoppedEvent)
		description: self serverProcessName;
		emit! !

!ZnSingleThreadedServer methodsFor: 'logging' stamp: ''!
logToTranscript
	"Setup logging to the Transcript.
	If logging was completely off, turn it on"

	logLevel isZero ifTrue: [ self loggingOn ].
	ZnLogEvent logToTranscript ! !

!ZnSingleThreadedServer methodsFor: 'logging' stamp: ''!
loggingOff
	"Turn logging completely off - do not generate log events."
	
	self logLevel: 0! !

!ZnSingleThreadedServer methodsFor: 'logging' stamp: ''!
loggingOn
	"Turn full logging on - generate all log events"
	
	self logLevel: 3! !

!ZnSingleThreadedServer methodsFor: 'private logging' stamp: ''!
newLogEvent: logEventClass
	^ logEventClass new
		serverId: self route;
		yourself! !

!ZnSingleThreadedServer methodsFor: 'private' stamp: ''!
noteAcceptWaitTimedOut
	"The accept wait on a server socket timed out (see #acceptWaitTimeout).
	This can be used as an opportunity to do some periodic tasks."
	
	self periodicTasks! !

!ZnSingleThreadedServer methodsFor: 'accessing' stamp: ''!
onRequestRespond: block
	"Convenience method to quickly set up a delegate.
	On any incoming request, block will be executed with the request 
	as argument to produce a valid response."
	
	self delegate: (ZnValueDelegate with: block)  ! !

!ZnSingleThreadedServer methodsFor: 'private' stamp: ''!
periodicTasks
	"Every #acceptWaitTimeout seconds this method is called.
	Note that during this time the server is not actively listening for connections."
	
	sessionManager ifNotNil: [ sessionManager cleanupInvalidSessions ]! !

!ZnSingleThreadedServer methodsFor: 'printing' stamp: ''!
printOn: stream
	super printOn: stream.
	stream nextPut: $(.
	stream nextPutAll: (self isRunning ifTrue: [ 'running' ] ifFalse: [ 'stopped' ]).
	self port ifNotNil: [ :port | stream space; print: port ].
	self bindingAddress ifNotNil: [ :bindingAddress | stream space; print: bindingAddress ].
	stream nextPut: $)! !

!ZnSingleThreadedServer methodsFor: 'accessing' stamp: ''!
process
	"Return the process that is running my main listening loop.
	Will be nil when I am not running"
	
	^ process! !

!ZnSingleThreadedServer methodsFor: 'request handling' stamp: ''!
readRequest: stream timing: timing
	"Read a request from stream.
	We add a virtual header containing the remote IP address of the client"

	| request initialMilliseconds |
	initialMilliseconds := Time millisecondClockValue.
	request := self withMaximumEntitySizeDo: [ 
		self withMaximumNumberOfDictionaryEntriesDo: [ 
			self withDefaultCharacterEncoderDo: [
				self reader value: stream ] ] ].
	request headers 
		at: ZnConstants remoteAddressHeader
		put: (ZnNetworkingUtils ipAddressToString: stream socket remoteAddress).
	timing requestDuration:  (self logRequestRead: request started: initialMilliseconds).
	^ request! !

!ZnSingleThreadedServer methodsFor: 'private' stamp: ''!
releaseServerSocket
	"Release our server socket"
	
	| address port |
	serverSocket ifNil: [ ^ self ].
	address := [ serverSocket address ] on: Error do: [ nil ].
	port := [ serverSocket port ]on: Error do: [ nil ].
	(Delay forMilliseconds: 10) wait.
	serverSocket destroy.
	serverSocket := nil.
	self logServerSocketReleasedAddress: address port: port ! !

!ZnSingleThreadedServer methodsFor: 'request handling' stamp: ''!
serveConnectionOn: listeningSocket
	"We wait up to acceptWaitTimeout seconds for an incoming connection.
	If we get one we wrap it in a SocketStream and #executeOneRequestResponseOn: on it.
	We do not fork a worker thread/process but stay in the current one for just one request/response cycle."

	| stream socket |
	socket := listeningSocket waitForAcceptFor: self acceptWaitTimeout.
	socket ifNil: [ ^ self noteAcceptWaitTimedOut ].
	stream := self socketStreamOn: socket.
	[ 
	[ self executeOneRequestResponseOn: stream ]
		ensure: [ 
			self logConnectionClosed: stream.
			stream close ] ]
		ifCurtailed: [ 
			socket destroy ]! !

!ZnSingleThreadedServer methodsFor: 'private' stamp: ''!
serverProcessName
	^ String streamContents: [ :stream |
		stream nextPutAll: self class name; nextPutAll: ' HTTP port '; print: self port ]! !

!ZnSingleThreadedServer methodsFor: 'accessing' stamp: ''!
serverSocket
	"Return the server socket that I am using.
	Will be nil when I am not running"
	
	^ serverSocket! !

!ZnSingleThreadedServer methodsFor: 'private' stamp: ''!
socketStreamOn: socket
	self logConnectionAccepted: socket.
	^ ZnNetworkingUtils socketStreamOn: socket! !

!ZnSingleThreadedServer methodsFor: 'public' stamp: ''!
start
	"Start me. I will start listening on my port for incoming HTTP connections.
	If I am running, I will first stop and thus effectively restart"
	
	self stop: false.
	self class default = self ifTrue: [ self register ].
	process := [ [ self listenLoop ] repeat ]
		forkAt: Processor highIOPriority 
		named: self serverProcessName.
	self logStarted.
! !

!ZnSingleThreadedServer methodsFor: 'public' stamp: ''!
stop: unregister
	"Stop me. I will stop listening on my port for incoming HTTP connections.
	If unregister is true, unregister me from the list of managed instances.
	Does nothing when I am not running"
	
	self isRunning ifFalse: [ ^ self ].
	process terminate. 
	process := nil.
	unregister ifTrue: [ self unregister ].
	self closeDelegate.
	self logStopped! !

!ZnSingleThreadedServer methodsFor: 'private' stamp: ''!
withDefaultCharacterEncoderDo: block
	^ self defaultEncoder = ZnDefaultCharacterEncoder value
			ifTrue: block
			ifFalse: [ 
				ZnDefaultCharacterEncoder 
					value: self defaultEncoder  
					during: block ]! !

!ZnSingleThreadedServer methodsFor: 'private' stamp: ''!
withMaximumEntitySizeDo: block
	^ self maximumEntitySize = ZnMaximumEntitySize value
			ifTrue: block
			ifFalse: [ 
				ZnMaximumEntitySize 
					value: self maximumEntitySize  
					during: block ]! !

!ZnSingleThreadedServer methodsFor: 'private' stamp: ''!
withMaximumNumberOfDictionaryEntriesDo: block
	^ self maximumNumberOfDictionaryEntries = ZnMaximumNumberOfDictionaryEntries value
			ifTrue: block
			ifFalse: [ 
				ZnMaximumNumberOfDictionaryEntries 
					value: self maximumNumberOfDictionaryEntries 
					during: block ]! !

!ZnSingleThreadedServer methodsFor: 'request handling' stamp: ''!
writeResponse: response on: stream timing: timing
	"Write response to stream and flush the stream"
	
	| initialMilliseconds |
	initialMilliseconds := Time millisecondClockValue.
	self withDefaultCharacterEncoderDo: [ response writeOn: stream ].
	stream flush.
	timing responseDuration: (self logResponseWritten: response started: initialMilliseconds)  ! !

Object subclass: #ZnServerSession
	instanceVariableNames: 'id attributes created modified'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Support'!

!ZnServerSession commentStamp: '' prior: 0!
I am ZnServerSession.
I represent an HTTP session in the context of a ZnServer.

To get a reference to the current session, send #session to a ZnMessage (request or response).

Part of Zinc HTTP Components.!

!ZnServerSession methodsFor: 'accessing' stamp: ''!
age
	^ DateAndTime now - created! !

!ZnServerSession methodsFor: 'accessing' stamp: ''!
attributeAt: key
	^ attributes at: key! !

!ZnServerSession methodsFor: 'accessing' stamp: ''!
attributeAt: key ifAbsent: block
	^ attributes at: key ifAbsent: block! !

!ZnServerSession methodsFor: 'accessing' stamp: ''!
attributeAt: key ifAbsentPut: block
	^ attributes at: key ifAbsentPut: block! !

!ZnServerSession methodsFor: 'accessing' stamp: ''!
attributeAt: key put: value
	^ attributes at: key put: value! !

!ZnServerSession methodsFor: 'accessing' stamp: ''!
attributeKeys
	^ attributes keys! !

!ZnServerSession methodsFor: 'accessing' stamp: ''!
created
	^ created! !

!ZnServerSession methodsFor: 'accessing' stamp: ''!
id
	^ id! !

!ZnServerSession methodsFor: 'initialize-release' stamp: ''!
id: anObject
	id := anObject! !

!ZnServerSession methodsFor: 'accessing' stamp: ''!
inactivity
	^ DateAndTime now - modified ! !

!ZnServerSession methodsFor: 'initialization' stamp: ''!
initialize
	super initialize.
	attributes := SmallDictionary new.
	created := modified := DateAndTime now! !

!ZnServerSession methodsFor: 'testing' stamp: ''!
isValid
	^ self inactivity < 1 hour! !

!ZnServerSession methodsFor: 'accessing' stamp: ''!
modified
	^ modified! !

!ZnServerSession methodsFor: 'printing' stamp: ''!
printOn: stream
	super printOn: stream.
	id
		ifNotNil: [ 
			stream
				nextPut: $(;
				print: id;
				nextPut: $) ]! !

!ZnServerSession methodsFor: 'accessing' stamp: ''!
removeAttribute: key
	attributes removeKey: key ifAbsent: [ ] ! !

!ZnServerSession methodsFor: 'accessing' stamp: ''!
touch
	modified := DateAndTime now! !

Object subclass: #ZnServerSessionManager
	instanceVariableNames: 'sessions'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Support'!

!ZnServerSessionManager commentStamp: '' prior: 0!
I am ZnServerSessionManager.
I manage a collection of ZnServerSession on behalf of a ZnServer.

Session ids are stored in cookies.

Part of Zinc HTTP Components.!

!ZnServerSessionManager methodsFor: 'accessing' stamp: ''!
cleanupInvalidSessions
	"Remove all expired sessions"
	
	sessions keysAndValuesRemove: [ :key :value | value isValid not ]! !

!ZnServerSessionManager methodsFor: 'accessing' stamp: ''!
cookieName
	^ 'session-id'! !

!ZnServerSessionManager methodsFor: 'initialization' stamp: ''!
initialize
	super initialize.
	sessions := Dictionary new! !

!ZnServerSessionManager methodsFor: 'private' stamp: ''!
newSessionForRequest: request
	| id session |
	id := self newSessionId.
	session := ZnServerSession new.
	session id: id.
	sessions at: id put: session.
	^ session! !

!ZnServerSessionManager methodsFor: 'private' stamp: ''!
newSessionId
	"Create a new session id.
	If the server has a route, append it at the end with a dot as separator."
	
	^ ZnCurrentServer value route
		ifNil: [ UUID new asString36 ]
		ifNotNil: [ :route | 
			String
				streamContents: [ :stream | 
					stream
						<< UUID new asString36;
						nextPut: $.;
						<< route ] ]! !

!ZnServerSessionManager methodsFor: 'accessing' stamp: ''!
removeSessionWithId: id
	"Remove the session with id.
	If there is no such session, do nothing"
	
	^ sessions removeKey: id ifAbsent: [ nil ]! !

!ZnServerSessionManager methodsFor: 'accessing' stamp: ''!
sessionFor: request
	"Bind an existing session to request or create a new session"

	| id |
	id := self sessionIdFromRequest: request.
	sessions
		at: id
		ifPresent: [ :session | 
			session isValid
				ifTrue: [ ^ session touch; yourself ]
				ifFalse: [ sessions removeKey: id ] ].
	^ self newSessionForRequest: request! !

!ZnServerSessionManager methodsFor: 'private' stamp: ''!
sessionIdFromRequest: request
	| cookie |
	cookie := request cookies 
		detect: [ :each | each name = self cookieName ] 
		ifNone: [ ^ nil ].
	^ cookie value! !

!ZnServerSessionManager methodsFor: 'accessing' stamp: ''!
setSession: session in: response
	"Bind session in response"

	| cookie |
	session ifNil: [ ^ self ].
	(cookie := ZnCookie name: self cookieName value: session id)
		path: '/'.
	response addCookie: cookie! !

Object subclass: #ZnServerTransactionTiming
	instanceVariableNames: 'requestDuration handlerDuration responseDuration'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Logging'!

!ZnServerTransactionTiming commentStamp: '' prior: 0!
I am ZnServerTransactionTiming.

I hold timing information related to the handling of an HTTP transaction by a server.
!

!ZnServerTransactionTiming methodsFor: 'accessing' stamp: ''!
handlerDuration
	^ handlerDuration! !

!ZnServerTransactionTiming methodsFor: 'accessing' stamp: ''!
handlerDuration: anObject
	handlerDuration := anObject! !

!ZnServerTransactionTiming methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	requestDuration := handlerDuration := responseDuration := 0! !

!ZnServerTransactionTiming methodsFor: 'accessing' stamp: ''!
requestDuration
	^ requestDuration! !

!ZnServerTransactionTiming methodsFor: 'accessing' stamp: ''!
requestDuration: anObject
	requestDuration := anObject! !

!ZnServerTransactionTiming methodsFor: 'accessing' stamp: ''!
responseDuration
	^ responseDuration! !

!ZnServerTransactionTiming methodsFor: 'accessing' stamp: ''!
responseDuration: anObject
	responseDuration := anObject! !

!ZnServerTransactionTiming methodsFor: 'accessing' stamp: ''!
totalDuration
	"requestDuration is unreliable because it contains idle wait time"
	
	^ handlerDuration + responseDuration ! !

Object subclass: #ZnStatusLine
	instanceVariableNames: 'version code reason'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Core'!

!ZnStatusLine commentStamp: '' prior: 0!
I am an HTTP Status Line.
I consist of an HTTP protocol version, an integer HTTP status code and an HTTP reason string.
I can be used for generating and parsing.

Part of Zinc HTTP Components.!

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
accepted
	^ self code: 202! !

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
badRequest
	^ self code: 400! !

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
code: code
	^ self new
		code: code;
		version: ZnConstants defaultHTTPVersion;
		yourself! !

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
created
	^ self code: 201! !

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
forbidden
	^ self code: 403! !

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
found
	^ self code: 302! !

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
internalServerError
	^ self code: 500! !

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
methodNotAllowed
	^ self code: 405! !

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
movedPermanently
	^ self code: 301! !

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
noContent
	^ self code: 204! !

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
notFound
	^ self code: 404! !

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
notImplemented
	^ self code: 501! !

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
notModified
	^ self code: 304! !

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
ok
	^ self code: 200! !

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
readFrom: stream
	^ self new
		readFrom: stream;
		yourself! !

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
redirect
	^ self code: 302! !

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
seeOther
	^ self code: 303! !

!ZnStatusLine class methodsFor: 'instance creation' stamp: ''!
unauthorized
	^ self code: 401! !

!ZnStatusLine methodsFor: 'comparing' stamp: ''!
= other
	self class = other class ifFalse: [ ^ false ].
	^ self code = other code 
		and: [ self version = other version
			and: [ self reason = other reason ] ]! !

!ZnStatusLine methodsFor: 'accessing' stamp: ''!
code
	^ code! !

!ZnStatusLine methodsFor: 'accessing' stamp: ''!
code: integer
	(ZnConstants httpStatusCodes includesKey: integer) 
		ifFalse: [
			(integer between: 100 and: 599)
				ifFalse: [
					(ZnUnknownHttpStatusCode code: integer) signal ] ].
	code := integer! !

!ZnStatusLine methodsFor: 'comparing' stamp: ''!
hash
	^ (self code hash bitXor: self version hash) bitXor: self reason hash! !

!ZnStatusLine methodsFor: 'printing' stamp: ''!
printCodeAndReasonOn: stream
	self code
		ifNotNil: [
			stream print: self code; space; nextPutAll: self reason ]! !

!ZnStatusLine methodsFor: 'printing' stamp: ''!
printOn: stream
	super printOn: stream.
	self code ifNil: [ ^ self ].
	stream nextPut: $(.
	self printCodeAndReasonOn: stream.
	stream nextPut: $)! !

!ZnStatusLine methodsFor: 'initialize-release' stamp: ''!
readFrom: stream
	| line lineStream httpCode |
	line := (ZnLineReader on: stream) nextLine.
	lineStream := line readStream.
	self version: (lineStream upTo: Character space).
	httpCode := lineStream upTo: Character space.
	httpCode := Integer readFrom: httpCode ifFail: [ (ZnUnknownHttpStatusCode code: httpCode) signal ].
	self code: httpCode.
	self reason: (lineStream upToEnd)
! !

!ZnStatusLine methodsFor: 'accessing' stamp: ''!
reason
	reason ifNil: [ 
		reason := ZnConstants httpStatusCodes 
			at: self code 
			ifAbsent: [ 'Unknown Status Code' ] ].
	^ reason! !

!ZnStatusLine methodsFor: 'accessing' stamp: ''!
reason: string
	reason := string! !

!ZnStatusLine methodsFor: 'accessing' stamp: ''!
version
	^ version! !

!ZnStatusLine methodsFor: 'accessing' stamp: ''!
version: string
	version := ZnConstants knownHTTPVersions 
		detect: [ :each | each = string ] 
		ifNone: [ (ZnUnknownHttpVersion version: string) signal ] ! !

!ZnStatusLine methodsFor: 'writing' stamp: ''!
writeOn: stream
	| writeStream |
	writeStream := ZnBivalentWriteStream on: stream.
	writeStream nextPutAll: self version; space.
	self code printOn: writeStream.
	writeStream space; nextPutAll: self reason; nextPutAll: String crlf ! !

Object subclass: #ZnUserAgentSession
	instanceVariableNames: 'cookiejar credentials currentRealm'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Support'!

!ZnUserAgentSession commentStamp: '' prior: 0!
I represent a session store for an HTTP client instance of ZnUserAgent. 

I hold the cookie jar for the session and any authentication credentials used by the client.

Part of Zinc HTTP Components.!

!ZnUserAgentSession class methodsFor: 'instance creation' stamp: ''!
withCookieJar: aCookieJar
	^ self new cookiejar: aCookieJar; yourself! !

!ZnUserAgentSession methodsFor: 'accessing' stamp: ''!
addCredential: aCredential
	self credentials at: aCredential domain, '@', aCredential realm put: aCredential ! !

!ZnUserAgentSession methodsFor: 'accessing' stamp: ''!
cookieJar
	^ cookiejar ifNil: [ cookiejar := ZnCookieJar new ]! !

!ZnUserAgentSession methodsFor: 'accessing' stamp: ''!
cookiejar: anObject
	cookiejar := anObject! !

!ZnUserAgentSession methodsFor: 'accessing' stamp: ''!
credentialAt: aRealm forUrl: urlObject
	^ self credentials at: urlObject authority, '@', aRealm ifAbsent: [ nil ]! !

!ZnUserAgentSession methodsFor: 'accessing' stamp: ''!
credentials
	^ credentials ifNil: [ credentials := Dictionary new ]! !

!ZnUserAgentSession methodsFor: 'accessing' stamp: ''!
credentials: anObject
	credentials := anObject! !

!ZnUserAgentSession methodsFor: 'accessing' stamp: ''!
credentialsForUrl: urlObject
	^ self credentials values select: [ :cred | urlObject authority includesSubstring: (cred domain) ]! !

!ZnUserAgentSession methodsFor: 'accessing' stamp: ''!
currentRealm
	^ currentRealm! !

!ZnUserAgentSession methodsFor: 'accessing' stamp: ''!
currentRealm: anObject
	currentRealm := anObject! !

!ZnUserAgentSession methodsFor: 'operations' stamp: ''!
resetCookies
	cookiejar ifNotNil: [ cookiejar removeAll ]! !

Object subclass: #ZnUtils
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Support'!

!ZnUtils commentStamp: '' prior: 0!
I am ZnUtils holding various utilities related to HTTP handling.

Part of Zinc HTTP Components.!

!ZnUtils class methodsFor: 'streaming' stamp: ''!
bufferedWriteStreamOn: stream
	^ (ZnBufferedWriteStream on: stream)
		sizeBuffer: self streamingBufferSize;
		yourself! !

!ZnUtils class methodsFor: 'conversions' stamp: ''!
capitalizeString: string
	| capitalExpected |
	capitalExpected := true.
	^ String new: string size streamContents: [ :stream |
		string do: [ :each |
			each isAlphaNumeric
				ifTrue: [
					capitalExpected 
						ifTrue: [ 
							stream nextPut: each asUppercase.
							capitalExpected := false. ]
						ifFalse: [ 
							stream nextPut: each asLowercase ] ] 
		 		ifFalse: [ 
					stream nextPut: each.
					capitalExpected := true ] ] ]! !

!ZnUtils class methodsFor: 'accessing' stamp: ''!
currentProcessID
	"Return a integer of at most 6 digits that can be used as a description for the current active process.
	This is a reduced hash that is not necessarily unique (but uniqueness is very probably)"
	
	^ Processor activeProcess hash \\ 999999! !

!ZnUtils class methodsFor: 'conversions' stamp: ''!
decodeBase64: string
	| bytes |
	bytes := ZnBase64Encoder new decode: string.
	^ ZnNullEncoder new decodeBytes: bytes! !

!ZnUtils class methodsFor: 'conversions' stamp: ''!
encodeBase64: string
	| bytes |
	bytes := ZnNullEncoder new encodeString: string.
	^ ZnBase64Encoder new encode: bytes! !

!ZnUtils class methodsFor: 'conversions' stamp: ''!
httpDate
	^ self httpDate: DateAndTime now! !

!ZnUtils class methodsFor: 'conversions' stamp: ''!
httpDate: object
	"Answers a string formatting object as time stamp for HTTP1.1 (RFC 1123). 
	Eg. Sun, 06 Nov 1994 08:49:37 GMT"

	| utc |
	utc := object asDateAndTime asUTC.
	^ String streamContents: [ :stream |
		stream nextPutAll: utc dayOfWeekAbbreviation; nextPut: $,; space.
		stream nextPutAll: (utc dayOfMonth printStringLength: 2 padded: true).
		stream space; nextPutAll: utc monthAbbreviation; space.
		utc year printOn: stream.
		stream space.
		utc printHMSOn: stream.
		stream nextPutAll: ' GMT' ]! !

!ZnUtils class methodsFor: 'testing' stamp: ''!
isCapitalizedString: string
	| capitalExpected |
	capitalExpected := true.
	string do: [ :each |
		each isAlphaNumeric
			ifTrue: [
				((capitalExpected and: [ each isUppercase ])
					or: [ capitalExpected not and: [ each isLowercase ] ])
					ifTrue: [ capitalExpected := false ]
					ifFalse: [ ^ false ] ] 
		 	ifFalse: [ capitalExpected := true ] ].
	^ true



! !

!ZnUtils class methodsFor: 'streaming' stamp: ''!
nextPutAll: collection on: stream
	| totalWritten toWrite total |
	total := collection size.
	"Do not enter the loop for small sizes"
	total < self streamingBufferSize
		ifTrue: [ ^ stream nextPutAll: collection ].
	totalWritten := 0.
	[ totalWritten < total ] whileTrue: [ 
		toWrite := self streamingBufferSize min: (total - totalWritten).
		stream next: toWrite putAll: collection startingAt: totalWritten + 1.
		totalWritten := totalWritten + toWrite.
		self signalProgress: totalWritten total: total ]! !

!ZnUtils class methodsFor: 'conversions' stamp: ''!
parseHttpDate: string
	"self parseHttpDate: 'Tue, 13 Sep 2011 08:04:49 GMT'."
	"self parseHttpDate: 'Tue, 13-Sep-2011 08:04:51 GMT'."
	"self parseHttpDate: 'Tue Jan 01 00:00:01 2036 GMT'."
	
	| tokens day month year hour minute second months map yearToken |
	tokens := (string findTokens: #( $ $- $: $, )) allButFirst.
	tokens last = 'GMT' ifTrue: [ tokens := tokens allButLast ].
	months := #(jan feb mar apr may jun jul aug sep oct nov dec).
	(Integer readFrom: tokens first ifFail: [ 0 ]) isZero
		ifTrue: [ map := #(2 1 6 3 4 5) ]
		ifFalse: [ map := #(1 2 3 4 5 6) ].
	day := Integer readFrom: (tokens at: map first) ifFail: [ 0 ].
	month := months indexOf: (((tokens at: map second) asLowercase) copyFrom: 1 to: 3) ifAbsent: [ 0 ].
	yearToken := tokens at: map third.
	year := yearToken isAllDigits
		ifTrue: [
			year := yearToken asNumber.
			yearToken size = 2
				ifTrue: [
					year < 50
						ifTrue: [ year + 2000 ]
						ifFalse: [ year + 1900 ] ]
				ifFalse: [ year ] ]
		ifFalse: [ 0 ].
	hour := Integer readFrom: (tokens at: map fourth) ifFail: [ 0 ].
	minute := Integer readFrom: (tokens at: map fifth) ifFail: [ 0 ].
	second := Integer readFrom: (tokens at: map sixth) ifFail: [ 0 ].
	^ DateAndTime year: year month: month day: day hour: hour minute: minute second: second offset: Duration zero! !

!ZnUtils class methodsFor: 'streaming' stamp: ''!
readUpToEnd: inputStream limit: limit
	| species bufferSize buffer totalRead outputStream |
	bufferSize := limit ifNil: [ self streamingBufferSize ] ifNotNil: [ self streamingBufferSize min: limit ].
	species := inputStream isBinary ifTrue: [ ByteArray ] ifFalse: [ String ].
	buffer := species new: bufferSize.
	totalRead := 0.
	outputStream := nil.
	[ inputStream atEnd ] whileFalse: [ | readCount |
		readCount := inputStream readInto: buffer startingAt: 1 count: bufferSize.
		totalRead := totalRead + readCount.
		(limit notNil and: [ totalRead > limit ])
			ifTrue: [ ZnEntityTooLarge signal ].
		outputStream ifNil: [ 
			inputStream atEnd
				ifTrue: [ ^ buffer copyFrom: 1 to: readCount ] 
				ifFalse: [ outputStream := (species new: bufferSize) writeStream ] ].
		outputStream next: readCount putAll: buffer startingAt: 1.
		self signalProgress: totalRead total: limit ].
	^ outputStream ifNil: [ species new ] ifNotNil: [ outputStream contents ]! !

!ZnUtils class methodsFor: 'streaming' stamp: ''!
signalProgress: amount total: total
	ZnSignalProgress enabled
		ifTrue: [ 
			total 
				ifNil: [ 
					HTTPProgress new 
						signal: ('Transferred {1} bytes ...' format: { amount humanReadableSIByteSize }) ]
				ifNotNil: [ 
					HTTPProgress new 
						total: total; 
						amount: amount;
						signal: 'Transferring...' ] ]! !

!ZnUtils class methodsFor: 'streaming' stamp: ''!
streamFrom: inputStream to: outputStream
	| buffer totalRead |
	buffer := (inputStream isBinary ifTrue: [ ByteArray ] ifFalse: [ String ]) new: self streamingBufferSize.
	totalRead := 0.
	[ inputStream atEnd ]
		whileFalse: [ | readCount |
			readCount := inputStream readInto: buffer startingAt: 1 count: buffer size.
			readCount > 0
				ifTrue: [
					totalRead := totalRead + readCount.
					outputStream next: readCount putAll: buffer startingAt: 1 ].
			inputStream atEnd
				ifFalse: [
					self signalProgress: totalRead total: nil. 
					outputStream flush ] ]! !

!ZnUtils class methodsFor: 'streaming' stamp: ''!
streamFrom: inputStream to: outputStream size: totalSize
	| bufferSize buffer leftToRead |
	bufferSize := self streamingBufferSize min: totalSize.
	buffer := (inputStream isBinary ifTrue: [ ByteArray ] ifFalse: [ String ]) new: bufferSize.
	leftToRead := totalSize.
	[ leftToRead > 0 ]
		whileTrue: [ | readCount |
			readCount := bufferSize min: leftToRead.
			inputStream next: readCount into: buffer.
			leftToRead := leftToRead - readCount.
			outputStream next: readCount putAll: buffer startingAt: 1.
			leftToRead > 0 
				ifTrue: [
					self signalProgress: (totalSize - leftToRead) total: totalSize.
					outputStream flush ] ]! !

!ZnUtils class methodsFor: 'streaming' stamp: ''!
streamingBufferSize
	"Return the byte count used as buffer size when streaming"
	
	^ 16384! !

!ZnUtils class methodsFor: 'conversions' stamp: ''!
trimString: string
	"Until String>>#trimBoth is available everywhere"
	
	| left right |
	left := 1.
	right := string size.
	[ left <= right and: [ (string at: left) isSeparator ] ]
		whileTrue: [ left := left + 1 ].
	[ left <= right and: [ (string at: right) isSeparator ] ]
		whileTrue: [ right := right - 1 ].
	^ string copyFrom: left to: right! !

Object subclass: #ZnValueDelegate
	instanceVariableNames: 'object'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-HTTP-Client-Server'!

!ZnValueDelegate commentStamp: '' prior: 0!
I am ZnValueDelegate.
I forward #handleRequest: messages to the object that I wrap using #value:

ZnServer default delegate: 
	(ZnValueDelegate with: [ :request | 
		ZnResponse ok: (ZnEntity with: 'You asked for ', request uri printString) ] ).

Part of Zinc HTTP Components.!

!ZnValueDelegate class methodsFor: 'instance creation' stamp: ''!
with: object
	^ self new
		object: object;
		yourself! !

!ZnValueDelegate methodsFor: 'public' stamp: ''!
handleRequest: request
	^ object value: request! !

!ZnValueDelegate methodsFor: 'initialize-release' stamp: ''!
object: anObject
	object := anObject! !

ZnLogEvent initialize!

ZnHeaders initialize!

ZnServer initialize!

ZnSingleThreadedServer initialize!

ZnNetworkingUtils initialize!

ZnConstants initialize!
